/*
 * linux/drivers/pcmcia/pxa2xx_korebot.c
 *
 * Author:	Pierre Bureau
 * Created:	May,21 2007
 * Copyright:	K-Team SA
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * Originally based upon linux/drivers/pcmcia/pxa2xx_lubbock.c
 *
 * Korebot PCMCIA specific routines.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>

#include <pcmcia/ss.h>

#include <asm/hardware.h>
#include <asm/irq.h>

#include <asm/arch/pxa-regs.h>
#include <asm/arch/korebot.h>

#include "soc_common.h"

#if defined(CONFIG_ARCH_KOREBOT_KTEAM)
/* GPIO Settings for K-TEAM Korebot Board Rev C3 */
#define KOREBOT_CD0             5
#define KOREBOT_CD1             4
#define KOREBOT_RDY_BUSY0       13
#define KOREBOT_RDY_BUSY1       14
#define KOREBOT_RST0            19
#define KOREBOT_RST1            21
#endif /* defined(CONFIG_ARCH_KOREBOT_KTEAM) */

#if defined(CONFIG_ARCH_KOREBOT_LE)
/* GPIO Settings for K-TEAM Korebot Board Rev C3 */
#define KOREBOT_CD0             5
#define KOREBOT_CD1             4
#define KOREBOT_RDY_BUSY0       13
#define KOREBOT_RDY_BUSY1       14
#define KOREBOT_RST0            19
#define KOREBOT_RST1            21
#endif /* defined(CONFIG_ARCH_KOREBOT_LE) */

#define GPIO_READ(gpio) (( (gpio) < 32 ? GPLR0 >> (gpio) : \
                      ((gpio) < 64 ? GPLR1 >> ((gpio)-32) : (GPLR2 >> ((gpio)-64)))) & 1)


#define KOREBOT_S0_CD_IRQ IRQ_GPIO(KOREBOT_CD0)
#define KOREBOT_S1_CD_IRQ IRQ_GPIO(KOREBOT_CD1)
#define KOREBOT_S0_STSCHG_IRQ IRQ_GPIO(KOREBOT_RDY_BUSY0)
#define KOREBOT_S1_STSCHG_IRQ IRQ_GPIO(KOREBOT_RDY_BUSY1)

static struct pcmcia_irqs irqs[] = {
	{ 0, KOREBOT_S0_CD_IRQ, "PCMCIA0 CD" },
//	{ 0, KOREBOT_S0_STSCHG_IRQ, "PCMCIA0 STSCHG" },
#ifndef CONFIG_ARCH_KOREBOT_LE
	{ 1, KOREBOT_S1_CD_IRQ, "PCMCIA1 CD" },
//	{ 1, KOREBOT_S1_STSCHG_IRQ, "PCMCIA1 STSCHG" },
#endif
};

static int korebot_pcmcia_hw_init(struct soc_pcmcia_socket *skt)
{
	/*
	 * Setup default state of GPIO outputs
	 * before we enable them as outputs.
	 */
	GPSR(GPIO48_nPOE) =
		GPIO_bit(GPIO48_nPOE) |
		GPIO_bit(GPIO49_nPWE) |
		GPIO_bit(GPIO50_nPIOR) |
		GPIO_bit(GPIO51_nPIOW) |
		GPIO_bit(GPIO85_nPCE_1) |
		GPIO_bit(GPIO54_nPCE_2);

	pxa_gpio_mode(GPIO48_nPOE_MD);
	pxa_gpio_mode(GPIO49_nPWE_MD);
	pxa_gpio_mode(GPIO50_nPIOR_MD);
	pxa_gpio_mode(GPIO51_nPIOW_MD);
	pxa_gpio_mode(GPIO85_nPCE_1_MD);
	pxa_gpio_mode(GPIO54_nPCE_2_MD);
	pxa_gpio_mode(GPIO79_pSKTSEL_MD);
	pxa_gpio_mode(GPIO55_nPREG_MD);
	pxa_gpio_mode(GPIO56_nPWAIT_MD);
	pxa_gpio_mode(GPIO57_nIOIS16_MD);

	printk( "Korebot CF Init\n");

	/* Output, Enable Reset */
	GPDR(KOREBOT_RST0) |= GPIO_bit(KOREBOT_RST0);
#ifndef CONFIG_ARCH_KOREBOT_LE
	GPDR(KOREBOT_RST1) |= GPIO_bit(KOREBOT_RST1);
#endif
	/* Clear Reset (Reset Low => not active) */
	GPCR(KOREBOT_RST0) = GPIO_bit(KOREBOT_RST0);
#ifndef CONFIG_ARCH_KOREBOT_LE
	GPCR(KOREBOT_RST1) = GPIO_bit(KOREBOT_RST1);
#endif
	/* Card detect */
	/* set IO to input and interrupt to falling edge */
	GPDR(KOREBOT_CD0) &= ~GPIO_bit(KOREBOT_CD0);
#ifndef CONFIG_ARCH_KOREBOT_LE
	GPDR(KOREBOT_CD1) &= ~GPIO_bit(KOREBOT_CD1);
#endif
	/* Card ready */
	GPDR(KOREBOT_RDY_BUSY0) &= ~GPIO_bit(KOREBOT_RDY_BUSY0);
#ifndef CONFIG_ARCH_KOREBOT_LE
	GPDR(KOREBOT_RDY_BUSY1) &= ~GPIO_bit(KOREBOT_RDY_BUSY1);
#endif

	skt->irq = (skt->nr == 0) ? KOREBOT_S0_STSCHG_IRQ : KOREBOT_S1_STSCHG_IRQ;
	printk("korebot_init: skt->irq=%d\n", skt->irq);
	return soc_pcmcia_request_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

static void korebot_pcmcia_hw_shutdown(struct soc_pcmcia_socket *skt)
{
	soc_pcmcia_free_irqs(skt, irqs, ARRAY_SIZE(irqs));
}

//static unsigned long mst_pcmcia_status[2];

static void korebot_pcmcia_socket_state(struct soc_pcmcia_socket *skt,
				    struct pcmcia_state *state)
{
#if 0
	unsigned long status, flip;

	flip = (status ^ mst_pcmcia_status[skt->nr]) & MST_PCMCIA_nSTSCHG_BVD1;

	/*
	 * Workaround for STSCHG which can't be deasserted:
	 * We therefore disable/enable corresponding IRQs
	 * as needed to avoid IRQ locks.
	 */
	if (flip) {
		mst_pcmcia_status[skt->nr] = status;
		if (status & MST_PCMCIA_nSTSCHG_BVD1)
			enable_irq( (skt->nr == 0) ? MAINSTONE_S0_STSCHG_IRQ
						   : MAINSTONE_S1_STSCHG_IRQ );
		else
			disable_irq( (skt->nr == 0) ? MAINSTONE_S0_STSCHG_IRQ
						    : MAINSTONE_S1_STSCHG_IRQ );
	}
#endif

	if(skt->nr == 0)
	{
	  state->detect=!GPIO_READ(KOREBOT_CD0);
	  state->ready=GPIO_READ(KOREBOT_RDY_BUSY0);
	  /* we don't have bvd, we are never write-protected, we are always at 3V */
	  state->bvd1=1;//GPIO_READ(KOREBOT_BVD1);
	  state->bvd2=1;//GPIO_READ(KOREBOT_BVD2);
	  state->wrprot=0;
	  state->vs_3v=1;
	  state->vs_Xv=0;
	}
#ifndef CONFIG_ARCH_KOREBOT_LE
	if(skt->nr == 1)
	{
	  state->detect=!GPIO_READ(KOREBOT_CD1);
	  state->ready=GPIO_READ(KOREBOT_RDY_BUSY1);
	  /* we don't have bvd, we are never write-protected, we are always at 3V */
	  state->bvd1=1;//GPIO_READ(KOREBOT_BVD1);
	  state->bvd2=1;//GPIO_READ(KOREBOT_BVD2);
	  state->wrprot=0;
	  state->vs_3v=1;
	  state->vs_Xv=0;
	}
#endif
}

static int korebot_pcmcia_configure_socket(struct soc_pcmcia_socket *skt,
				       const socket_state_t *state)
{
	unsigned long power = 0;
	int ret = 0;

	switch (state->Vcc) {
	case 0:  power |= KOREBOT_PCMCIA_PWR_VCC_0;  break;
	case 33: power |= KOREBOT_PCMCIA_PWR_VCC_33; break;
	case 50: power |= KOREBOT_PCMCIA_PWR_VCC_50; break;
	default:
		 printk(KERN_ERR "%s(): bad Vcc %u\n",
				 __FUNCTION__, state->Vcc);
		 ret = -1;
	}

	switch (state->Vpp) {
	case 0:   power |= KOREBOT_PCMCIA_PWR_VPP_0;   break;
	case 120: power |= KOREBOT_PCMCIA_PWR_VPP_120; break;
	default:
		  if(state->Vpp == state->Vcc) {
			  power |= KOREBOT_PCMCIA_PWR_VPP_VCC;
		  } else {
			  printk(KERN_ERR "%s(): bad Vpp %u\n",
					  __FUNCTION__, state->Vpp);
			  ret = -1;
		  }
	}

	if (state->flags & SS_RESET)
	{
	  printk(KERN_NOTICE "%s(): Reset on sock %d\n", __FUNCTION__, skt->nr);
	  if(skt->nr)
	    GPSR(KOREBOT_RST1) = GPIO_bit(KOREBOT_RST1);
	  else
	    GPSR(KOREBOT_RST0) = GPIO_bit(KOREBOT_RST0);
	}
	else
	{
	  printk(KERN_NOTICE "%s(): Reset off sock %d\n", __FUNCTION__, skt->nr);
	  if(skt->nr)
	    GPCR(KOREBOT_RST1) = GPIO_bit(KOREBOT_RST1);
	  else
	    GPCR(KOREBOT_RST0) = GPIO_bit(KOREBOT_RST0);
	}

	return ret;
}

static void korebot_pcmcia_socket_init(struct soc_pcmcia_socket *skt)
{
}

static void korebot_pcmcia_socket_suspend(struct soc_pcmcia_socket *skt)
{
}

static struct pcmcia_low_level mst_pcmcia_ops = {
	.owner			= THIS_MODULE,
	.hw_init		= korebot_pcmcia_hw_init,
	.hw_shutdown		= korebot_pcmcia_hw_shutdown,
	.socket_state		= korebot_pcmcia_socket_state,
	.configure_socket	= korebot_pcmcia_configure_socket,
	.socket_init		= korebot_pcmcia_socket_init,
	.socket_suspend		= korebot_pcmcia_socket_suspend,
#ifndef CONFIG_ARCH_KOREBOT_LE
	.nr			= 1,
#else
	.nr			= 2,
#endif
};

static struct platform_device *korebot_pcmcia_device;

static int __init korebot_pcmcia_init(void)
{
	int ret;

	korebot_pcmcia_device = platform_device_alloc("pxa2xx-pcmcia", -1);
	if (!korebot_pcmcia_device)
		return -ENOMEM;

	korebot_pcmcia_device->dev.platform_data = &mst_pcmcia_ops;

	ret = platform_device_add(korebot_pcmcia_device);

	if (ret)
		platform_device_put(korebot_pcmcia_device);

	return ret;
}

static void __exit korebot_pcmcia_exit(void)
{
	platform_device_unregister(korebot_pcmcia_device);
}

fs_initcall(korebot_pcmcia_init);
module_exit(korebot_pcmcia_exit);

MODULE_LICENSE("GPL");
