/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: function_tester.c
 * Author: Biao Jia		2/01/02
 * Description:
 *    Function side driver suitable for testing.
 *    EP_ADDR can be changed to do IN or OUT transfer
 *    EP_TYPE can be changed to do control, iso, bulk and interrupt transfer
 */
#include "os.h"
#include "stack_urb.h"

#include "otg.h"
#include "test_t.h"
#include "otg_device_request.h"
#include "otg_fd.h"

#ifdef SIM_OTG
#include "otg_chip.h"
#include "otg_sim_device.h"
#endif

#define FUNCTION_TESTER_MAJOR 183
#define PKG_MAX	8
#define BUF_MAX	64
#define EP_ADDR	(0x01)	/* 0x81 - IN, 0x01 - OUT */
#define EP_TYPE	1		/* 0 - control, 1 - iso, 2 - bulk, 3 - interrupt */

static test_t *device = 0;
static otg_urb_t function_tester_urb;
static char function_tester_buffer[BUF_MAX];
static DECLARE_WAIT_QUEUE_HEAD(function_tester_wait);
static int prev_length = 0;
static int prev_pos = 0;

static otg_fdevice_t tester_device;

static ssize_t
otg_function_tester_Read(struct file* file, char* buf, size_t size, loff_t* ppos)
{
	int length;
	int sz;

	info("Read: size=%d buf=%p", size, buf);
	length = 0;
	if (prev_length)
	{
		sz = prev_length;
		if (sz > size)
		{
			sz = size;
		}

		if (OS_CopyToUser(buf, &function_tester_buffer[prev_pos], prev_length))
		{
			err("Read: copy_to_user failed");
			return -EFAULT;
		}

		length = sz;
		prev_length -= sz;
		prev_pos += sz;
	}

	while (!length)
	{
		otg_fd_Submit(&function_tester_urb);
		interruptible_sleep_on(&function_tester_wait);
		if (signal_pending(current))
		{
			return -EINTR;
		}

		if (function_tester_urb.status)
		{
			err("Read: transfer failed with %d", function_tester_urb.status);
			return function_tester_urb.status;
		}

		sz = function_tester_urb.actual_length;
		info("Read: sz=%d bytes transfered", sz);
		if (sz > size)
		{
			prev_length = sz - size;
			prev_pos = sz - prev_length;
			sz = prev_pos;
		}

		if (OS_CopyToUser(buf, function_tester_buffer, sz))
		{
			err("Read: copy_to_user failed");
			return -EFAULT;
		}

		length = sz;
	}

	info("Read: length=%d", length);
	return length;
}

static ssize_t
otg_function_tester_Write(struct file* fp, const char* buf, size_t size, loff_t* ppos)
{
	int length;
	info("Write: size=%d buf=%p", size, buf);

	length = 0;
	while (length < size)
	{
		int sz;

info("Write: size=%d length=%d", size, length);
		sz = size - length;
		if (sz > BUF_MAX)
		{
			sz = BUF_MAX;
		}

		if (OS_CopyFromUser(function_tester_buffer, &buf[length], sz))
		{
			err("Write: copy_from_user failed");
			return -EFAULT;
		}

		function_tester_urb.transfer_length = sz;

		otg_fd_Submit(&function_tester_urb);
		interruptible_sleep_on(&function_tester_wait);
		if (signal_pending(current))
		{
			break;
		}

		if (function_tester_urb.status)
		{
			err("Read: transfer failed with %d", function_tester_urb.status);
			return function_tester_urb.status;
		}

		length += function_tester_urb.actual_length;
	}

	if (length && !(length % PKG_MAX))
	{
		info("Write: send a zero sized packet ");
		/* Send a zero length package */
		function_tester_urb.transfer_length = 0;

		otg_fd_Submit(&function_tester_urb);
		interruptible_sleep_on(&function_tester_wait);
		if (signal_pending(current))
		{
			return -EINTR;
		}
	}

	return length;
}

static int
otg_function_tester_Open(struct inode* inode, struct file *filp)
{
	int minor;
	
	info("Open: called");
	minor = MINOR(inode->i_rdev);

	prev_length = 0;

	return 0;
}

static int
otg_function_tester_Release(struct inode* inode, struct file *filp)
{
	info("Release: called");
	return 0;
}

static struct file_operations function_tester_fops =
{
	read: otg_function_tester_Read,
	write: otg_function_tester_Write,
	open: otg_function_tester_Open,
	release: otg_function_tester_Release,
};

static void
otg_function_tester_CallBack(otg_urb_t *urb)
{
	info("CallBack: urb=%p actual_length=%d status=0x%0x",
		urb, urb->actual_length, urb->status);

	wake_up(&function_tester_wait);
/*	wake_up_interruptible(&function_tester_wait); */
}

static int
otg_function_tester_StandardRequest(otg_device_request_t *request, void *private)
{
	info("StandardRequest: bmRequestType=0x%0x bRequest=0x%0x",
		request->bmRequestType, request->bRequest);
	info("       wValue=0x%0x wIndex=0x%0x wLength=0x%0x",
		request->wValue, request->wIndex, request->wLength);
	
	request->wLength = 0;

	return 0;
}

static int
otg_function_tester_ClassRequest(otg_device_request_t *request, void *private)
{
	info("ClassRequest: bmRequestType=0x%0x bRequest=0x%0x",
		request->bmRequestType, request->bRequest);
	info("       wValue=0x%0x wIndex=0x%0x wLength=0x%0x",
		request->wValue, request->wIndex, request->wLength);
	
	request->wLength = 0;

	return 0;
}

static int
otg_function_tester_VendorRequest(otg_device_request_t *request, void *private)
{
	info("VendorRequest: bmRequestType=0x%0x bRequest=0x%0x",
		request->bmRequestType, request->bRequest);
	info("		wValue=0x%0x wIndex=0x%0x wLength=0x%0x",
		request->wValue, request->wIndex, request->wLength);
	
	request->wLength = 0;

	return 0;
}

static int
otg_function_tester_start(void)
{
	test_t *dev;
	int rc;
	int len;
	int strlen;

	dev = device;
	dev->deviceDesc[0] = otg_device_desc_Max; /* bLength */
	dev->deviceDesc[1] = 0x01; /* bDescriptorType, device */
	dev->deviceDesc[2] = 0x10; /* bcdUSB lower byte */
	dev->deviceDesc[3] = 0x01; /* bcdUSB higher byte */
	dev->deviceDesc[4] = 0x00; /* bDeviceClass, hid */
	dev->deviceDesc[5] = 0x00; /* bDeviceSubClass */
	dev->deviceDesc[6] = 0x00; /* bDeviceProtocol */
	dev->deviceDesc[7] = 0x08; /* bMaxPacketSize0, 64 bytes */
	dev->deviceDesc[8] = 0x11; /* idVendor lower byte */
	dev->deviceDesc[9] = 0x55; /* idVendor higher byte */
	dev->deviceDesc[10] = 0x00; /* idProduct lower byte */
	dev->deviceDesc[11] = 0x10; /* idProduct higher byte */
	dev->deviceDesc[12] = 0x34; /* bcdDevice lower byte */
	dev->deviceDesc[13] = 0x12; /* bcdDevice higher byte */
	dev->deviceDesc[14] = 0x01; /* iManufacturer */
	dev->deviceDesc[15] = 0x02; /* iProduct */
	dev->deviceDesc[16] = 0x03; /* iSeriaNumber */
	dev->deviceDesc[17] = 0x01; /* bNumConfigurations */

	len = 0;
	dev->configDesc[len++] = otg_config_desc_Max; /* bLength */
	dev->configDesc[len++] = 0x02; /* bDescriptorType, configuration */
	dev->configDesc[len++] = 0x00; /* wTotalLength lower byte */
	dev->configDesc[len++] = 0x00; /* wTotalLength higher byte */
	dev->configDesc[len++] = 0x01; /* bNumInterfaces */
	dev->configDesc[len++] = 0x01; /* bConfigurationValue */
	dev->configDesc[len++] = 0x00; /* iConfiguration */
	dev->configDesc[len++] = 0xA0; /* bmAttributes, Bus powered,Remote Wakeup */
	dev->configDesc[len++] = 0x01; /* MaxPower */

	dev->configDesc[len++] = otg_interface_desc_Max;
	dev->configDesc[len++] = 0x04; /* bDescriptorType, Interface */
	dev->configDesc[len++] = 0x00; /* bInterfaceNumber */
	dev->configDesc[len++] = 0x00; /* bAlternateSetting */
	dev->configDesc[len++] = 0x01; /* bNumEndpoints */
	dev->configDesc[len++] = 0xF0; /* bInterfaceClass, random */
	dev->configDesc[len++] = 0x01; /* bInterfaceSubClass, boot inferface */
	dev->configDesc[len++] = 0x01; /* bInterfaceProtocol */
	dev->configDesc[len++] = 0x00; /* iInterface */

	/* Interrupt In endpoint */
	dev->configDesc[len++] = otg_ep_desc_Max;
	dev->configDesc[len++] = 0x05; /* bDescriptorType, Endpoint */
	dev->configDesc[len++] = EP_ADDR; /* bEndpointAddress */
	dev->configDesc[len++] = EP_TYPE; /* bmAttributes */
	dev->configDesc[len++] = PKG_MAX & 0xFF; /* wMaxPacketSize lower byte */
							/* wMaxPacketSize higher byte */
	dev->configDesc[len++] = (PKG_MAX >> 8) & 0xFF;
	dev->configDesc[len++] = 0xF4; /* bInterval */

	/* Fillin wTotalLength */
	dev->configDesc[otg_config_desc_wTotalLength] = len & 0xFF;
	dev->configDesc[otg_config_desc_wTotalLength + 1] = (len >> 8) & 0xFF;

	dev->strDesc[0] = 4;	/* bLength */
	dev->strDesc[1] = 3;	/* bDescriptorType, String */
	dev->strDesc[2] = 0x09; /* LangID codes, English */
	dev->strDesc[3] = 0x00;

	strlen = 4;
	dev->strDesc[strlen++] = 8;	/* bLength */
	dev->strDesc[strlen++] = 3;	/* bDescriptorType, String */
	dev->strDesc[strlen++] = 'T'; /* T */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'D'; /* D */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'I'; /* I */
	dev->strDesc[strlen++] = 0x00;

	dev->strDesc[strlen++] = 32;	/* bLength */
	dev->strDesc[strlen++] = 3;		/* bDescriptorType, String */
	dev->strDesc[strlen++] = 'O'; /* O */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'T'; /* T */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'G'; /* G */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '2'; /* 2 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '4'; /* 4 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '3'; /* 3 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = ' '; /* space */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'F'; /* F */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'u'; /* u */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'n'; /* n */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'c'; /* c */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 't'; /* t */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'i'; /* i */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'o'; /* o */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = 'n'; /* n */
	dev->strDesc[strlen++] = 0x00;

	dev->strDesc[strlen++] = 10;	/* bLength */
	dev->strDesc[strlen++] = 3;		/* bDescriptorType, String */
	dev->strDesc[strlen++] = '1'; /* 1 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '0'; /* 0 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '0'; /* 0 */
	dev->strDesc[strlen++] = 0x00;
	dev->strDesc[strlen++] = '0'; /* 0 */
	dev->strDesc[strlen++] = 0x00;

	otg_urb_Init
	(
		&function_tester_urb,
		EP_ADDR,
		function_tester_buffer,
		PKG_MAX,
		otg_function_tester_CallBack
	);

	otg_urb_SetDevice(&function_tester_urb, &tester_device);

	OS_MemSet(&tester_device, 0, sizeof(tester_device));
	tester_device.strDesc = dev->strDesc;
	tester_device.strDescLength = strlen;
	tester_device.deviceDesc = dev->deviceDesc;
	tester_device.deviceDescLength = otg_device_desc_Max;
	tester_device.configDesc = dev->configDesc;
	tester_device.configDescLength = len;
	tester_device.standardSetup = otg_function_tester_StandardRequest;
	tester_device.classSetup = otg_function_tester_ClassRequest;
	tester_device.vendorSetup = otg_function_tester_VendorRequest;

	rc = otg_fd_Register(&tester_device);
	if (rc < 0)
	{
		err("Register failed");
		return -1;
	}

	init_waitqueue_head(&function_tester_wait);

	return 0;
}

static int
otg_function_tester_InitOtg(void)
{
	device = OS_Malloc(sizeof(test_t));
	if (!device)
	{
		return -1;
	}

	OS_MemSet(device, 0, sizeof(test_t));

	return 0;
}

static void
otg_function_tester_ExitOtg(void)
{
	otg_fd_Unregister(&tester_device);

	if (device)
	{
		OS_Free(device);
		device = 0;
	}
}

static int __init otg_function_tester_init(void)
{
#ifdef SIM_OTG
	otg_chip_t *chip;
#endif
	int rc;

	rc = otg_function_tester_InitOtg();
	if (rc)
	{
		return rc;
	}

#if 0 // TODO: fix devfs
    if (devfs_register_chrdev(FUNCTION_TESTER_MAJOR, "function_tester", &function_tester_fops))
	{
		err("unable t get major %d for function_tester devices", FUNCTION_TESTER_MAJOR);
		otg_function_tester_ExitOtg();
		return -EBUSY;
	}
#endif

#ifdef SIM_OTG
	chip = otg_chip_GetChip();
	chip->ports[1] = otg_sim_device_New();
	if (!chip->ports[1])
	{
		info("Failed to allocate memory");
		return 0;
	}
#endif

	rc = otg_function_tester_start();
	if (rc)
	{
		err("OTG function tester failed");
		otg_function_tester_ExitOtg();
	}

#ifdef SIM_OTG
	otg_chip_RhPortStatusChange(1);
#endif

	return 0;
}

static void __exit otg_function_tester_exit(void)
{
#ifdef SIM_OTG
	otg_chip_t *chip;
#endif

	devfs_unregister_chrdev(FUNCTION_TESTER_MAJOR, "function_tester");
	otg_function_tester_ExitOtg();

#ifdef SIM_OTG
	chip = otg_chip_GetChip();
	if (chip->ports[1])
	{
		otg_chip_RhPortStatusChange(1);
		otg_sim_device_Delete(chip->ports[1]);
		chip->ports[1] = 0;
	}
#endif
}

module_init(otg_function_tester_init);
module_exit(otg_function_tester_exit);
