/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: host_tester.c
 * Author: Biao Jia		2/01/02
 * Description:
 *    Host side driver used to drive function_tester.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

#define HOST_TESTER_MAJOR	182
#define HOST_TESTER_MINOR	32

#define BUF_MAX				(16 * 1024) /* has to be 2^n */
#define PACKET_SIZE_MAX		1024
#define BULK_MAX			256
#define ISO_PACKET_MAX		16

static char buffer[BUF_MAX];
static int read_pos = 0;
static int write_pos = 0;
static struct usb_device *device = 0;
static struct usb_endpoint_descriptor *ep = 0;
static struct urb *host_tester_urb = 0;
static wait_queue_head_t host_tester_wait;
static int submitted = 0;
static int no_more_data = 0;

static char bulk_buffer[PACKET_SIZE_MAX * ISO_PACKET_MAX];

static int stop = 0;

static ssize_t
host_tester_ReadBuffer(char *buf, size_t size, int to_user)
{
	int length;
	int sz;
	int remain;

	info("ReadBuffer: size=%d buf=%p", size, buf);
	length = (write_pos - read_pos) & (BUF_MAX - 1);
	if (length > size)
	{
		length = size;
	}

	sz = 0;
	if ((read_pos + length) >= BUF_MAX)
	{
		sz = BUF_MAX - read_pos;
		if (to_user)
		{
			if (OS_CopyToUser(buf, &buffer[read_pos], sz))
			{
				err("ReadBuffer: copy_to_user failed");
				return -EFAULT;
			}
		}
		else
		{
			memcpy(buf, &buffer[read_pos], sz);
		}

		read_pos = 0;
	}

	remain = length - sz;
	if (remain)
	{
		if (to_user)
		{
			if (OS_CopyToUser(&buf[sz], &buffer[read_pos], remain))
			{
				err("ReadBuffer: copy_to_user failed");
				return -EFAULT;
			}
		}
		else
		{
			memcpy(&buf[sz], &buffer[read_pos], remain);
		}

		read_pos += remain;
	}

	info("ReadBuffer: length=%d read_pos=%d write_pos=%d",
		length, read_pos, write_pos);
	return length;
}

static ssize_t
host_tester_WriteBuffer(const char *buf, size_t size, int from_user)
{
	int length;
	int sz;
	int remain;

	info("WriteBuffer: size=%d buf=%p", size, buf);
	length = (read_pos - write_pos - 1) & (BUF_MAX - 1);
	if (length > size)
	{
		length = size;
	}

	sz = 0;
	if ((write_pos + length) >= BUF_MAX)
	{
		sz = BUF_MAX - write_pos;
		if (from_user)
		{
			if (OS_CopyFromUser(&buffer[write_pos], buf, sz))
			{
				err("WriteBuffer: copy_from_user failed");
				return -EFAULT;
			}
		}
		else
		{
			memcpy(&buffer[write_pos], buf, sz);
		}
		write_pos = 0;
	}

	remain = length - sz;
	if (remain)
	{
		if (from_user)
		{
			if (OS_CopyFromUser(&buffer[write_pos], &buf[sz], remain))
			{
				err("WriteBuffer: copy_from_user failed");
				return -EFAULT;
			}
		}
		else
		{
			memcpy(&buffer[write_pos], &buf[sz], remain);
		}

		write_pos += remain;
	}

	info("WriteBuffer: length=%d read_pos=%d write_pos=%d",
		length, read_pos, write_pos);
	return length;
}

void host_tester_Transfer(void);

/* Maybe called from interrupt handler.
 * Can't unlink urb.
 */
static void
host_test_CallBack(struct urb *urb)
{
	int length;
	int i;

	if (urb->status)
	{
		err("CallBack: status=%d", urb->status);

		if (!urb->dev)
		{
			/* Called with unlink_urb */
			return;
		}

		wake_up_interruptible(&host_tester_wait);

		no_more_data = 1;

		return;
	}
	else if (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
	{
		if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) !=
			USB_ENDPOINT_XFER_ISOC)
		{
			length = host_tester_WriteBuffer(bulk_buffer,
						urb->actual_length, 0);
		}
		else
		{
			iso_packet_descriptor_t *packets;
			
			packets = urb->iso_frame_desc;
			for(i = 0; i < host_tester_urb->number_of_packets; i++)
			{
				info("CallBack: %d status=%d length=%d",
					i, packets[i].status, packets[i].actual_length);

				if (packets[i].status)
				{
					info("CallBack: %d status=%d", i, packets[i].status);
					continue;
				}

				length = host_tester_WriteBuffer
						 (
						 	&bulk_buffer[packets[i].offset],
							packets[i].actual_length,
							0
						 );

				if (length <= 0)
				{
					break;
				}
			}
		}
	}
	else
	{
		if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) ==
			USB_ENDPOINT_XFER_INT)
		{
			length = host_tester_ReadBuffer(bulk_buffer, BULK_MAX, 0);
			if (length < 0)
			{
				err("CallBack: error %d ReadBuffer", length);
			}
		}
	}
	
	info("CallBack: submitted=%d", submitted);
	if ((ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK) != 
		USB_ENDPOINT_XFER_INT)
	{
		submitted = 0;

		usb_free_urb(host_tester_urb);
		host_tester_urb = 0;
	}

	wake_up_interruptible(&host_tester_wait);
}

static void
host_tester_Bulk(void)
{
	int length;
	int pipe;

	if (submitted)
	{
		return;
	}

	host_tester_urb = usb_alloc_urb(0);
	if (!host_tester_urb)
	{
		err("Bulk: can't alloc urb");
		return;
	}

	if (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
	{
		length = BULK_MAX;
		pipe = usb_rcvbulkpipe(device, ep->bEndpointAddress);
	}
	else
	{
		length = host_tester_ReadBuffer(bulk_buffer, BULK_MAX, 0);
		if (length <= 0)
		{
			return;
		}

		pipe = usb_sndbulkpipe(device, ep->bEndpointAddress);
	}


	FILL_BULK_URB
	(
		host_tester_urb,
		device,
		pipe,
		bulk_buffer,
		length,
		host_test_CallBack,
		0
	);

	usb_submit_urb(host_tester_urb);
	submitted = 1;
}

static void
host_tester_Intr(void)
{
	int length;
	int pipe;

	if (submitted)
	{
		return;
	}

	host_tester_urb = usb_alloc_urb(0);
	if (!host_tester_urb)
	{
		err("Intr: can't alloc urb");
		return;
	}

	if (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
	{
		length = BULK_MAX;
		pipe = usb_rcvintpipe(device, ep->bEndpointAddress);
	}
	else
	{
		length = host_tester_ReadBuffer(bulk_buffer, BULK_MAX, 0);
		if (length <= 0)
		{
			return;
		}

		pipe = usb_sndintpipe(device, ep->bEndpointAddress);
	}

	FILL_INT_URB
	(
		host_tester_urb,
		device,
		pipe,
		bulk_buffer,
		length,
		host_test_CallBack,
		0,
		ep->bInterval
	);

	usb_submit_urb(host_tester_urb);
	submitted = 1;
}

static void
host_tester_Iso(void)
{
	int length;
	int pipe;
	int frames;
	int i;
	iso_packet_descriptor_t *packets;

	if (submitted)
	{
		return;
	}

	host_tester_urb = usb_alloc_urb(ISO_PACKET_MAX);
	if (!host_tester_urb)
	{
		err("Iso: can't alloc urb");
		return;
	}

	packets = host_tester_urb->iso_frame_desc;
	if (ep->bEndpointAddress & USB_ENDPOINT_DIR_MASK)
	{
		length = 0;
		for(i = 0; i < ISO_PACKET_MAX; i++)
		{
			packets[i].offset = length;
			packets[i].length = ep->wMaxPacketSize;
			length += ep->wMaxPacketSize;
		}

		frames = ISO_PACKET_MAX;
		pipe = usb_rcvisocpipe(device, ep->bEndpointAddress);
	}
	else
	{
		int sz;

		frames = 0;
		length = 0;
		for(i = 0; i < ISO_PACKET_MAX; i++)
		{
			sz = host_tester_ReadBuffer(&bulk_buffer[length],
						ep->wMaxPacketSize, 0);

			if (sz <= 0)
			{
				break;
			}

			packets[i].offset = length;
			packets[i].length = sz;

			length += sz;
			frames++;
		}

		if (!frames)
		{
			return;
		}

		pipe = usb_sndisocpipe(device, ep->bEndpointAddress);
	}

	spin_lock_init(&host_tester_urb->lock);
	host_tester_urb->dev = device;
	host_tester_urb->pipe = pipe;
	host_tester_urb->transfer_buffer = bulk_buffer;
	host_tester_urb->transfer_buffer_length = length;
	host_tester_urb->complete = host_test_CallBack;
	host_tester_urb->context = 0;
	host_tester_urb->transfer_flags = USB_ISO_ASAP;
	host_tester_urb->number_of_packets = frames;

	for(i = 0; i < frames; i++)
	{
		info("Iso: %d offset=0x%0x length=%d", i, 
			packets[i].offset, packets[i].length);
	}

	usb_submit_urb(host_tester_urb);
	submitted = 1;
}

void
host_tester_Transfer(void)
{
	switch (ep->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
	{
		case USB_ENDPOINT_XFER_ISOC:
			host_tester_Iso();
			break;
		case USB_ENDPOINT_XFER_BULK:
			host_tester_Bulk();
			break;
		case USB_ENDPOINT_XFER_INT:
			host_tester_Intr();
			break;
	}
}

static ssize_t
host_tester_Read(struct file *fp, char *buf, size_t size, loff_t *ppos)
{
	int length;
	int sz;

	if (!ep)
	{
		return -ENODEV;
	}

	if (stop)
	{
		return 0;
	}

	info("Read: buf=%p size=%d", buf, size);
	length = BUF_MAX;
	if (length > size)
	{
		length = size;
	}

	sz = 0;
	while (!sz)
	{
		sz = host_tester_ReadBuffer(buf, length, 1);
		info("Read: %d bytes read. submitted=%d", sz, submitted);
		if (sz)
		{
			break;
		}

		if (submitted)
		{
			if (signal_pending(current))
			{
				return -EINTR;
			}

			interruptible_sleep_on(&host_tester_wait);
			if (no_more_data)
			{
				return 0;
			}
		}
		else
		{
			host_tester_Transfer();
		}
	}

	if (sz > 24)
	{
		stop = 1;
	}

	return sz;
}

static ssize_t
host_tester_Write(struct file *fp, const char *buf, size_t size, loff_t *ppos)
{
	int length;

	if (!ep)
	{
		return -ENODEV;
	}

	info("Write: buf=%p size=%d", buf, size);
	length = 0;
	while (length < size)
	{
		int sz;

		sz = host_tester_WriteBuffer(&buf[length], size - length, 1);
		if (sz < 0)
		{
			return sz;
		}

		length += sz;
		if (sz > 0)
		{
			continue;
		}

		if (submitted)
		{
			if (signal_pending(current))
			{
				return -EINTR;
			}

			interruptible_sleep_on(&host_tester_wait);

			if (!ep)
			{
				return -ENODEV;
			}

			if (no_more_data)
			{
				return -EREMOTEIO;
			}
		}
		else
		{
			host_tester_Transfer();
		}
	}

	if (!submitted)
	{
		if (!ep)
		{
			return -ENODEV;
		}

		host_tester_Transfer();
	}

	return length;
}

static int
host_tester_Open(struct inode *inode, struct file *fp)
{
	info("Open: called");
	read_pos = 0;
	write_pos = 0;
	no_more_data = 0;

	return 0;
}

static int
host_tester_Release(struct inode *inode, struct file *fp)
{
	info("Release: called");
	no_more_data = 1;

	return 0;
}

static struct file_operations host_tester_fops = 
{
	owner:	THIS_MODULE,
	read:	host_tester_Read,
	write:	host_tester_Write,
	open:	host_tester_Open,
	release:	host_tester_Release
};

static void *
host_tester_Probe
(
	struct usb_device *dev,
	unsigned int ifnum,
	const struct usb_device_id *id
)
{
	ep = dev->actconfig->interface[ifnum].altsetting[0].endpoint;
	device = dev;
	submitted = 0;

	init_waitqueue_head(&host_tester_wait);
	info("Probe: a test device is found");

	return buffer;
}

static void
host_tester_Disconnect(struct usb_device *dev, void *ptr)
{
	ep = 0;
	device = 0;

	if (submitted)
	{
		if (host_tester_urb)
		{
			usb_unlink_urb(host_tester_urb);
			usb_free_urb(host_tester_urb);
			host_tester_urb = 0;
			submitted = 0;
		}
	}
}

static struct usb_device_id host_tester_ids[] =
{
	{USB_DEVICE(0x5511, 0x1000)},
	{}
};

static struct usb_driver host_tester_driver =
{
	name:	"host_tester_driver",
	probe:	host_tester_Probe,
	disconnect:	host_tester_Disconnect,
	fops:	&host_tester_fops,
	minor:	HOST_TESTER_MINOR,
	id_table:	host_tester_ids
};

static int __init host_tester_init(void)
{
	if (usb_register(&host_tester_driver))
	{
		return -1;
	}

#if 0 // TODO: fix devfs
	if (devfs_register_chrdev(HOST_TESTER_MAJOR,
		"host_tester", &host_tester_fops))
	{
		err("init: unable to get major %d", HOST_TESTER_MAJOR);
		usb_deregister(&host_tester_driver);
		return -EBUSY;
	}
#endif

	info("tester driver registered");
	return 0;
}

static void __exit host_tester_exit(void)
{
	usb_deregister(&host_tester_driver);
#if 0 // TODO: fix devfs
	devfs_unregister_chrdev(HOST_TESTER_MAJOR, "host_tester");
#endif
	info("tester driver removed");
}

module_init(host_tester_init);
module_exit(host_tester_exit);
