/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg.h
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 chip object function definitions.
 */
#ifndef OTG_H
#define OTG_H

#include "os.h"

#include "uhcTypes.h"
#include "otg_config.h"
#include "otg_t.h"
#include "otg_mode_t.h"
#include "otg_dma_t.h"
#include "otg_etd_t.h"
#include "otg_intr_t.h"
#include "otg_io_config_t.h"
#include "otg_mem_t.h"
#include "otg_hc_t.h"

#ifdef SIM_OTG
#include "otg_chip.h"
#endif

#ifdef DEBUG_OTG_OTHERS
#define otg_LockChip(otg) { info("otg_LockChip"); spin_lock(&otg->chip_lock); }
#define otg_UnlockChip(otg) { info("otg_UnLockChip"); spin_unlock(&otg->chip_lock); }
#define otg_LockChipIntr(otg, flags) { spin_lock_irqsave(&otg->chip_lock, flags); info("otg_LockChipIntr %lx", flags); }
#define otg_UnlockChipIntr(otg, flags) { info("otg_UnlockChipIntr %lx", flags); spin_unlock_irqrestore(&otg->chip_lock, flags); }
#else
#define otg_LockChip(otg) spin_lock(&otg->chip_lock)
#define otg_UnlockChip(otg) spin_unlock(&otg->chip_lock)
#define otg_LockChipIntr(otg, flags) spin_lock_irqsave(&otg->chip_lock, flags)
#define otg_UnlockChipIntr(otg, flags) spin_unlock_irqrestore(&otg->chip_lock, flags)
#endif

otg_t *otg_New(otg_hc_t* otg_hc);
void otg_Delete(otg_t *otg);

int otg_Init(otg_t *otg, unsigned long address);
void otg_Exit(otg_t *otg);
void otg_Reset(otg_t *otg);
irqreturn_t otg_DoIntr(int intr_num, void* priv);

otg_t *otg_FindChipIntr(void); /* find the chip which generated the intrrupt */

#ifdef DEBUG_ETD_ENABLE
void *otg_GetMemMap(void);
#endif

#ifdef DEBUG_OTG_CHIP
char *otg_GetRegName(int reg);
void otg_DumpReg(otg_t *otg);
#endif

#ifdef DO_PROFILING
void otg_TimeStart(otg_t *otg);
void otg_TimeStop(otg_t *otg);
#endif

#ifdef OTG_16_BIT_DATA_BUS
static inline void
otg_WriteReg16(void* addr, UINT16 val)
{
	writew(val, addr);
}
#endif

static inline void
otg_WriteReg(void* addr, UINT32 val)
{
	/* Caller of this function must disable interrupt */
#ifdef OTG_16_BIT_DATA_BUS
	writew(val, addr);
	writew(val >> 16, addr);
#else
#ifdef SIM_OTG
	void otg_chip_WriteReg(void*, UINT32);

	otg_chip_WriteReg(addr, val);
#else
#ifdef DEBUG_ETD_ENABLE
if ((addr - otg_GetMemMap())==(otg_HcRegBankStart_c + otg_hc_etd_EnableClear_c))
{
	info("ETD Clear: 0x%0x", val);
}
#endif
	writel(val, addr);
#endif /* SIM_OTG */

#endif /* #ifdef OTG_16_BIT_DATA_BUS */
}

#ifdef OTG_16_BIT_DATA_BUS
static inline UINT16
otg_ReadReg16(void* addr)
{
	return readw(addr);
}
#endif

static inline UINT32
otg_ReadReg(void* addr)
{
	UINT32 w;

#ifdef OTG_16_BIT_DATA_BUS
	/* Caller of this function must disable interrupt */
	w = readw(addr) & 0xFFFF;
	w |= readw(addr) << 16;
#else
#ifdef SIM_OTG
	UINT32 otg_chip_ReadReg(void *);

	w = otg_chip_ReadReg(addr);
#else
	w = readl(addr);
#endif /* SIM_OTG */

#endif /* #ifdef OTG_16_BIT_DATA_BUS */

	return w;
}

static inline void
otg_WriteChipReg(otg_t *otg, int reg_off, UINT32 val)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
#ifdef DEBUG_READ_PIO
info("WriteChipReg: 0x%0x:%s = 0x%0x", reg_off, otg_GetRegName(reg_off), val);
#else
if (otg_PIO_dataPort_c != reg_off)
{
info("WriteChipReg: 0x%0x:%s = 0x%0x", reg_off, otg_GetRegName(reg_off), val);
}
#endif
#endif
#endif
	otg_WriteReg(otg->mem_map + reg_off, val);
}

static inline void
otg_WriteHostReg(otg_t *otg, int reg_off, UINT32 val)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
info("WriteHostReg: 0x%0x:%s = 0x%0x", reg_off,
	otg_GetRegName(reg_off + otg_HcRegBankStart_c), val);
#endif
#endif

	otg_WriteReg(otg->mem_map + otg_HcRegBankStart_c + reg_off, val);
}

static inline void
otg_WriteFunctionReg(otg_t *otg, int reg_off, UINT32 val)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
info("WriteFunctionReg: 0x%0x:%s = 0x%0x", reg_off,
	otg_GetRegName(reg_off + otg_FcRegBankStart_c), val);
#endif
#endif
	otg_WriteReg(otg->mem_map + otg_FcRegBankStart_c + reg_off, val);
}

static inline UINT32
otg_ReadChipReg(otg_t *otg, int reg_off)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
UINT32 w;
w = otg_ReadReg(otg->mem_map + reg_off);
#ifdef DEBUG_READ_PIO
info("ReadChipReg: 0x%0x:%s = 0x%0x", reg_off, otg_GetRegName(reg_off), w);
#else
if (otg_PIO_dataPort_c != reg_off)
{
info("ReadChipReg: 0x%0x:%s = 0x%0x", reg_off, otg_GetRegName(reg_off), w);
}
#endif
return w;
#endif
#endif
	return otg_ReadReg(otg->mem_map + reg_off);
}

static inline UINT32
otg_ReadHostReg(otg_t *otg, int reg_off)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
UINT32 w;
w = otg_ReadReg(otg->mem_map + otg_HcRegBankStart_c + reg_off);
if (reg_off != otg_hc_rh_port_Status0_c &&
	reg_off != otg_hc_rh_port_Status1_c &&
	reg_off != otg_hc_rh_port_Status2_c)
{
	info("ReadHostReg: 0x%0x:%s = 0x%0x", reg_off,
		otg_GetRegName(reg_off + otg_HcRegBankStart_c), w);
}
return w;
#endif
#endif
	return otg_ReadReg(otg->mem_map + otg_HcRegBankStart_c + reg_off);
}

static inline UINT32
otg_ReadFunctionReg(otg_t *otg, int reg_off)
{
#ifndef SIM_OTG
#ifdef DEBUG_OTG_CHIP
UINT32 w;
w = otg_ReadReg(otg->mem_map + otg_FcRegBankStart_c + reg_off);
info("ReadFunctionReg: 0x%0x:%s = 0x%0x", reg_off,
	otg_GetRegName(reg_off + otg_FcRegBankStart_c), w);
return w;
#endif
#endif
	return otg_ReadReg(otg->mem_map + otg_FcRegBankStart_c + reg_off);
}

#ifndef OTG_FUNCTION_ONLY
static inline otg_hc_t *
otg_GetHost(otg_t *otg)
{
	return otg->hc;
}
#endif

#ifndef OTG_HOST_ONLY
static inline otg_fc_t *
otg_GetFunction(otg_t *otg)
{
	return &(otg->fc);
}
#endif

static inline otg_intr_t *
otg_GetIntr(otg_t *otg)
{
	return &otg->intr;
}

#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
static inline otg_device_t *
otg_GetDevice(otg_t *otg)
{
	return &otg->device;
}
#endif
#endif

#include "otg_mode.h"
#include "otg_dma.h"
#include "otg_etd.h"
#include "otg_intr.h"
#include "otg_io_config.h"
#include "otg_mem.h"
#include "otg_hc.h"
#include "otg_fc.h"
#include "otg_device.h"

static inline int
otg_GetIntrNo(otg_t *otg)
{
	return otg_intr_GetIntrNo(&otg->intr);
}

static inline void
otg_SetIntrNo(otg_t *otg, int no)
{
	otg_intr_SetIntrNo(&otg->intr, no);
}

static inline void
otg_SetHardwareHNP(otg_t *otg)
{
	otg_mode_SetHardwareHNP(&otg->mode);
}

static inline void
otg_SetSoftwareHNP(otg_t *otg)
{
	otg_mode_SetSoftwareHNP(&otg->mode);
}

static inline void
otg_ClearBHnpEnable(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_ClearBHnpEnable(&otg->device);
#endif
}

static inline void
otg_SetBHnpEnable(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_SetBHnpEnable(&otg->device);
#endif
}

static inline void
otg_SetBHnpSupport(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_SetBHnpSupport(&otg->device);
#endif
}

static inline void
otg_ClearBHnpSupport(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_ClearBHnpSupport(&otg->device);
#endif 
}


static inline void
otg_TransferHostToB(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_ReleaseBus(&otg->device);
	otg_device_SetAvBHnpEnable(&otg->device);
	otg_hc_SetOtgPortSuspend(otg->hc);
#ifdef OTG_HNP_DEBUG
info("TransferHostToB: HNPStatus=0x%0x", otg->device.status);
#endif
#endif
}

static inline void
otg_BHnpEnabled(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
#ifdef OTG_HNP_DEBUG
info("BHnpEnabled: ");
#endif
	otg_device_BHnpEnabled(&otg->device);
#endif
}

static inline void
otg_EnableSofIntr(otg_t *otg)
{
	(otg->sofCount)++;
	if (otg->sofCount == 1)
	{
		otg_hc_EnableSofIntr(otg->hc);
	}
}

static inline void
otg_DisableSofIntr(otg_t *otg)
{
	(otg->sofCount)--;
	if (!otg->sofCount)
	{
		otg_hc_DisableSofIntr(otg->hc);
	}
}

static inline void
otg_DoSofIntr(otg_t *otg)
{
#ifndef OTG_HOST_ONLY
	otg_device_DoSofIntr(&otg->device);
#endif 
#ifdef OTG_ISO_SHARE_MEMORY
	otg_hc_DoSofIntr(otg->hc);
#endif
}



#endif /* #ifndef OTG_H */
