/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_chip.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 simulator.
 */
#include <linux/usb.h>
#include "otg.h"
#include "otg_chip.h"
#include "otg_sim_device.h"

#define max_frame_bytes 1024

static int process_id = -1;
static otg_chip_t *chip = 0;
static DECLARE_MUTEX_LOCKED(kotg_chip_exited);
static UINT8 frame[max_frame_bytes];
static int frame_write;
static int frame_read;

static char* regNames[] =
{
	"HardwareMode",			/* 0 */
	"MNPControlStatus",		/* 4 */
	"OTGTimer1",			/* 8 */
	"OTGTimer2",			/* C */
	"ClockControl",			/* 10 */
	"DMA0Config",			/* 14 */
	"DMA0ETD",				/* 18 */
	"DMA1Config",			/* 1C */
	"DMA1ETD",				/* 20 */
	"DMA0DataPort",			/* 24 */
	"DMA1DataPort",			/* 28 */
	"PIODataPort",			/* 2C */
	"FuncRamStart",			/* 30 */
	"ChipIntrStatus",		/* 34 */
	"ChipIntrEnable",		/* 38 */
	"ChipIntrClear",		/* 3C */
	"ChipIntrSet",			/* 40 */
	"IOReady",				/* 44 */
	"IOConfig1",			/* 48 */
	"IOConfig2",			/* 4C */
	"ShortFrame",			/* 50 */
	"NoRegister54",			/* 54 */
	"NoRegister58",			/* 58 */
	"NoRegister5C",			/* 5C */
	"NoRegister60",			/* 60 */
	"NoRegister64",			/* 64 */
	"NoRegister68",			/* 68 */
	"NoRegister6C",			/* 6C */
	"NoRegister70",			/* 70 */
	"NoRegister74",			/* 74 */
	"NoRegister78",			/* 78 */
	"NoRegister7C",			/* 7C */
	"HcRevision",			/* 80 */
	"HcControl",			/* 84 */
	"HcCommandStatus",		/* 88 */
	"HcIntrStatus",			/* 8C */
	"HcIntrEnbale",			/* 90 */
	"HcIntrDisable",		/* 94 */
	"HcETDIntrStatus",		/* 98 */
	"HcETDIntrEnable",		/* 9C */
	"HcETDIntrDisable",		/* A0 */
	"HcStartAddr",			/* A4 */
	"HcXYBufferStatus",		/* A8 */
	"HcXYBufferSet",		/* AC */
	"HcXYBufferClear",		/* B0 */
	"HcNoRegister34",		/* B4 */
	"HcETDEnableSet",		/* B8 */
	"HcETDEnableClear",		/* BC */
	"HcETDEnableStatus",	/* C0 */
	"HcNoRegister44",		/* C4 */
	"HcNoRegister48",		/* C8 */
	"HcDirectRouting",		/* CC */
	"HcImmediateIntr",		/* D0 */
	"HcDoneQueue",			/* D4 */
	"HcFmInterval",			/* D8 */
	"HcFmRemaining",		/* DC */
	"HcFmNumber",			/* E0 */
	"HcLSThreshold",		/* E4 */
	"HcRhDescriptorA",		/* E8 */
	"HcRhDescriptorB",		/* EC */
	"HcRhStatus",			/* F0 */
	"HcRhPortStatus0",		/* F4 */
	"HcRhPortStatus1",		/* F8 */
	"HcRhPortStatus2",		/* FC */
	"FcDeviceAddress",		/* 100 */
	"FcCommand",			/* 104 */
	"FcEP0Control",			/* 108 */
	"FcEP0PacketSize",		/* 10C */
	"FcEP1Config",			/* 110 */
	"FcEP1BufferStart",		/* 114 */
	"FcEP1PacketControl",	/* 118 */
	"FcNoRegister1C",		/* 11C */
	"FcEP2Config",			/* 120 */
	"FcEP2BufferStart",		/* 124 */
	"FcEP2PacketControl",	/* 128 */
	"FcNoRegister2C",		/* 12C */
	"FcEP3Config",			/* 130 */
	"FcEP3BufferStart",		/* 134 */
	"FcEP3PacketControl",	/* 138 */
	"FcNoRegister3C",		/* 13C */
	"FcEP4Config",			/* 140 */
	"FcEP4BufferStart",		/* 144 */
	"FcEP4PacketControl",	/* 148 */
	"FcNoRegister4C",		/* 14C */
	"FcEP5Config",			/* 150 */
	"FcEP5BufferStart",		/* 154 */
	"FcEP5PacketControl",	/* 158 */
	"FcNoRegister5C",		/* 15C */
	"FcEP6Config",			/* 160 */
	"FcEP6BufferStart",		/* 164 */
	"FcEP6PacketControl",	/* 168 */
	"FcFcNoRegister6C",		/* 16C */
	"FcEP7Config",			/* 170 */
	"FcEP7BufferStart",		/* 174 */
	"FcEP7PacketControl",	/* 178 */
	"FcNoRegister7C",		/* 17C */
	"FcNoRegister80",		/* 180 */
	"FcNoRegister84",		/* 184 */
	"FcNoRegister88",		/* 188 */
	"FcNoRegister8C",		/* 18C */
	"FcNoRegister90",		/* 190 */
	"FcNoRegister94",		/* 194 */
	"FcNoRegister98",		/* 198 */
	"FcNoRegister9C",		/* 19C */
	"FcBufferFilled",		/* 1A0 */
	"FcBufferFilledSet",	/* 1A4 */
	"FcBufferFilledClear",	/* 1A8 */
	"FcNoRegisterAC",		/* 1AC */
	"FcEPIntrStatus",		/* 1B0 */
	"FcEPIntrEnable",		/* 1B4 */
	"FcEPIntrClear",		/* 1B8 */
	"FcEPIntrSet",			/* 1BC */
	"FcSysIntrStatus",		/* 1C0 */
	"FcSysIntrEnable",		/* 1C4 */
	"FcSysIntrClear",		/* 1C8 */
	"FcSysIntrSet",			/* 1CC */
	"FcStartAddr",			/* 1D0 */
	"FcNoRegisterD4",		/* 1D4 */
	"FcNoRegisterD8",		/* 1D8 */
	"FcNoRegisterDC",		/* 1DC */
	"FcChipId",				/* 1E0 */
	"FcFrameNo",			/* 1E4 */
	"FcUnlockRegister",		/* 1E8 */
	"FcNoRegisterEC",		/* 1EC */
};

char *
otg_chip_GetRegName(void *addr)
{
	return regNames[((unsigned long)addr - (unsigned long)chip)>>2];
}

char *
otg_chip_GetRegNameByAddr(int addr)
{
	return regNames[addr>>2];
}

otg_chip_t *
otg_chip_GetChip(void)
{
	return chip;
}

static inline void
otg_chip_SetReg(int reg, UINT32 w)
{
	chip->reg[reg>>2] |= w;
}

UINT32
otg_chip_GetReg(int reg)
{
	return chip->reg[reg>>2];
}

static inline void
otg_chip_ClearReg(int reg, UINT32 w)
{
	chip->reg[reg>>2] &= ~w;
}

static inline UINT32
otg_chip_ReadHostReg(int reg)
{
	return chip->reg[(reg + otg_HcRegBankStart_c) >> 2];
}

static inline void
otg_chip_WriteHostReg(int reg, UINT32 val)
{
	chip->reg[(reg + otg_HcRegBankStart_c) >> 2] = val;
}

static inline UINT32
otg_chip_ReadFcReg(int reg)
{
	return chip->reg[(reg + otg_FcRegBankStart_c) >> 2];
}

void
otg_chip_WriteFcReg(int reg, UINT32 val)
{
	chip->reg[(reg + otg_FcRegBankStart_c) >> 2] = val;
}

static inline void
otg_chip_SetHostReg(int reg, UINT32 val)
{
	chip->reg[(reg + otg_HcRegBankStart_c) >> 2] |= val;
}

static inline void
otg_chip_ClearHostReg(int reg, UINT32 val)
{
	chip->reg[(reg + otg_HcRegBankStart_c) >> 2] &= ~val;
}

void
otg_chip_WriteMem(int addr, UINT8 val)
{
	UINT32 w;

	w = chip->mem[addr>>2];
	switch (addr & 0x03)
	{
		case 0:
			w &= ~ 0xFF;
			w |= val;
			break;
		case 1:
			w &= ~0xFF00;
			w |= val << 8;
			break;
		case 2:
			w &= ~0xFF0000;
			w |= val << 16;
			break;
		case 3:
			w &= ~0xFF000000;
			w |= val << 24;
#ifdef DEBUG_OTG_CHIP
			info("WriteMem: addr=0x%0x w=0x%0x", addr & ~3, w);
#endif
			break;
	}

	chip->mem[addr>>2] = w;
}

UINT8
otg_chip_ReadMem(int addr)
{
	UINT32 w;
	UINT8 val = 0;

	w = chip->mem[addr>>2];
	switch (addr & 0x03)
	{
		case 0:
			val = w & 0xFF;
			break;
		case 1:
			val = (w >> 8) & 0xFF;
			break;
		case 2:
			val = (w >> 16) & 0xFF;
			break;
		case 3:
			val = (w >> 24) & 0xFF;
			break;
	}

	return val;
}

static inline void
otg_chip_WriteEtd(int etd, int word, UINT32 val)
{
	chip->etd[etd * 8 + word] = val;
}

static inline UINT32
otg_chip_ReadEtd(int etd, int word)
{
	return chip->etd[etd * 8 + word];
}

static int
otg_chip_HasData(void)
{
	return frame_read < frame_write;	
}

static int
otg_chip_HasMoreData(int bytes)
{
	return (frame_read + bytes - 1) < frame_write;	
}

static void
otg_chip_WriteFrame(UINT8 data)
{
	if (frame_write >= max_frame_bytes)
	{
		return;
	}

	frame[frame_write++] = data;
}

static UINT8
otg_chip_ReadFrame(void)
{
	if (frame_read >= max_frame_bytes)
	{
		return 0;
	}

	return frame[frame_read++];
}

static int
otg_chip_IsInEtd(int index)
{
	UINT32 w;

	w = chip->etd[index << 3] & otg_etd_Direction_c;
	if (w == otg_etd_Direction_IN_c)
	{
		return 1;
	}
	else if (w  == otg_etd_Direction_OUT_c)
	{
		return 0;
	}

	/* Check the DP field of Dword2 */
	w = chip->etd[(index << 3) + 2] & otg_td_ctrl_Pid_c;
	if (w == otg_td_ctrl_Pid_In_c)
	{
		return 1;
	}
	else
	{
		return 0;
	}
}

static void
otg_chip_StartTimer(int fun, int arg, int interval)
{
	static void otg_chip_Timeout(unsigned long data);
	struct timer_list *timer;

	timer = &chip->timer;
	if (chip->timer_fun)
	{
		/* There is timer pending. Let's trigger it */
		otg_chip_Timeout(0);
		del_timer(timer);
	}

	chip->timer_fun = fun;
	chip->timer_arg = arg;

	init_timer(timer);
	timer->function = otg_chip_Timeout;
	timer->data = 0;
	timer->expires = jiffies + interval;
	add_timer(timer);
}

static void
otg_chip_Timeout(unsigned long data)
{
	UINT32 w;

	switch (chip->timer_fun)
	{
		case otg_chip_PortResume:
			w = chip->reg[chip->timer_arg >> 2];
			w &= ~otg_hc_rh_port_Status_PSS_c;
			w |= otg_hc_rh_port_Status_PSSC_c;
			w |= otg_hc_rh_port_Status_PES_c;
			chip->reg[chip->timer_arg >> 2] = w;
			break;
		case otg_chip_PortReset:
			w = chip->reg[chip->timer_arg >> 2];
			w &= ~otg_hc_rh_port_Status_PRS_c;
			w |= otg_hc_rh_port_Status_PRSC_c;
			w |= otg_hc_rh_port_Status_PES_c;
			w &= ~otg_hc_rh_port_Status_PSS_c;
			chip->reg[chip->timer_arg >>2] = w;
			break;
	}

	chip->timer_fun = 0;
}

void
otg_chip_WriteStartAddr(UINT32 val)
{
	int max_addr;

	chip->mem_addr = val;

	if (val & 0x8000)
	{
		max_addr = otg_mem_Size_c;
	}
	else
	{
		max_addr = otg_chip_EtdBankSize * 4;
	}

	if ((val & 0x3FFF) >= max_addr)
	{
		err("WriteStartAddr: Write to illegal address 0x%x", val);
	}

	otg_chip_ClearReg(otg_IOReady_c, otg_IOReady_NotReady);
}

void
otg_chip_WritePIO(UINT32 val)
{
	int addr;
	int max_addr;
	UINT32 *pdata;

	otg_chip_SetReg(otg_IOReady_c, otg_IOReady_NotReady);

	if (!(chip->mem_addr & 0x4000))
	{
		err("WritePIO: Write while address is setup for read");
		return;
	}

	addr = chip->mem_addr & 0x3FFF;
	if (chip->mem_addr & 0x8000)
	{
		max_addr = otg_mem_Size_c;
		pdata = chip->mem;
	}
	else
	{
		max_addr = otg_chip_EtdBankSize * 4;
		pdata = chip->etd;
	}

	if (addr >= max_addr)
	{
		err("WritePIO: Write to illegal address 0x%x", addr);
		return;
	}

	pdata[addr >> 2] = val;
	chip->mem_addr += 4;
}

static void
otg_chip_ClearSetPower(int reg, int mode)
{
	UINT32 descA;
	UINT32 descB;
	UINT32 status;
	int change;

	descA = otg_chip_ReadHostReg(otg_hc_rh_DescriptorA_c);
	descB = otg_chip_ReadHostReg(otg_hc_rh_DescriptorB_c);
	status = otg_chip_ReadHostReg(reg);
	change = 0;
	if (descA & otg_hc_rh_Da_PSM_c)
	{
		/* per-port power mode */
		UINT32 w;

		w = 0x02 << otg_hc_rh_Db_PPCMLoc_c;
		if (reg == otg_hc_rh_port_Status2_c)
		{
			w <<= 1;
		}

		if (!(descB & w))
		{
			change = 1;
		}
	}
	else
	{
		/* global power mode */
		change = 1;
	}

	if (change)
	{
		if (mode)
		{
			/* Set Power */
			status |= otg_hc_rh_port_Status_PPS_c;
		}
		else
		{
			/* Clear Power */
			status &= ~otg_hc_rh_port_Status_PPS_c;
			status &= ~otg_hc_rh_port_Status_CCS_c;
			status &= ~otg_hc_rh_port_Status_PES_c;
			status &= ~otg_hc_rh_port_Status_PSS_c;
			status &= ~otg_hc_rh_port_Status_PRS_c;
		}
	}

	otg_chip_WriteHostReg(reg, status);
}

void
otg_chip_WriteRhStatus(UINT32 val)
{
	UINT32 status;

	status = otg_chip_ReadHostReg(otg_hc_rh_Status_c);
	if (val & otg_hc_rh_Status_LPS_c)
	{
		/* ClearGlobalPower */
		otg_chip_ClearSetPower(otg_hc_rh_port_Status1_c, 0);
		otg_chip_ClearSetPower(otg_hc_rh_port_Status2_c, 0);
		status &= ~otg_hc_rh_Status_LPS_c;
	}

	if (val & otg_hc_rh_Status_DRWE_c)
	{
		/* Set Device RemoveWakeupEnbale*/
		status |= otg_hc_rh_Status_DRWE_c;
	}

	if (val & otg_hc_rh_Status_LPSC_c)
	{
		otg_chip_ClearSetPower(otg_hc_rh_port_Status1_c, 1);
		otg_chip_ClearSetPower(otg_hc_rh_port_Status2_c, 1);
		status |= otg_hc_rh_Status_LPS_c;
	}

	if (val & otg_hc_rh_Status_OCIC_c)
	{
		status &= ~otg_hc_rh_Status_OCIC_c;
	}

	if (val & otg_hc_rh_Status_CRWE_c)
	{
		status &= ~otg_hc_rh_Status_DRWE_c;
	}

	otg_chip_WriteHostReg(otg_hc_rh_Status_c, status);
}

void
otg_chip_WriteRhPortStatus(int reg, UINT32 val)
{
	UINT32 status;

	status = otg_chip_ReadHostReg(reg);
	if (val & otg_hc_rh_port_Status_CCS_c)
	{
		status &= ~otg_hc_rh_port_Status_PES_c;
	}

	if (val & otg_hc_rh_port_Status_PES_c)
	{
		if (status & otg_hc_rh_port_Status_CCS_c)
		{
			status |= otg_hc_rh_port_Status_PES_c;
		}
	}

	if (val & otg_hc_rh_port_Status_PSS_c)
	{
		if (status & otg_hc_rh_port_Status_CCS_c)
		{
			status |= otg_hc_rh_port_Status_PSS_c;
		}
	}

	if (val & otg_hc_rh_port_Status_POCI_c)
	{
		otg_chip_StartTimer
		(
			otg_chip_PortResume,
			reg + otg_HcRegBankStart_c,
			3 * HZ / 1000
		);
	}

	if (val & otg_hc_rh_port_Status_PRS_c)
	{
		if (status & otg_hc_rh_port_Status_CCS_c)
		{
			status |= otg_hc_rh_port_Status_PRS_c;
			otg_chip_StartTimer
			(
				otg_chip_PortReset,
				reg + otg_HcRegBankStart_c,
				10 * HZ / 1000
			);
		}
		else
		{
			status |= otg_hc_rh_port_Status_CSC_c;
		}
	}

	if (val & otg_hc_rh_port_Status_PPS_c)
	{
		otg_chip_ClearSetPower(reg, 1);
	}

	if (val & otg_hc_rh_port_Status_LSDA_c)
	{
		otg_chip_ClearSetPower(reg, 0);
	}

	if (val & otg_hc_rh_port_Status_CSC_c)
	{
		status &= ~otg_hc_rh_port_Status_CSC_c;
	}

	if (val & otg_hc_rh_port_Status_PESC_c)
	{
		status &= ~otg_hc_rh_port_Status_PESC_c;
	}

	if (val & otg_hc_rh_port_Status_PSSC_c)
	{
		status &= ~otg_hc_rh_port_Status_PSSC_c;
	}

	if (val & otg_hc_rh_port_Status_PORTOCIC_c)
	{
		status &= ~otg_hc_rh_port_Status_PORTOCIC_c;
	}

	if (val & otg_hc_rh_port_Status_PRSC_c)
	{
		status &= ~otg_hc_rh_port_Status_PRSC_c;
	}

	otg_chip_WriteHostReg(reg, status);
}

static void
otg_chip_EnableEtd(UINT32 val)
{
	UINT32 w;
	int i;

	for(i = 0, w = 1; i < 16; i++, w <<= 1)
	{
		if (val & w)
		{
			chip->etdPacketDone[i] = 0;
		}
	}

	w = otg_chip_ReadHostReg(otg_hc_etd_EnableStatus_c);
	w |= val;
	otg_chip_WriteHostReg(otg_hc_etd_EnableStatus_c, w);
}

static void
otg_chip_HcReset(void)
{
	/* Reset host controller */
	chip->reg[(otg_hc_Revision_c + otg_HcRegBankStart_c)>>2] = 0x86;
	chip->reg[(otg_hc_Control_c + otg_HcRegBankStart_c)>>2] = otg_hc_USBSuspend_c;
	chip->reg[(otg_hc_CmdStatus_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_intr_Status_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_intr_Enable_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_intr_Disable_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_IntrStatus_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_IntrEnable_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_IntrDisable_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_AccessStartAddr_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_FilledStatus_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_FilledSet_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_FilledClear_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_EnableSet_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_EnableClear_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_EnableStatus_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_DirectRouting_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_ImmediateInt_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_etd_DoneQueue_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_FmInterval_c + otg_HcRegBankStart_c)>>2] = 0x1EDF;
	chip->reg[(otg_hc_FmRemaining_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_FmNumber_c + otg_HcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_hc_LSThreshold_c + otg_HcRegBankStart_c)>>2] = 0x0628;
}

static void
otg_chip_WriteFilled(int set_fill, int val)
{
	int index;
	UINT32 l;
	UINT32 h;
	UINT32 status;

	l = 1;
	h = l << 16;
	for (index = 0; index < 16; index++)
	{
		if (val & (l | h))
		{
			if (set_fill)
			{
				if (!otg_chip_IsInEtd(index))
				{
					status = otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c);
					if (val & l)
					{
						status |= l;
					}
					else
					{
						status |= h;
					}
					otg_chip_WriteHostReg(otg_hc_etd_FilledStatus_c, status);
				}
			}
			else
			{
				if (otg_chip_IsInEtd(index))
				{
					status = otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c);
					if (val & l)
					{
						status &= ~l;
					}
					else
					{
						status &= ~h;
					}
					otg_chip_WriteHostReg(otg_hc_etd_FilledStatus_c, status);
				}
			}
		}

		l <<= 1;
		h <<= 1;
	}
}

void
otg_chip_SetEpFillBit(int ep, UINT32 b, int set)
{
	UINT32 w;

	w = otg_chip_ReadFcReg(otg_ep_Config(ep));
	if (set)
	{
		if (!(w & otg_ep_ConfigDirection))
		{
			/* OUT endpoint */
			return;
		}

		w = otg_chip_ReadFcReg(otg_ep_FillStatus);
		w |= b;
		otg_chip_WriteFcReg(otg_ep_FillStatus, w);
	}
	else
	{
		if (w & otg_ep_ConfigDirection)
		{
			/* IN endpoint */
			return;
		}

		w = otg_chip_ReadFcReg(otg_ep_FillStatus);
		w &= ~b;
		otg_chip_WriteFcReg(otg_ep_FillStatus, w);
	}
}

void
otg_chip_SetEpFill(UINT32 val, int set)
{
	int i;
	UINT32 b;

	b = 1;
	for(i = 0; i < 7; i++)
	{
		b <<= 1;
		if (val & b)
		{
			otg_chip_SetEpFillBit(i, b, set);
		}
	}

	b = 0x10000;
	for(i = 0; i < 7; i++)
	{
		b <<= 1;
		if (val & b)
		{
			otg_chip_SetEpFillBit(i, b, set);
		}
	}
}

void
otg_chip_WriteEp0Ctrl(int val)
{
	UINT32 w;

	w = otg_chip_ReadFcReg(otg_ep0_Ctrl);
	if (val & otg_ep0_CtrlOut)
	{
		w &= ~otg_ep0_CtrlOut;
		val &= ~otg_ep0_CtrlOut;
	}

	if (val & otg_ep0_CtrlSetup)
	{
		w &= ~otg_ep0_CtrlSetup;
		val &= ~otg_ep0_CtrlSetup;
	}

	w |= val;
	otg_chip_WriteFcReg(otg_ep0_Ctrl, w);
}

void
otg_chip_FunctionReset(void)
{
	chip->reg[(otg_ep0_DeviceAddr + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_Command + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep0_Ctrl + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep0_Size + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(0) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(0) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(0) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(1) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(1) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(1) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(2) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(2) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(2) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(3) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(3) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(3) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(4) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(4) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(4) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(5) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(5) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(5) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Config(6) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Buffer(6) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_Ctrl(6) + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_FillStatus + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_FillSet + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_FillClear + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_intr_Status + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_intr_Enable + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_intr_Clear + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_ep_intr_Set + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_intr_Status + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_intr_Enable + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_intr_Clear + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_intr_Set + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_StartAddr + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_ChipId + otg_FcRegBankStart_c)>>2] = 0x0082;
	chip->reg[(otg_fc_FrameNo + otg_FcRegBankStart_c)>>2] = 0;
	chip->reg[(otg_fc_UnlockReg + otg_FcRegBankStart_c)>>2] = 0xEB82;
}

void
otg_chip_WriteReg(void* addr, UINT32 val)
{
	int reg;
	UINT32 w;

	reg = (unsigned long)addr - (unsigned long)chip;
#ifdef DEBUG_OTG_CHIP
info("WriteReg: Frame:0x%0x 0x%0x:%s with 0x%x", 
	otg_chip_ReadHostReg(otg_hc_FmNumber_c), reg, regNames[reg>>2], val);
#endif

	switch (reg)
	{
		case otg_device_HNPStatus_c:
		case otg_device_HNPTimer1_c:
		case otg_device_HNPTimer2_c:
		case otg_ClockCtl_c:
		case otg_dma_Config0_c:
		case otg_dma_Etd0_c:
		case otg_dma_Config1_c:
		case otg_dma_Etd1_c:
		case otg_dma_DataPort0_c:
		case otg_dma_DataPort1_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_PIO_dataPort_c:
			chip->reg[reg>>2] = val;
			otg_chip_WritePIO(val);
			break;
		case otg_FncRamStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_intr_Status_c:
			break;
		case otg_intr_Enable_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_intr_Clear_c:
			otg_chip_ClearReg(otg_intr_Status_c, val);
			break;
		case otg_intr_Set_c	:
			otg_chip_SetReg(otg_intr_Status_c, val);
			break;
		case otg_IOReady_c:
		case otg_io_config_Reg1_c:
		case otg_io_config_Reg2_c:
		case otg_ShortFrame_c:
		case otg_hc_Revision_c + otg_HcRegBankStart_c:
		case otg_hc_Control_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_hc_CmdStatus_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			if (val & otg_hc_CmdStatus_SwReset_c)
			{
				otg_chip_HcReset();
			}
			break;
		case otg_hc_intr_Status_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] &= ~val;
			if (val & otg_hc_intr_RDQ_c)
			{
				/* Clear out done queue */
				chip->reg[(otg_hc_etd_DoneQueue_c + otg_HcRegBankStart_c)>>2]
					= 0;
			}
			break;
		case otg_hc_intr_Enable_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] |= val;
			break;
		case otg_hc_intr_Disable_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] |= val;
			w = otg_chip_ReadHostReg(otg_hc_intr_Enable_c);
			w &= ~val;
			otg_chip_WriteHostReg(otg_hc_intr_Enable_c, w);
			break;
		case otg_hc_etd_IntrStatus_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] &= ~val;
			break;
		case otg_hc_etd_IntrEnable_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] |= val;
			break;
		case otg_hc_etd_IntrDisable_c + otg_HcRegBankStart_c:
			w = otg_chip_ReadHostReg(otg_hc_etd_IntrEnable_c);
			w &= ~val;
			otg_chip_WriteHostReg(otg_hc_etd_IntrEnable_c, w);
			break;
		case otg_hc_AccessStartAddr_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			otg_chip_WriteStartAddr(val);
			break;
		case otg_hc_etd_FilledStatus_c + otg_HcRegBankStart_c:
			break;
		case otg_hc_etd_FilledSet_c + otg_HcRegBankStart_c:
			otg_chip_WriteFilled(1, val);
			break;
		case otg_hc_etd_FilledClear_c + otg_HcRegBankStart_c:
			otg_chip_WriteFilled(0, val);
			break;
		case otg_hc_etd_EnableSet_c + otg_HcRegBankStart_c:
			otg_chip_EnableEtd(val);
			break;
		case otg_hc_etd_EnableClear_c + otg_HcRegBankStart_c:
			w = otg_chip_ReadHostReg(otg_hc_etd_EnableStatus_c);
			w &= ~val;
			otg_chip_WriteHostReg(otg_hc_etd_EnableStatus_c, w);
			break;
		case otg_hc_etd_EnableStatus_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_hc_DirectRouting_c + otg_HcRegBankStart_c:
		case otg_hc_etd_ImmediateInt_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_hc_etd_DoneQueue_c + otg_HcRegBankStart_c:
			break;
		case otg_hc_FmInterval_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_hc_FmRemaining_c + otg_HcRegBankStart_c:
		case otg_hc_FmNumber_c + otg_HcRegBankStart_c:
			break;
		case otg_hc_LSThreshold_c + otg_HcRegBankStart_c:
		case otg_hc_rh_DescriptorA_c + otg_HcRegBankStart_c:
		case otg_hc_rh_DescriptorB_c + otg_HcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_hc_rh_Status_c + otg_HcRegBankStart_c:
			otg_chip_WriteRhStatus(val);
			break;
		case otg_hc_rh_port_Status0_c + otg_HcRegBankStart_c:
			otg_chip_WriteRhPortStatus(otg_hc_rh_port_Status0_c, val);
			break;
		case otg_hc_rh_port_Status1_c + otg_HcRegBankStart_c:
			otg_chip_WriteRhPortStatus(otg_hc_rh_port_Status1_c, val);
			break;
		case otg_hc_rh_port_Status2_c + otg_HcRegBankStart_c:
			otg_chip_WriteRhPortStatus(otg_hc_rh_port_Status2_c, val);
			break;
		case otg_ep0_DeviceAddr + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_fc_Command + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			if (val & otg_fc_CommandReset)
			{
				otg_chip_FunctionReset();
			}

			break;
		case otg_ep0_Ctrl + otg_FcRegBankStart_c:
			otg_chip_WriteEp0Ctrl(val);
			break;
		case otg_ep0_Size + otg_FcRegBankStart_c:
			w = chip->reg[reg>>2];
			w &= ~otg_ep0_SizeIn;
			w |= val & otg_ep0_SizeIn;
			chip->reg[reg>>2] = val;
			break;
		case otg_ep_Config(0) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(0) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(0) + otg_FcRegBankStart_c:
		case otg_ep_Config(1) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(1) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(1) + otg_FcRegBankStart_c:
		case otg_ep_Config(2) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(2) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(2) + otg_FcRegBankStart_c:
		case otg_ep_Config(3) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(3) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(3) + otg_FcRegBankStart_c:
		case otg_ep_Config(4) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(4) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(4) + otg_FcRegBankStart_c:
		case otg_ep_Config(5) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(5) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(5) + otg_FcRegBankStart_c:
		case otg_ep_Config(6) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(6) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(6) + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_ep_FillStatus + otg_FcRegBankStart_c:
			break;
		case otg_ep_FillSet + otg_FcRegBankStart_c:
			otg_chip_SetEpFill(val, 1);
			break;
		case otg_ep_FillClear + otg_FcRegBankStart_c:
			otg_chip_SetEpFill(val, 0);
			break;
		case otg_ep_intr_Status + otg_FcRegBankStart_c:
			break;
		case otg_ep_intr_Enable + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_ep_intr_Clear + otg_FcRegBankStart_c:
			w = otg_chip_ReadFcReg(otg_ep_intr_Status);
			w &= ~val;
			otg_chip_WriteFcReg(otg_ep_intr_Status, w);
			break;
		case otg_ep_intr_Set + otg_FcRegBankStart_c:
			w = otg_chip_ReadFcReg(otg_ep_intr_Status);
			w |= val;
			otg_chip_WriteFcReg(otg_ep_intr_Status, w);
			break;
		case otg_fc_intr_Status + otg_FcRegBankStart_c:
			break;
		case otg_fc_intr_Enable + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		case otg_fc_intr_Clear + otg_FcRegBankStart_c:
			w = otg_chip_ReadFcReg(otg_fc_intr_Status);
			w &= ~val;
			otg_chip_WriteFcReg(otg_fc_intr_Status, w);
			break;
		case otg_fc_intr_Set + otg_FcRegBankStart_c:
			w = otg_chip_ReadFcReg(otg_fc_intr_Status);
			w |= val;
			otg_chip_WriteFcReg(otg_fc_intr_Status, w);
			break;
		case otg_fc_StartAddr + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			otg_chip_WriteStartAddr(val | 0x8000);
			break;
		case otg_fc_ChipId + otg_FcRegBankStart_c:
		case otg_fc_FrameNo + otg_FcRegBankStart_c:
		case otg_fc_UnlockReg + otg_FcRegBankStart_c:
			chip->reg[reg>>2] = val;
			break;
		default:
			chip->reg[reg>>2] = val;
			break;
	}
}

UINT32
otg_chip_ReadPIO(void)
{
	int addr;
	int max_addr;
	UINT32 *pdata;
	UINT32 val;

	otg_chip_SetReg(otg_IOReady_c, 0x1);

	if (chip->mem_addr & 0x4000)
	{
		err("Read while address is setup for write");
		return 0;
	}

	addr = chip->mem_addr & 0x3FFF;
	if (chip->mem_addr & 0x8000)
	{
		max_addr = otg_mem_Size_c;
		pdata = chip->mem;
	}
	else
	{
		max_addr = otg_chip_EtdBankSize * 4;
		pdata = chip->etd;
	}

	if (addr >= max_addr)
	{
		err("Write to illegal address 0x%x", addr);
	}

	val = pdata[addr >> 2];
	chip->mem_addr += 4;
	return val;
}

UINT32
otg_chip_ReadReg(void* addr)
{
	int reg;
	UINT32 val;

	reg = (unsigned long)addr - (unsigned long)chip;
	val = chip->reg[reg>>2];
	switch (reg)
	{
		case otg_device_HNPStatus_c:
		case otg_device_HNPTimer1_c:
		case otg_device_HNPTimer2_c:
		case otg_ClockCtl_c:
		case otg_dma_Config0_c:
		case otg_dma_Etd0_c:
		case otg_dma_Config1_c:
		case otg_dma_Etd1_c:
		case otg_dma_DataPort0_c:
		case otg_dma_DataPort1_c:
			break;
		case otg_PIO_dataPort_c:
			val = otg_chip_ReadPIO();
			break;
		case otg_FncRamStart_c:
		case otg_intr_Status_c:
		case otg_intr_Enable_c:
		case otg_intr_Clear_c:
		case otg_intr_Set_c	:
		case otg_IOReady_c:
		case otg_io_config_Reg1_c:
		case otg_io_config_Reg2_c:
		case otg_ShortFrame_c:
		case otg_hc_Revision_c + otg_HcRegBankStart_c:
		case otg_hc_Control_c + otg_HcRegBankStart_c:
		case otg_hc_CmdStatus_c + otg_HcRegBankStart_c:
		case otg_hc_intr_Status_c + otg_HcRegBankStart_c:
		case otg_hc_intr_Enable_c + otg_HcRegBankStart_c:
			break;
		case otg_hc_intr_Disable_c + otg_HcRegBankStart_c:
			val = otg_chip_ReadHostReg(otg_hc_intr_Enable_c);
			break;
		case otg_hc_etd_IntrStatus_c + otg_HcRegBankStart_c:
		case otg_hc_etd_IntrEnable_c + otg_HcRegBankStart_c:
			break;
		case otg_hc_etd_IntrDisable_c + otg_HcRegBankStart_c:
			val = otg_chip_ReadHostReg(otg_hc_etd_IntrEnable_c);
			break;
		case otg_hc_AccessStartAddr_c + otg_HcRegBankStart_c:
			val = chip->mem_addr;
			break;
		case otg_hc_etd_FilledStatus_c + otg_HcRegBankStart_c:
		case otg_hc_etd_FilledSet_c + otg_HcRegBankStart_c:
		case otg_hc_etd_FilledClear_c + otg_HcRegBankStart_c:
		case otg_hc_etd_EnableSet_c + otg_HcRegBankStart_c:
		case otg_hc_etd_EnableClear_c + otg_HcRegBankStart_c:
		case otg_hc_etd_EnableStatus_c + otg_HcRegBankStart_c:
		case otg_hc_DirectRouting_c + otg_HcRegBankStart_c:
		case otg_hc_etd_ImmediateInt_c + otg_HcRegBankStart_c:
		case otg_hc_etd_DoneQueue_c + otg_HcRegBankStart_c:
		case otg_hc_FmInterval_c + otg_HcRegBankStart_c:
		case otg_hc_FmRemaining_c + otg_HcRegBankStart_c:
		case otg_hc_FmNumber_c + otg_HcRegBankStart_c:
		case otg_hc_LSThreshold_c + otg_HcRegBankStart_c:
		case otg_ep0_DeviceAddr + otg_FcRegBankStart_c:
		case otg_fc_Command + otg_FcRegBankStart_c:
		case otg_ep0_Ctrl + otg_FcRegBankStart_c:
		case otg_ep0_Size + otg_FcRegBankStart_c:
		case otg_ep_Config(0) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(0) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(0) + otg_FcRegBankStart_c:
		case otg_ep_Config(1) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(1) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(1) + otg_FcRegBankStart_c:
		case otg_ep_Config(2) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(2) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(2) + otg_FcRegBankStart_c:
		case otg_ep_Config(3) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(3) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(3) + otg_FcRegBankStart_c:
		case otg_ep_Config(4) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(4) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(4) + otg_FcRegBankStart_c:
		case otg_ep_Config(5) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(5) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(5) + otg_FcRegBankStart_c:
		case otg_ep_Config(6) + otg_FcRegBankStart_c:
		case otg_ep_Buffer(6) + otg_FcRegBankStart_c:
		case otg_ep_Ctrl(6) + otg_FcRegBankStart_c:
		case otg_ep_FillStatus + otg_FcRegBankStart_c:
		case otg_ep_FillSet + otg_FcRegBankStart_c:
		case otg_ep_FillClear + otg_FcRegBankStart_c:
		case otg_ep_intr_Status + otg_FcRegBankStart_c:
		case otg_ep_intr_Enable + otg_FcRegBankStart_c:
		case otg_ep_intr_Clear + otg_FcRegBankStart_c:
		case otg_ep_intr_Set + otg_FcRegBankStart_c:
		case otg_fc_intr_Status + otg_FcRegBankStart_c:
		case otg_fc_intr_Enable + otg_FcRegBankStart_c:
		case otg_fc_intr_Clear + otg_FcRegBankStart_c:
		case otg_fc_intr_Set + otg_FcRegBankStart_c:
		case otg_fc_StartAddr + otg_FcRegBankStart_c:
		case otg_fc_ChipId + otg_FcRegBankStart_c:
		case otg_fc_FrameNo + otg_FcRegBankStart_c:
		case otg_fc_UnlockReg + otg_FcRegBankStart_c:
			break;
		default:
			break;
	}

#ifdef DEBUG_OTG_CHIP
info("ReadReg: Frame:0x%0x 0x%0x:%s get 0x%x",
	otg_chip_ReadHostReg(otg_hc_FmNumber_c), reg, regNames[reg>>2], val);
#endif

	return val;
}

static void
otg_chip_WriteDataToggle(UINT32 *w)
{
	if ((*w & otg_td_ctrl_DataToggle_c)==otg_td_ctrl_DataToggle0_c)
	{
		otg_chip_WriteFrame(otg_chip_TokenDATA0);
		*w &= ~otg_td_ctrl_DataToggle_c;
		*w |= otg_td_ctrl_DataToggle1_c;
	}
	else
	{
		otg_chip_WriteFrame(otg_chip_TokenDATA1);
		*w &= ~otg_td_ctrl_DataToggle_c;
		*w |= otg_td_ctrl_DataToggle0_c;
	}
}

static void
otg_chip_DoOut(UINT32 w0, UINT32 w1, UINT32 *w2, UINT32 *w3, int index,
otg_sim_device_t *dev)
{
	int mem_addr;
	int packets;
	int count;
	int last_size;
	int max_length;
	int yBuffer;
	int i;
	UINT32 filled;

#ifdef DEBUG_OTG_CHIP
info("DoOut: index=%d dev=%p FilledStatus=0x%0x",
	index, dev, otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c));
#endif
	otg_chip_WriteDataToggle(w2);
	packets = (*w3 & otg_td_ctrl_PacketCount_c) >> otg_td_ctrl_PacketCountLoc_c;
	count = (*w3 & otg_td_ctrl_NumPacketsPerBuff_c) >>
			otg_td_ctrl_NumPacketsPerBuffLoc_c;
	last_size = (*w3 & otg_td_ctrl_LastPacketSize_c) >>
			otg_td_ctrl_LastPacketSizeLoc_c;

	count++; /* zero is one packet */
	if (count > packets)
	{
		count = packets;
	}

	if (!count)
	{
#ifdef DEBUG_OTG_CHIP
info("DoOut: packet count is zero");
#endif
		return;
	}

	mem_addr = w1 & otg_etd_BufferField_c;
	yBuffer = 0;
	filled = 1 << index;
	if ((int)(chip->etdPacketDone[index] / count) & 1)
	{
		yBuffer = 1;
		filled <<= 16;
		mem_addr = (w1 >> otg_etd_YBufferLoc_c) & otg_etd_BufferField_c;
	}

	max_length = (w0 & otg_etd_MaxPacketSize_c) >> otg_etd_MaxPacketSizeLoc_c;

#ifdef DEBUG_OTG_CHIP
info("DoOut: count=%d packets=%d mem_addr=0x%0x last_size=0x%0x filled=0x%0x",
	count, packets, mem_addr, last_size, filled);
#endif
	(*w3)--;
	if (*w3 & otg_td_ctrl_PacketCount_c) /* not last packet */
	{
		last_size = max_length;
	}

	otg_chip_WriteFrame(last_size & 0xFF);
	otg_chip_WriteFrame((last_size >> 8) & 0xFF);
	for (i = 0; i < last_size; i++)
	{
		otg_chip_WriteFrame(otg_chip_ReadMem(mem_addr));
		mem_addr++;
	}

	otg_sim_device_DoFrame(dev);

	*w2 &= ~otg_td_ctrl_CompletionCode_c;
	if (otg_chip_HasData())
	{
		UINT8 b;

		b = otg_chip_ReadFrame();
		switch (b)
		{
			case otg_chip_TokenACK:
				*w2 |= otg_etd_CC_NoError_c;
#ifdef DEBUG_OTG_CHIP
info("ACK received");
#endif
				break;
			case otg_chip_TokenNAK:
				/* Need to find a good error code */
#ifdef DEBUG_OTG_CHIP
info("NAK received");
#endif
				*w2 |= otg_etd_CC_NoError_c;
				break;
			case otg_chip_TokenSTALL:
#ifdef DEBUG_OTG_CHIP
info("Stall received");
#endif
				*w2 |= otg_etd_CC_Stall_c;
				break;
			default:
				err("DoOut: Invalid Token 0x%x", b);
#ifdef DEBUG_OTG_CHIP
info("DoOut: Dump of rest of messages:");
while (otg_chip_HasData())
{
	UINT8 b;

	b = otg_chip_ReadFrame();
	info("DoOut: === 0x%0x", b);
}
#endif
		}
	}
	else
	{
		*w2 |= otg_etd_CC_DeviceNotResponding_c;
	}

	chip->etdPacketDone[index]++;

	if (!(*w3 & otg_td_ctrl_PacketCount_c))
	{
		/* Last packet. Retire ETD */
		otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
		otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_FilledStatus_c, filled);
	}
	else if (!(chip->etdPacketDone[index] % count))
	{
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_FilledStatus_c, filled);
	}
}

static void
otg_chip_DoIn(UINT32 w0, UINT32 w1, UINT32 *w2, UINT32 *w3,
int index, otg_sim_device_t *dev)
{
	int mem_addr;
	int packets;
	int count;
	int last_size;
	int max_length;
	int i;
	int yBuffer;
	UINT8 b;
	UINT32 filled;

#ifdef DEBUG_OTG_CHIP
info("DoIn");
#endif

	otg_sim_device_DoFrame(dev);

	packets = (*w3 & otg_td_ctrl_PacketCount_c) >> otg_td_ctrl_PacketCountLoc_c;
	count = (*w3 & otg_td_ctrl_NumPacketsPerBuff_c) >> 
		otg_td_ctrl_NumPacketsPerBuffLoc_c;
	count++;

	max_length = (w0 & otg_etd_MaxPacketSize_c) >> otg_etd_MaxPacketSizeLoc_c;
	if (count > packets)
	{
		count = packets;
	}

	if (!count)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIn: packet count is zero");
#endif
		return;
	}

	mem_addr = w1 & otg_etd_BufferField_c;
	yBuffer = 0;
	filled = 1 << index;
	if ((int)(chip->etdPacketDone[index] / count) & 1)
	{
		yBuffer = 1;
		filled <<= 16;
		mem_addr = (w1 >> otg_etd_YBufferLoc_c) & otg_etd_BufferField_c;
	}

	if (otg_chip_HasData())
	{
		b = otg_chip_ReadFrame();
		if (b == otg_chip_TokenNAK)
		{
			/* NAK is simulator means device is not responding */
#ifdef DEBUG_OTG_CHIP
			info("DoIn: NAK received");
#endif

			*w2 &= ~otg_td_ctrl_CompletionCode_c;
			*w2 |= otg_etd_CC_DeviceNotResponding_c;
			otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
			otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
			otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);

			return;
		}

		if (b != otg_chip_TokenDATA0 && b != otg_chip_TokenDATA1)
		{
			err("DoIn: Invalid toke received 0x%0x", b);
		}
	}

	last_size = otg_chip_ReadFrame();
	last_size |= otg_chip_ReadFrame() << 8;

	for (i = 0; i < last_size; i++)
	{
		b = otg_chip_ReadFrame();
		otg_chip_WriteMem(mem_addr, b);
		mem_addr++;
	}

	(*w3)--;

	*w2 &= ~otg_td_ctrl_CompletionCode_c;
	*w2 |= otg_etd_CC_NoError_c;

	chip->etdPacketDone[index]++;

	if (!(*w3 & otg_td_ctrl_PacketCount_c) || last_size < max_length)
	{
		/* last packet */
		*w3 &= ~otg_td_ctrl_LastPacketSize_c;
		*w3 |= last_size << otg_td_ctrl_LastPacketSizeLoc_c;

		/* Last packet. Retire ETD */
		otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
		otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);
		otg_chip_SetHostReg(otg_hc_etd_FilledStatus_c, filled);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
	}
	else if (!(chip->etdPacketDone[index] % count))
	{
		otg_chip_SetHostReg(otg_hc_etd_FilledStatus_c, filled);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
	}

	otg_chip_WriteFrame(otg_chip_TokenACK);
	otg_sim_device_DoFrame(dev);
}

static void
otg_chip_DoCtrl(int index, UINT32 dword0, otg_sim_device_t *dev)
{
	UINT32 dword1;
	UINT32 dword2;
	UINT32 dword3;
	UINT32 w;
	int addr;

	dword1 = otg_chip_ReadEtd(index, 1);
	dword2 = otg_chip_ReadEtd(index, 2);
	dword3 = otg_chip_ReadEtd(index, 3);
#ifdef DEBUG_OTG_CHIP
info("DoCtrl: ETD%d dword0=0x%0x dword1=0x%0x dword2=0x%0x dword3=0x%0x",
	index, dword0, dword1, dword2, dword3);
#endif

	if (dword3 & otg_td_ctrl_PacketCount_c)
	{
		if (!otg_chip_IsInEtd(index))
		{
			/* If there is packet to be transfered but no data
			 * trigger an interrupt */
			w = otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c);
			if (!(w & (1 << index)))
			{
#ifdef DEBUG_OTG_CHIP
info("DoCtrl: no data trigger interrupt. FilledStatus=0x%0x", w);
#endif
				otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
				return;
			}
		}
	}

	switch (dword2 & otg_td_ctrl_Pid_c)
	{
		case otg_td_ctrl_Pid_Setup_c:
			otg_chip_WriteFrame(otg_chip_TokenSETUP);
			break;
		case otg_td_ctrl_Pid_In_c:
			otg_chip_WriteFrame(otg_chip_TokenIN);
			break;
		case otg_td_ctrl_Pid_Out_c:
			otg_chip_WriteFrame(otg_chip_TokenOUT);
			break;
		default:
			err("Invalid PID 0x%x in Etd %d", dword2, index);
			return;
	}

	addr = dword0 & (otg_etd_FunctionAddress_c | otg_etd_EndpointNumber_c);
	otg_chip_WriteFrame(addr & 0xFF);
	otg_chip_WriteFrame((addr >> 8) & 0xFF);
	switch (dword2 & otg_td_ctrl_Pid_c)
	{
		case otg_td_ctrl_Pid_Setup_c:
		case otg_td_ctrl_Pid_Out_c:
			otg_chip_DoOut(dword0, dword1, &dword2, &dword3, index, dev);
			otg_chip_WriteEtd(index, 2, dword2);
			otg_chip_WriteEtd(index, 3, dword3);
			break;
		case otg_td_ctrl_Pid_In_c:
			otg_chip_DoIn(dword0, dword1, &dword2, &dword3, index, dev);
			otg_chip_WriteEtd(index, 2, dword2);
			otg_chip_WriteEtd(index, 3, dword3);
			break;
	}
}

static void
otg_chip_DoBulk(int index, UINT32 dword0, otg_sim_device_t *dev)
{
	UINT32 dword1;
	UINT32 dword2;
	UINT32 dword3;
	UINT32 w;
	int addr;
	int isOUT;

	dword1 = otg_chip_ReadEtd(index, 1);
	dword2 = otg_chip_ReadEtd(index, 2);
	dword3 = otg_chip_ReadEtd(index, 3);

#ifdef DEBUG_OTG_CHIP
info("DoBulk: ETD%d dword0=0x%0x dword1=0x%0x dword2=0x%0x dword3=0x%0x",
	index, dword0, dword1, dword2, dword3);
#endif

	if (dword3 & otg_td_ctrl_PacketCount_c)
	{
		if (!otg_chip_IsInEtd(index))
		{
			UINT32 mask;
			int count;

			/* If there is packet to be transfered but no data
			 * trigger an interrupt */
			count = (dword3 & otg_td_ctrl_NumPacketsPerBuff_c) >>
					otg_td_ctrl_NumPacketsPerBuffLoc_c;
			mask = 1 << index;
			if ((int)(chip->etdPacketDone[index] / count) & 1)
			{
				mask <<= 16;
			}

			w = otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c);
			if (!(w & mask))
			{
#ifdef DEBUG_OTG_CHIP
info("DoBulk: no data trigger interrupt. Filled=0x%0x mask=0x%0x", w, mask);
#endif
				otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
				return;
			}
		}
	}

	switch (dword0 & otg_etd_Direction_c)
	{
		case otg_etd_Direction_OUT_c:
			isOUT = 1;
			break;
		case otg_etd_Direction_IN_c:
			isOUT = 0;
			break;
		default:
			switch (dword2 & otg_td_ctrl_Pid_c)
			{
				case otg_td_ctrl_Pid_Out_c:
					isOUT = 1;
					break;
				case otg_td_ctrl_Pid_In_c:
					isOUT = 0;
					break;
				default:
					err("Invalid PID 0x%x in Etd %d", dword2, index);
					return;
			}
	}

	if (isOUT)
	{
		otg_chip_WriteFrame(otg_chip_TokenOUT);
	}
	else
	{
		otg_chip_WriteFrame(otg_chip_TokenIN);
	}

	addr = dword0 & (otg_etd_FunctionAddress_c | otg_etd_EndpointNumber_c);
	otg_chip_WriteFrame(addr & 0xFF);
	otg_chip_WriteFrame((addr >> 8) & 0xFF);

	if (isOUT)
	{
		otg_chip_DoOut(dword0, dword1, &dword2, &dword3, index, dev);
	}
	else
	{
		otg_chip_DoIn(dword0, dword1, &dword2, &dword3, index, dev);
	}

	otg_chip_WriteEtd(index, 2, dword2);
	otg_chip_WriteEtd(index, 3, dword3);
}

static void
otg_chip_DoIntr(int index, UINT32 dword0, otg_sim_device_t *dev)
{
	UINT32 dword2;
	int m;
	int t;

	dword2 = otg_chip_ReadEtd(index, 2);
	m = (dword2 & otg_td_ctrl_RelPollingPos_c)>>otg_td_ctrl_RelPollingPosLoc_c;
	t = otg_chip_ReadHostReg(otg_hc_FmNumber_c);
	if ((m & 0xFF) != (t & 0xFF))
	{
		/* Time not reached yet */
		return;
	}

#ifdef DEBUG_OTG_CHIP
	info("DoIntr: doing Interrupt transfer");
#endif
	otg_chip_DoBulk(index, dword0, dev);
}

static void
otg_chip_DoIsoOut(UINT32 w0, UINT32 w1, UINT32 w2, int index,
otg_sim_device_t *dev)
{
	int mem_addr;
	int packets;
	int count;
	int size;
	int max_length;
	int yBuffer;
	int i;
	UINT32 filled;
	UINT32 w3;
	int frame;

#ifdef DEBUG_OTG_CHIP
info("DoIsoOut: index=%d dev=%p FilledStatus=0x%0x",
	index, dev, otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c));
#endif
	packets = (w2 & otg_td_iso_FrameCount_c) >> otg_td_iso_FrameCountLoc_c;
	frame = otg_chip_ReadHostReg(otg_hc_FmNumber_c) -
		(w2 & otg_td_iso_StartingFrame_c);

	w3 = otg_chip_ReadEtd(index, (frame >> 1) + 3);
	otg_chip_WriteDataToggle(&w2);
	count = (w2 & otg_td_iso_NumPacketsPerBuff_c) >>
			otg_td_iso_NumPacketsPerBuffLoc_c;
	count++;

	if (frame > packets)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIsoOut: this ETD should retire.");
#endif
	}

	if (count > (packets + 1))
	{
		count = packets + 1;
	}

	if (!count)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIsoOut: packet count is zero");
#endif
		return;
	}

	if (frame & 1)
	{
		yBuffer = 1;
		filled = 1 << (16 + index);
		size = (w3 >> otg_td_iso_UpperDwordLoc_c) & otg_td_iso_LowerLenPSW_c;
		mem_addr = (w1 >> otg_etd_YBufferLoc_c) & otg_etd_BufferField_c;
	}
	else
	{
		yBuffer = 0;
		filled = 1 << index;
		size = w3 & otg_td_iso_LowerLenPSW_c;
		mem_addr = w1 & otg_etd_BufferField_c;
	}

	max_length = (w0 & otg_etd_MaxPacketSize_c) >> otg_etd_MaxPacketSizeLoc_c;

#ifdef DEBUG_OTG_CHIP
info("DoIsoOut: count=%d packets=%d mem_addr=0x%0x size=0x%0x filled=0x%0x",
	count, packets, mem_addr, size, filled);
#endif

	otg_chip_WriteFrame(size & 0xFF);
	otg_chip_WriteFrame((size >> 8) & 0xFF);
	for (i = 0; i < size; i++)
	{
		otg_chip_WriteFrame(otg_chip_ReadMem(mem_addr));
		mem_addr++;
	}

	otg_sim_device_DoFrame(dev);

	w2 &= ~(otg_td_iso_CompletionCode_c << 16);
	w2 |= otg_etd_CC_NoError_c;

	if (frame & 1)
	{
		w3 &= ~(otg_td_iso_CompletionCode_c << 16);
		w3 |= otg_etd_CC_NoError_c;
	}
	else
	{
		w3 &= ~otg_td_iso_CompletionCode_c;
		w3 |= (otg_etd_CC_NoError_c >> 16) & otg_td_iso_CompletionCode_c;
	}

	otg_chip_WriteEtd(index, 2, w2);
	otg_chip_WriteEtd(index, 3 + (frame >> 1), w3);

	if (packets <= frame)
	{
		/* Last packet. Retire ETD */
		otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
		otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_FilledStatus_c, filled);
	}
	else if (!(frame % count))
	{
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_FilledStatus_c, filled);
	}
}

static void
otg_chip_DoIsoIn(UINT32 w0, UINT32 w1, UINT32 w2, int index,
otg_sim_device_t *dev)
{
	int mem_addr;
	int packets;
	int count;
	int size;
	int max_length;
	int i;
	int yBuffer;
	UINT8 b;
	UINT32 filled;
	UINT32 w3;
	int frame;

#ifdef DEBUG_OTG_CHIP
info("DoIsoIn");
#endif

	otg_sim_device_DoFrame(dev);

	packets = (w2 & otg_td_iso_FrameCount_c) >> otg_td_iso_FrameCountLoc_c;
	frame = otg_chip_ReadHostReg(otg_hc_FmNumber_c) -
		(w2 & otg_td_iso_StartingFrame_c);

	w3 = otg_chip_ReadEtd(index, (frame >> 1) + 3);
	count = (w2 & otg_td_iso_NumPacketsPerBuff_c) >>
			otg_td_iso_NumPacketsPerBuffLoc_c;
	count++;

	if (frame > packets)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIsoIn: data overrun. this ETD should be retired");
#endif
	}

	if (count > (packets + 1))
	{
		count = packets + 1;
	}

	if (!count)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIsoIn: packet count is zero");
#endif
		return;
	}

	if (frame & 1)
	{
		yBuffer = 1;
		filled = 1 << (16 + index);
		mem_addr = (w1 >> otg_etd_YBufferLoc_c) & otg_etd_BufferField_c;
		max_length = (w3 >> otg_td_iso_UpperDwordLoc_c) &
					otg_td_iso_LowerLenPSW_c;
	}
	else
	{
		mem_addr = w1 & otg_etd_BufferField_c;
		yBuffer = 0;
		filled = 1 << index;
		max_length = w3 & otg_td_iso_LowerLenPSW_c;
	}

	if (otg_chip_HasData())
	{
		b = otg_chip_ReadFrame();
		if (b == otg_chip_TokenNAK)
		{
			/* NAK is simulator means device is not responding */
#ifdef DEBUG_OTG_CHIP
			info("DoIsoIn: NAK received");
#endif

			w2 &= ~otg_td_ctrl_CompletionCode_c;
			w2 |= otg_etd_CC_DeviceNotResponding_c;
			if (frame & 1)
			{
				w3 &= otg_td_iso_CompletionCode_c << 16;
				w3 |= otg_etd_CC_DeviceNotResponding_c;
			}
			else
			{
				w3 &= otg_td_iso_CompletionCode_c;
				w3 |= (otg_etd_CC_DeviceNotResponding_c >>
					otg_td_iso_UpperDwordLoc_c) & otg_td_iso_CompletionCode_c;
			}

			otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
			otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
			otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);

			return;
		}

		if (b != otg_chip_TokenDATA0 && b != otg_chip_TokenDATA1)
		{
			err("DoIsoIn: Invalid toke received 0x%0x", b);
		}
	}

	size = otg_chip_ReadFrame();
	size |= otg_chip_ReadFrame() << 8;
#ifdef DEBUG_OTG_CHIP
info("DoIsoIn: size=%d", size);
#endif
	for (i = 0; i < size; i++)
	{
		b = otg_chip_ReadFrame();
		otg_chip_WriteMem(mem_addr, b);
		mem_addr++;
	}

	w2 &= ~otg_td_ctrl_CompletionCode_c;
	w2 |= otg_etd_CC_NoError_c;

	if (frame & 1)
	{
		w3 &= ~((otg_td_iso_InSize_c | otg_td_iso_CompletionCode_c) <<
			otg_td_iso_UpperDwordLoc_c);
		w3 |= (size << otg_td_iso_UpperDwordLoc_c) | otg_etd_CC_NoError_c;
	}
	else
	{
		w3 &= ~(otg_td_iso_InSize_c | otg_td_iso_CompletionCode_c);
		w3 |= size | ((otg_etd_CC_NoError_c >> otg_td_iso_UpperDwordLoc_c) &
			otg_td_iso_CompletionCode_c);
	}

	if ((packets <= frame) || size < max_length)
	{
		/* Last packet. Retire ETD */
		otg_chip_SetHostReg(otg_hc_etd_DoneQueue_c, 1 << index);
		otg_chip_ClearHostReg(otg_hc_etd_EnableStatus_c, 1 << index);
		otg_chip_SetHostReg(otg_hc_intr_Status_c, otg_hc_intr_RDQ_c);
		otg_chip_SetHostReg(otg_hc_etd_FilledStatus_c, filled);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
	}
	else if (!(frame % count))
	{
		otg_chip_SetHostReg(otg_hc_etd_FilledStatus_c, filled);
		otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
	}

	otg_chip_WriteEtd(index, 2, w2);
	otg_chip_WriteEtd(index, 3 + (frame >> 1), w3);
}


static void
otg_chip_DoIso(int index, UINT32 dword0, otg_sim_device_t *dev)
{
	UINT32 dword1;
	UINT32 dword2;
	UINT32 w;
	int addr;
	int isOUT;
	int frame;
	int frames;

	dword1 = otg_chip_ReadEtd(index, 1);
	dword2 = otg_chip_ReadEtd(index, 2);

	frame = otg_chip_ReadHostReg(otg_hc_FmNumber_c) -
		(dword2 & otg_td_iso_StartingFrame_c);

	if (frame < 0)
	{
		return;
	}
	
	frames = (dword2 & otg_td_iso_FrameCount_c) >> otg_td_iso_FrameCountLoc_c;
#ifdef DEBUG_OTG_CHIP
info("DoIso: ETD%d dword0=0x%0x dword1=0x%0x dword2=0x%0x",
	index, dword0, dword1, dword2);
#endif

	if (frame > 7)
	{
#ifdef DEBUG_OTG_CHIP
		info("DoIso: Iso frame missed frame=%d", frame);
#endif
		return;
	}

	if (frames)
	{
		if (!otg_chip_IsInEtd(index))
		{
			UINT32 mask;

			mask = 1 << index;
			if (frame & 1)
			{
				mask <<= 16;
			}

			w = otg_chip_ReadHostReg(otg_hc_etd_FilledStatus_c);
			if (!(w & mask))
			{
#ifdef DEBUG_OTG_CHIP
info("DoIso: no data trigger interrupt. Filled=0x%0x mask=0x%0x", w, mask);
#endif
				otg_chip_SetHostReg(otg_hc_etd_IntrStatus_c, 1 << index);
				return;
			}
		}
	}

	switch (dword0 & otg_etd_Direction_c)
	{
		case otg_etd_Direction_OUT_c:
			isOUT = 1;
			break;
		case otg_etd_Direction_IN_c:
			isOUT = 0;
			break;
		default:
			switch (dword2 & otg_td_ctrl_Pid_c)
			{
				case otg_td_ctrl_Pid_Out_c:
					isOUT = 1;
					break;
				case otg_td_ctrl_Pid_In_c:
					isOUT = 0;
					break;
				default:
					err("Invalid PID 0x%x in Etd %d", dword2, index);
					return;
			}
	}

	if (isOUT)
	{
		otg_chip_WriteFrame(otg_chip_TokenOUT);
	}
	else
	{
		otg_chip_WriteFrame(otg_chip_TokenIN);
	}

	addr = dword0 & (otg_etd_FunctionAddress_c | otg_etd_EndpointNumber_c);
	otg_chip_WriteFrame(addr & 0xFF);
	otg_chip_WriteFrame((addr >> 8) & 0xFF);

	if (isOUT)
	{
		otg_chip_DoIsoOut(dword0, dword1, dword2, index, dev);
	}
	else
	{
		otg_chip_DoIsoIn(dword0, dword1, dword2, index, dev);
	}
}

static void
otg_chip_DoEtd(int index)
{
	UINT32 dword0;
	otg_sim_device_t *dev;
	int i;

	dword0 = otg_chip_ReadEtd(index, 0);
	for (i = 0; i < 3; i++)
	{
		if (!chip->ports[i])
		{
			continue;
		}
	
		dev = chip->ports[i];
		switch (dword0 & otg_etd_Format_c)
		{
			case otg_etd_Format_Ctrl_c:
				otg_chip_DoCtrl(index, dword0, dev);
				break;
			case otg_etd_Format_Iso_c:
				otg_chip_DoIso(index, dword0, dev);
				break;
			case otg_etd_Format_Bulk_c:
				otg_chip_DoBulk(index, dword0, dev);
				break;
			case otg_etd_Format_Intr_c:
				otg_chip_DoIntr(index, dword0, dev);
				break;
		}
	}
}

static void
otg_chip_DoEvent(void)
{
	int i;
	UINT32 w;
	UINT32 status;

	status = otg_chip_ReadHostReg(otg_hc_Control_c);
	frame_write = 0;
	frame_read = 0;
	if ((status & otg_hc_HostState_c) == otg_hc_USBOperational_c)
	{
		/* Increase frame count */
		w = otg_chip_ReadHostReg(otg_hc_FmNumber_c);
		w++;
		w &= otg_hc_FmMask_c;
		otg_chip_WriteHostReg(otg_hc_FmNumber_c, w);

		/* Send SOF */
		otg_chip_WriteFrame(otg_chip_TokenSOF);
		otg_chip_WriteFrame(w & 0xFF);
		otg_chip_WriteFrame((w >> 8)& 0xFF); /* ignore CRC */

		w = 1;
		status = otg_chip_ReadHostReg(otg_hc_etd_EnableStatus_c);
		for (i = 0; i < 16; i++)
		{
			if (status & w)
			{
				otg_chip_DoEtd(i);
			}

			w <<= 1;
		}

		if (otg_chip_ReadHostReg(otg_hc_etd_IntrStatus_c) &
			(otg_chip_ReadHostReg(otg_hc_etd_IntrEnable_c) |
			(otg_chip_ReadHostReg(otg_hc_etd_IntrEnable_c) << 16)))
		{
			otg_chip_SetReg(otg_intr_Status_c, otg_intr_H_c);
		}
		else if (otg_chip_ReadHostReg(otg_hc_intr_Status_c) &
			otg_chip_ReadHostReg(otg_hc_intr_Enable_c))
		{
			otg_chip_SetReg(otg_intr_Status_c, otg_intr_H_c);
		}

		if (otg_chip_ReadFcReg(otg_ep_intr_Status) &
			otg_chip_ReadFcReg(otg_ep_intr_Enable))
		{
			otg_chip_SetReg(otg_intr_Status_c, otg_intr_F_c);
		}
		else if (otg_chip_ReadFcReg(otg_fc_intr_Status) &
			otg_chip_ReadFcReg(otg_fc_intr_Enable))
		{
			otg_chip_SetReg(otg_intr_Status_c, otg_intr_F_c);
		}

		if (otg_chip_GetReg(otg_intr_Status_c) &
			otg_chip_GetReg(otg_intr_Enable_c))
		{
			(*otg_hcd_GetISR())(7, otg_hcd_GetDevice(), 0);
		}
	}
}

static int
otg_chip_thread(void *p)
{
	lock_kernel();

	daemonize();
	strcpy(current->comm, "kotg_chipd");
	do
	{
		otg_chip_DoEvent();
		unlock_kernel();
		wait_ms(1);
		lock_kernel();
	} while (!signal_pending(current));

	unlock_kernel();
	up_and_exit(&kotg_chip_exited, 0);
}

int
otg_chip_Init(void)
{
	chip = (otg_chip_t *)OS_Malloc(sizeof(otg_chip_t));
	if (!chip)
	{
		err("Failed to allocate memory");
		return -1;
	}

#ifdef DEBUG_OTG_CHIP
	info("chip=%p", chip);
#endif

	OS_MemSet(chip, 0, sizeof(otg_chip_t));

	chip->timer_fun = 0; /* No timer function */

	process_id = kernel_thread(otg_chip_thread, NULL,
		CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
	
	if (process_id > 0)
	{
		return 0;
	}

	process_id = -1;
	err("Failed to start otg_chip_thread");
	return -1;
}

void
otg_chip_Exit(void)
{
	if (process_id > 0)
	{
		kill_proc(process_id, SIGTERM, 1);
		down(&kotg_chip_exited);
		process_id = -1;
	}
	else
	{
		err("Wrong process_id %d", process_id);
	}

	if (chip)
	{
		OS_Free(chip);
		chip = 0;
	}
}

int
otg_chip_RhPortStatusChange(int port)
{
	int reg;
	UINT32 w;

	switch (port)
	{
		case 0:
			reg = otg_hc_rh_port_Status0_c;
			break;
		case 1:
			reg = otg_hc_rh_port_Status1_c;
			break;
		case 2:
			reg = otg_hc_rh_port_Status2_c;
			break;
		default:
			reg = otg_hc_rh_port_Status0_c;
			break;
	}

	w = otg_chip_ReadHostReg(reg);
	w ^= otg_hc_rh_port_Status_CCS_c;
	w |= otg_hc_rh_port_Status_CSC_c;
	otg_chip_WriteHostReg(reg, w);

	return 0;
}

/*
static int __init otg_chip_init(void)
{
	otg_chip_Init();

	return 0;
}

static void __exit otg_chip_exit(void)
{
	otg_chip_Exit();
}

module_init(otg_chip_init);
module_exit(otg_chip_exit);
*/

EXPORT_SYMBOL(otg_chip_RhPortStatusChange);
EXPORT_SYMBOL(otg_chip_GetChip);
EXPORT_SYMBOL(otg_chip_ReadReg);
EXPORT_SYMBOL(otg_chip_WriteReg);
EXPORT_SYMBOL(otg_chip_HasData);
EXPORT_SYMBOL(otg_chip_WriteFrame);
EXPORT_SYMBOL(otg_chip_ReadFrame);
EXPORT_SYMBOL(otg_chip_HasMoreData);
EXPORT_SYMBOL(otg_chip_WriteMem);
EXPORT_SYMBOL(otg_chip_ReadMem);
EXPORT_SYMBOL(otg_chip_GetReg);
EXPORT_SYMBOL(otg_chip_WriteFcReg);
