/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_driver.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    Host side OTG device driver.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

#include "otg_driver.h"

static DECLARE_WAIT_QUEUE_HEAD(kotgd_wait);
static int kotgd_pid = 0;
static OS_CompletionCreate(kotgd_exited);

struct otgDevice
{
	stack_device_t *dev;
	stack_urb_t statusUrb;

	UINT16 status;

	int supportHNP;
	int supportSRP;
	int hnpPort;
} otgDevice;

static int
otg_driver_SetFeature(stack_device_t *dev, int feature)
{
	return usb_control_msg
			(
				dev,
				usb_sndctrlpipe(dev, 0),
				USB_REQ_SET_FEATURE,
				USB_TYPE_STANDARD | USB_RECIP_DEVICE,
				feature,
				0,
				0,
				0,
				HZ * 5
			);
}

static int
otg_driver_GetOTGDescriptor(stack_device_t *dev, struct otgDevice *otg)
{
	int rc;
	UINT8 data[otg_driver_OtgDescLen];

	rc = usb_control_msg
		(
			dev,
			usb_rcvctrlpipe(dev, 0),
			USB_REQ_GET_DESCRIPTOR,
			USB_DIR_IN,
			otg_driver_OtgDescType << 8,
			0,
			data,
			otg_driver_OtgDescLen,
			HZ
		);

	if (rc < 0)
	{
		return -1;
	}

#ifdef OTG_HNP_DEBUG
info("GetOTGDescriptor: data[0]=0x%0x data[1]=0x%0x data[2]=0x%0x",
	data[0], data[1], data[2]);
#endif
	if (data[0] != otg_driver_OtgDescLen ||
		data[1] != otg_driver_OtgDescType)
	{
		return -1;
	}

	return data[2];
}

static void
otg_driver_Irq(stack_urb_t *urb)
{
	struct otgDevice *device = stack_urb_GetContext(urb);

	/* If the device request host, do the change */
	if (stack_urb_GetStatus(urb))
	{
		return;
	}

	info("Irq: received valid data. status = 0x%0x", device->status);
	if (device->status & 1)
	{
		info("Irq: B-device is ready to take over host");
		wake_up(&kotgd_wait);
	}
}

static void *
otg_driver_Probe
(
	stack_device_t *dev,
	unsigned int ifnum,
	const struct usb_device_id *id
)
{
	struct usb_interface_descriptor *intf;
	unsigned char *extra;
	stack_bus_t *bus;
	stack_device_t *root_hub;
	struct usb_endpoint_descriptor *endpoint;
	otg_hc_t *hc;

	info("Probe: otg=%p dev=%p ifnum=%d", &otgDevice, dev, ifnum);

	intf = dev->actconfig->interface[ifnum].altsetting;
	info("Probe: intf=%p", intf);
	if (!intf)
	{
		err("no interface");
		return 0;
	}

	endpoint = intf->endpoint;
	if (!endpoint)
	{
		return 0;
	}

	extra = intf->extra;
	if (!extra || intf->extralen < otg_driver_OtgDescLen)
	{
		return 0;
	}

	info("Probe: 0x%0x 0x%0x 0x%0x", *extra, *(extra + 1), *(extra + 2));

	if (*extra != otg_driver_OtgDescLen)
	{
		return 0;
	}

	if (*(extra + 1) != otg_driver_OtgDescType)
	{
		return 0;
	}

	if (*(extra + 2) & otg_driver_OtgSrpSupport)
	{
		otgDevice.supportSRP = 1;
	}

	/* Temporary. Need to change usbd structure to handle this */
	bus = otg_hcd_GetUsbBus();
	hc = otg_hcd_GetHostController();
#ifdef OTG_HNP_DEBUG
info("Probe: hc=%p otg=%p", hc, hc->otg);
#endif

	if (*(extra + 2) & otg_driver_OtgHnpSupport)
	{
		otgDevice.supportHNP = 1;
		otg_SetBHnpSupport(hc->otg);
	}
	else
	{
		otgDevice.supportHNP = 0;
		otg_ClearBHnpSupport(hc->otg);
	}

	otgDevice.dev = dev;

	bus = dev->bus;
	root_hub = bus->root_hub;

	if (root_hub->children[0] == dev)
	{
		otgDevice.hnpPort = 1;
		if (otgDevice.supportHNP)
		{
			otg_driver_SetFeature(dev, otg_ep0_FEATURE_A_HNP_SUPPORT);
		}
	}
	else
	{
		otgDevice.hnpPort = 0;
		if (otgDevice.supportHNP)
		{
			otg_driver_SetFeature(dev, otg_ep0_FEATURE_A_ALT_HNP_SUPPORT);
		}
	}

	info("Probe: hnpPort=%d", otgDevice.hnpPort);

	FILL_INT_URB
	(
		&otgDevice.statusUrb,
		otgDevice.dev,
		usb_rcvintpipe(otgDevice.dev, endpoint->bEndpointAddress),
		&otgDevice.status,
		2,
		otg_driver_Irq,
		&otgDevice,
		endpoint->bInterval
	);

	if (usb_submit_urb(&otgDevice.statusUrb))
	{
		err("Probe: submit statuUrb failed");
	}

	return &otgDevice;
}

static void
otg_driver_Disconnect(stack_device_t *dev, void *ptr)
{
	struct otgDevice *otg;

	otg = (struct otgDevice *)ptr;
	info("Disconnect: otg device disconnected otg=%p", otg);
	if (!otg)
	{
		return;
	}

	otg->dev = 0;

	usb_unlink_urb(&otgDevice.statusUrb);
}

static struct usb_device_id otgDriverId [] =
{
	{ USB_DEVICE(0x2848, 0x0243) },
	{}
};

static struct usb_driver otgDriver =
{
	name:		"otg_driver",
	probe:		otg_driver_Probe,
	disconnect:	otg_driver_Disconnect,
	id_table:	otgDriverId,
};

int
otg_driver_ReleaseBus(void)
{
	stack_bus_t *bus;
	otg_hc_t *hc;
	stack_device_t *root_hub;
	stack_device_t *dev;
	int hnp;
	int i;
	int rc;

	bus = otg_hcd_GetUsbBus();
	hc = otg_hcd_GetHostController();
	if (!bus || !hc)
	{
		/* No host controller */
		return otg_driver_ErrorNoHost;
	}

#ifdef OTG_HNP_DEBUG
info("ReleaseBus: hc=%p", hc);
#endif
	root_hub = bus->root_hub;
	dev = root_hub->children[otg_hc_rh_port_HNPPort];
	if (!dev)
	{
#ifdef OTG_HNP_DEBUG
info("ReleaseBus: no device on HNP port");
#endif
		for(i = 0; i < otg_hc_rh_MaxPorts; i++)
		{
			if (i == otg_hc_rh_port_HNPPort)
			{
				continue;
			}

			dev = root_hub->children[i];
			if (!dev)
			{
				continue;
			}

			hnp = otg_driver_GetOTGDescriptor(dev, &otgDevice);
			if (hnp & otg_driver_OtgHnpSupport ||
				hnp & otg_driver_OtgSrpSupport)
			{
				/* HNP supported on a different port */
				otg_driver_SetFeature(dev, otg_ep0_FEATURE_A_ALT_HNP_SUPPORT);
				return otg_driver_ErrorAltHnp;
			}
		}

		/* No HNP device found */
		return otg_driver_ErrorNoDevice;
	}

	hnp = otg_driver_GetOTGDescriptor(dev, &otgDevice);
#ifdef OTG_HNP_DEBUG
info("ReleaseBus: hnp=%x", hnp);
#endif
	if (hnp < 0)
	{
		/* No HNP device found */
		return otg_driver_ErrorNoDevice;
	}

	if (!(hnp & otg_driver_OtgHnpSupport) &&
		!(hnp & otg_driver_OtgSrpSupport))
	{
		/* No HNP device found */
		return otg_driver_ErrorNoDevice;
	}

	if (hnp & otg_driver_OtgSrpSupport)
	{
		otgDevice.supportSRP = 1;
	}

	if (hnp & otg_driver_OtgHnpSupport)
	{
		otgDevice.supportHNP = 1;
		otg_SetBHnpSupport(hc->otg);
	}
	else
	{
		otgDevice.supportHNP = 0;
		otg_ClearBHnpSupport(hc->otg);
	}

#ifdef OTG_HNP_DEBUG
info("ReleaseBus: hnp=%x", hnp);
#endif
	otgDevice.dev = dev;
	otgDevice.hnpPort = 1;

	rc = otg_driver_SetFeature(dev, otg_ep0_FEATURE_A_HNP_SUPPORT);
	if (rc < 0)
	{
		/* B-device not respond */
#ifdef OTG_HNP_DEBUG
info("ReleaseBus: Set A HNP Support failed");
#endif
		return otg_driver_ErrorNotRespond;
	}

	if (otg_driver_SetFeature(dev, otg_ep0_FEATURE_B_HNP_ENABLE))
	{
		/* B-device not respond */
#ifdef OTG_HNP_DEBUG
info("ReleaseBus: Set B HNP failed");
#endif
		return otg_driver_ErrorNotRespond;
	}
	else
	{
#ifdef OTG_HNP_DEBUG
info("ReleaseBus: hc=%p otg=%p", hc, hc->otg);
#endif
		otg_TransferHostToB(hc->otg);
	}

	return 0;
}

static int otg_driver_Thread(void *otg)
{
	lock_kernel();
	daemonize();
	strcpy(current->comm, "kotgd");

	do
	{
		unlock_kernel();
		interruptible_sleep_on(&kotgd_wait);
		if (otgDevice.status & 1)
		{
			otg_driver_ReleaseBus();
		}
		lock_kernel();
	} while (!signal_pending(current));

	unlock_kernel();
	info("Thread: kotgd exiting");
	OS_CompletionSignal(&kotgd_exited, 0);
}

static int __init otg_driver_init(void)
{
	int pid;

	if (usb_register(&otgDriver))
	{
		return -1;
	}

	OS_MemSet(&otgDevice, 0, sizeof(otgDevice));

	pid = kernel_thread
		(
			otg_driver_Thread,
			NULL,
			CLONE_FS | CLONE_FILES | CLONE_SIGHAND
		);
	if (pid >= 0)
	{
		kotgd_pid = pid;
		return 0;
	}

	usb_deregister(&otgDriver);
	err("init: start otg driver failed");

	return -1;
}

static void __exit otg_driver_exit(void)
{
	int ret;

	ret = kill_proc(kotgd_pid, SIGTERM, 1);
	OS_CompletionWait(&kotgd_exited);

	usb_deregister(&otgDriver);
}

module_init(otg_driver_init);
module_exit(otg_driver_exit);

#ifndef MODULE
EXPORT_SYMBOL(otg_driver_ReleaseBus);
#endif
