/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_ep.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    Function side endpoint object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

void
otg_ep_Init(otg_ep_t *ep_bank, otg_fc_t *fc)
{
	otg_ep_t *ep;
	int i;

	ep = ep_bank;
	for (i = 0; i < otg_fc_MaxEP; i++)
	{
		ep->fc = fc;
		ep->desc = 0;
		ep->urb = 0;

		ep->addrX = -1;
		ep->addrY = -1;

		ep->index = i;

		ep++;
	}
}

void
otg_ep_Reset(otg_ep_t *ep_bank)
{
	otg_ep_t *ep;
	int i;

	ep = ep_bank;
	for (i = 0; i < otg_fc_MaxEP; i++)
	{
		otg_ep_Disable(ep);
		ep->halted = 0;

		ep++;
	}
}

void
otg_ep_DoIntr(otg_ep_t *ep, otg_ep_intr_t *intr, UINT32 filled)
{
	UINT32 w;

	if (!ep->dataPtr)
	{
		/* Donothing, hardware will send Nak automatically */
		return;
	}

	w = otg_ep_intr_CTRL1 << ep->index;
	if (otg_ep_intr_IsTriggered(intr, w))
	{
		otg_ep_intr_ClearIntr(intr, w);
		otg_ep_DoTCIntr(ep);
		return;
	}

	w = otg_ep_FillStatusBits << ep->index;
	if (otg_ep_desc_IsDirectionIn(ep->desc))
	{
		otg_ep_SendData(ep);
		if (!(filled ^ w))
		{
			otg_ep_SendData(ep);
		}
	}
	else
	{
		otg_ep_RecvData(ep, ep->packetSize, 0);
		if (!(filled ^ w))
		{
			otg_ep_RecvData(ep, ep->packetSize, 0);
		}
	}
}

void
otg_ep_DoTCIntr(otg_ep_t *ep)
{
	otg_urb_t *urb;
	UINT32 w;
	int size;

	urb = ep->urb;
	w = otg_fc_ReadReg(ep->fc, otg_ep_Ctrl(ep->index));
	size = (w >> otg_ep_CtrlLPSLoc) & 1023;
	if (!otg_ep_desc_IsDirectionIn(ep->desc))
	{
		/* read in the last packet */
		otg_ep_RecvData(ep, size, 1);
	}

	if (!ep->dataLen)
	{
		/* Zero length packet */
		urb->actual_length = 0;
	}
	else
	{
		urb->actual_length = (ep->currentPacket - 1) * ep->packetSize + size;
	}

	urb->status = 0;
	ep->urb = 0;
	if (urb->complete)
	{
		urb->complete(urb);
	}
}

int
otg_ep_Set(otg_ep_t *ep, otg_ep_desc_t *ep_desc)
{
	int size;

#ifdef DEBUG_OTG_CHIP
	info("Set: ep=%p ep_desc=%p", ep, ep_desc);
#endif
	ep->desc = ep_desc;
	ep->dataLen = 0;
	ep->dataPos = 0;
	ep->dataPtr = 0;

	if (ep->urb)
	{
		otg_urb_t *urb;

		/* ??? may need lock */
		urb = ep->urb;
		ep->urb = 0;
#ifdef DEBUG_OTG_CHIP
	info("Set: urb=%p", urb);
#endif
		urb->status = otg_urb_ErrorResetByHost;
		if (urb->complete)
		{
			urb->complete(urb);
		}
	}

	size = otg_ep_desc_GetMaxPacketSize(ep_desc);
	ep->addrX = otg_fc_AllocateMem(ep->fc, size);
	if (ep->addrX < 0)
	{
		ep->addrX = -1;
		return -ENOMEM;
	}

#ifdef DEBUG_OTG_CHIP
	info("Set: addrX=0x%0x", ep->addrX);
#endif
	ep->addrY = otg_fc_AllocateMem(ep->fc, size);
	if (ep->addrY < 0)
	{
		otg_fc_ReleaseMem(ep->fc, ep->addrX);
		ep->addrX = -1;
		ep->addrY = -1;
		return -ENOMEM;
	}

	ep->halted = 0;

#ifdef DEBUG_OTG_CHIP
	info("Set: addrY=0x%0x", ep->addrY);
#endif

	/* Set PacketCount to zero will cause NAK on OUT and IN packet */
	otg_fc_WriteReg(ep->fc, otg_ep_Ctrl(ep->index), 0);
	otg_ep_FillConfig(ep);
	if (otg_ep_desc_IsDirectionIn(ep->desc))
	{
		otg_fc_WriteReg(ep->fc, otg_ep_FillClear, 0x10001 << ep->index);
	}

	otg_ep_Enable(ep);

	return 0;
}

void
otg_ep_Release(otg_ep_t *ep)
{
#ifdef DEBUG_OTG_CHIP
info("Release: addrX=0x%0x addrY=0x%0x", ep->addrX, ep->addrY);
#endif
	if (ep->addrX >= 0)
	{
		otg_fc_ReleaseMem(ep->fc, ep->addrX);
		ep->addrX = -1;
	}

	if (ep->addrY >= 0)
	{
		otg_fc_ReleaseMem(ep->fc, ep->addrY);
		ep->addrY = -1;
	}

	ep->desc = 0;
}

int
otg_ep_Submit(otg_ep_t *ep, otg_urb_t *urb)
{
	int rc;

	if (ep->urb)
	{
#ifdef DEBUG_OTG_SUBMIT
info("Submit: ep is busy. urb=%p", urb);
#endif
		return -EBUSY;
	}

	ep->urb = urb;
	ep->urb->status = otg_urb_ErrorInProgress;
	ep->dataPtr = urb->transfer_buffer;
	ep->dataLen = urb->transfer_length;
	ep->dataPos = 0;

#ifdef DEBUG_OTG_SUBMIT
info("Submit: ep->index=%d", ep->index);
#endif
	/* setup control and buffer words and download them to otg chip */
	rc = otg_ep_FillCtrl(ep);
	if (rc)
	{
		ep->urb = 0;
		ep->urb->status = -EAGAIN;
		return rc;
	}

	if (otg_ep_desc_IsDirectionIn(ep->desc))
	{
		if (!ep->dataLen)
		{
			/* Send zero length packet */
			otg_fc_WriteReg(ep->fc, otg_ep_FillSet, 2 << ep->index);
		}
		else
		{
			/* Download data to otg chip */
			otg_ep_SendData(ep);
			if (ep->dataPos < ep->dataLen)
			{
				/* More data to send, prepare the second buffer */
				otg_ep_SendData(ep);
			}
		}
	}

	return 0;
}

void
otg_ep_SendData(otg_ep_t *ep)
{
	int addr;
	int size;
	int buffer;

#ifdef DEBUG_OTG_SUBMIT
info("SendData: dataPos=%d dataLen=%d dataPtr=%p",
	ep->dataPos, ep->dataLen, ep->dataPtr);
#endif
	addr = ep->addrX;
	buffer = 2;
	if (ep->currentPacket & 1)
	{
		/* Send to Y buffer */
		addr = ep->addrY;
		buffer = 0x20000;
	}

	if (ep->dataPos >= ep->dataLen)
	{
		/* Transfer finished, no more data */
#ifdef DEBUG_OTG_SUBMIT
		info("SendData: no more data to send");
#endif
		otg_ep_DoTCIntr(ep);
		return;
	}

	size = ep->dataLen - ep->dataPos;
	if (size > ep->packetSize)
	{
		size = ep->packetSize;
	}

#ifdef DEBUG_OTG_SUBMIT
info("SendData: addr=0x%0x size=%d", addr, size);
#endif
	otg_mem_FcWrite(ep->fc, addr, ep->dataPtr, size);
	otg_fc_WriteReg(ep->fc, otg_ep_FillSet, buffer << ep->index);

	ep->dataPos += size;
	ep->dataPtr = (void *)((unsigned long)ep->dataPtr + size);
	ep->currentPacket++;
#ifdef DEBUG_OTG_SUBMIT
info("SendData: dataPos=%d dataPtr=%p", ep->dataPos, ep->dataPtr);
#endif
}

void
otg_ep_RecvData(otg_ep_t *ep, int size, int is_last_pkg)
{
	int addr;
	int buffer;

#ifdef DEBUG_OTG_SUBMIT
info("RecvData: dataPos=%d dataLen=%d dataPtr=%p",
	ep->dataPos, ep->dataLen, ep->dataPtr);
#endif
	if (ep->dataPos >= ep->dataLen)
	{
		/* Transfer finished, no more data */
		info("RecvData: no more data to receive");
		return;
	}

	addr = ep->addrX;
	buffer = 2;
	if (ep->currentPacket & 1)
	{
		/* Send to Y buffer */
		addr = ep->addrY;
		buffer = 0x20000;
	}

#ifdef DEBUG_OTG_SUBMIT
info("RecvData: addr=0x%0x size=%d", addr, size);
#endif

	otg_mem_FcRead(ep->fc, addr, ep->dataPtr, size);

	ep->dataPos += size;
	ep->dataPtr = (void *)((unsigned long)ep->dataPtr + size);
	ep->currentPacket++;

	/* Workaround a hardware bug */
	if (!is_last_pkg && ep->dataPos < ep->dataLen)
	{
		otg_fc_WriteReg(ep->fc, otg_ep_FillClear, buffer << ep->index);
	}
}

int
otg_ep_FillCtrl(otg_ep_t *ep)
{
	UINT32 w;
	int size;
	int no_packets;

#ifdef DEBUG_OTG_SUBMIT
info("FillCtrl: ep=%p", ep);
#endif
	if (ep->addrX < 0)
	{
		err("FillCtrl: X buffer for ep %d not allocated", ep->index);
		return -1;
	}

	size = otg_ep_desc_GetMaxPacketSize(ep->desc);
	no_packets = (ep->dataLen - 1) / size + 1;
	w = ep->addrX >> 2;
	if (no_packets > 1)
	{
		if (ep->addrY < 0)
		{
			err("FillCtrl: Y buffer for ep %d not allocated", ep->index);
			return -1;
		}

		w |= (ep->addrY >> 2) << otg_ep_BufferYLoc;
	}
	
	otg_fc_WriteReg(ep->fc, otg_ep_Buffer(ep->index), w);

	if (otg_ep_desc_IsIso1(ep->desc))
	{
		w = size;
		w |= size << 16;
	}
	else
	{
		int lastsize;

		/* 1 packet per buffer. */
		w = no_packets;

		lastsize = ep->dataLen % size;
		if (ep->dataLen && !lastsize)
		{
			lastsize = size;
		}
		w |= lastsize << otg_ep_CtrlLPSLoc;
	}
	
	otg_fc_WriteReg(ep->fc, otg_ep_Ctrl(ep->index), w);

	/* Workaround a hardware bug */
	if (!otg_ep_desc_IsDirectionIn(ep->desc))
	{
		otg_fc_WriteReg(ep->fc, otg_ep_FillClear, 0x20002 << ep->index);
	}

	w = otg_fc_ReadReg(ep->fc, otg_ep_Config(ep->index));

	w |= otg_ep_ConfigNewTransfer;
	otg_fc_WriteReg(ep->fc, otg_ep_Config(ep->index), w);

	ep->packetSize = size;
	ep->currentPacket = 0;

	return 0;
}

int
otg_ep_FillConfig(otg_ep_t *ep)
{
	UINT32 w;
	int size;

	size = otg_ep_desc_GetMaxPacketSize(ep->desc);
	w = size << otg_ep_ConfigMaxPacketSizeLoc;

	if (otg_ep_desc_IsDirectionIn(ep->desc))
	{
		w |= otg_ep_ConfigDirection;
	}

	if (otg_ep_desc_IsIso1(ep->desc))
	{
		w |= otg_ep_ConfigTypeIso1;
	}
	
	otg_fc_WriteReg(ep->fc, otg_ep_Config(ep->index), w);

	return 0;
}
