/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_ep0.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    Function side endpoint 0 object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"
#include "otg_device_request.h"

void
otg_ep0_Init(otg_ep0_t *ep0, otg_fc_t *fc)
{
	ep0->fc = fc;
	ep0->setupMem = otg_fc_GetMemStart(fc) + otg_ep0_SetupBuffer;
	ep0->outMem = otg_fc_GetMemStart(fc) + otg_ep0_OutBuffer;
	ep0->inMem = otg_fc_GetMemStart(fc) + otg_ep0_InBuffer;

	ep0->standardSetup = 0;
	ep0->classSetup = 0;
	ep0->vendorSetup = 0;
}

void
otg_ep0_Reset(otg_ep0_t *ep0)
{
	ep0->stage = otg_ep0_StageIdle;
	ep0->dataLen = 0;
	ep0->dataPtr = 0;

	ep0->remoteWakeupEnable = 0; /* Default to disable */
	ep0->selfPowered = 1; /* This will always be self-powered */
	ep0->halted = 0; /* This will always be self-powered */
	ep0->bHnpEnable = 0;
	ep0->aHnpSupport = 0;
	ep0->aAltHnpSupport = 0;
}

void
otg_ep0_DoSetupOverWritten(otg_ep0_t *ep0)
{
	otg_ep0_DoSetup(ep0);
}

void
otg_ep0_DoSetup(otg_ep0_t *ep0)
{
	UINT8 *setup;

	if (ep0->stage != otg_ep0_StageIdle)
	{
		/* May need to clearup previous transfer */

		ep0->stage = otg_ep0_StageIdle;
		ep0->dataLen = 0;
		ep0->dataPtr = 0;
	}

	if (ep0->stage != otg_ep0_StageIdle)
	{
		/* Something wrong. */
		err("Setup: ep0 stage not idle");

		/* May need to cleanup */
		ep0->stage = otg_ep0_StageIdle;
	}

	setup = ep0->dataBuffer;
	otg_mem_FcRead(ep0->fc, ep0->setupMem, setup, 8);
	otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlSetup);

	ep0->bRequestType = setup[otg_ep0_bRequestType];
	ep0->bRequest = setup[otg_ep0_bRequest];
	ep0->wValue = setup[otg_ep0_wValue];
	ep0->wValue |= setup[otg_ep0_wValue + 1] << 8;
	ep0->wIndex = setup[otg_ep0_wIndex];
	ep0->wIndex |= setup[otg_ep0_wIndex + 1] << 8;
	ep0->wLength = setup[otg_ep0_wLength];
	ep0->wLength |= setup[otg_ep0_wLength + 1] << 8;

	if (otg_ep0_IsClassRequest(ep0->bRequestType) ||
		otg_ep0_IsVendorRequest(ep0->bRequestType))
	{
		otg_ep0_ClassRequest(ep0);
		return;
	}

	switch (ep0->bRequest)
	{
		case otg_ep0_RequestGetStatus:
			otg_ep0_GetStatus(ep0);
			break;
		case otg_ep0_RequestClearFeature:
			otg_ep0_Feature(ep0, 0);
			break;
		case otg_ep0_RequestSetFeature:
			otg_ep0_Feature(ep0, 1);
			break;
		case otg_ep0_RequestSetAddress:
			otg_ep0_SetAddress(ep0);
			break;
		case otg_ep0_RequestGetDescriptor:
			otg_ep0_GetDescriptor(ep0);
			break;
		case otg_ep0_RequestSetDescriptor:
			ep0->dataPos = 0;
			ep0->dataPtr = ep0->data;
			ep0->dataLen = otg_ep0_DataMax;
			ep0->stage = otg_ep0_StageOut;
			break;
		case otg_ep0_RequestGetConfiguration:
			otg_ep0_GetConfiguration(ep0);
			break;
		case otg_ep0_RequestSetConfiguration:
			otg_ep0_SetConfiguration(ep0);
			break;
		case otg_ep0_RequestGetInterface:
			otg_ep0_GetInterface(ep0);
			break;
		case otg_ep0_RequestSetInterface:
			otg_ep0_SetInterface(ep0);
			break;
		case otg_ep0_RequestSyncFrame:
			otg_ep0_SyncFrame(ep0);
			break;
	}
}

void
otg_ep0_DoSetupDone(otg_ep0_t *ep0)
{
	switch (ep0->bRequest)
	{
		case otg_ep0_RequestSetDescriptor:
			otg_ep0_SetDescriptor(ep0);
			break;
		default:
			break;
	}
}

void
otg_ep0_DoOut(otg_ep0_t *ep0)
{
	int size;

	otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlOut);

	if (ep0->stage == otg_ep0_StageIn)
	{
		otg_ep0_SendAck(ep0);
		ep0->stage = otg_ep0_StageIdle;
		return;
	}

	if (ep0->stage != otg_ep0_StageOut)
	{
		err("DoOut: stage is not StageOut");
	}

	size = otg_fc_ReadReg(ep0->fc, otg_ep0_Size) & otg_ep0_SizeOut;
	otg_ep0_RecvData(ep0, size);

	if (ep0->dataPos >= ep0->wLength)
	{
		if (otg_ep0_IsClassRequest(ep0->bRequestType) ||
			otg_ep0_IsVendorRequest(ep0->bRequestType))
		{
			otg_ep0_ClassRequestDone(ep0);
		}
		else
		{
			otg_ep0_DoSetupDone(ep0);
		}
	}
}

void
otg_ep0_DoIn(otg_ep0_t *ep0)
{
	if (ep0->stage == otg_ep0_StageOut)
	{
		otg_ep0_SendAck(ep0);
		ep0->stage = otg_ep0_StageIdle;
		return;
	}

	if (ep0->stage != otg_ep0_StageIn)
	{
		err("DoIn: stage is not StageIn");
	}

	if (ep0->dataPos >= ep0->wLength)
	{
#ifdef DEBUG_OTG_CHIP
info("DoIn: Done. dataPos=%d wLength=%d", ep0->dataPos, ep0->wLength);
#endif
		return;
	}

	otg_ep0_SendData(ep0);
}

void
otg_ep0_SendData(otg_ep0_t *ep0)
{
	int len;
	UINT8 *buf;
#ifdef DEBUG_OTG_CHIP
	UINT32 w;
#endif

	/* At this point:
	 *  ep0->dataPtr points to the data buffer.
	 *  ep0->wLength is the number of bytes requested by host
	 *  ep0->dataPos is the number of bytes already sent
	 *  ep0->dataLen is the number of bytes in buffer ep0->dataPtr
	 */
#ifdef DEBUG_OTG_CHIP
info("SendData: enter stage=%d dataLen=%d dataPos=%d dataPtr=%p",
	ep0->stage, ep0->dataLen, ep0->dataPos, ep0->dataPtr);
#endif
	if (ep0->dataPos >= ep0->wLength || ep0->dataLen <= 0)
	{
		/* Done. Send zero length packet */
		otg_fc_WriteReg(ep0->fc, otg_ep0_Size, 0);
		otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlIn);
		return;
	}

	if (!ep0->dataPtr)
	{
		err("SendData: data pointer is null");
#ifdef DEBUG_OTG_CHIP
info("SendData: wLength=0x%0x wIndex=0x%0x wValue=0x%0x",
	ep0->wLength, ep0->wIndex, ep0->wValue);
#endif
		otg_fc_WriteReg(ep0->fc, otg_ep0_Size, 0);
		otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlIn);
		return;
	}

	len = ep0->wLength - ep0->dataPos;
	if (len > otg_ep0_BufferSize)
	{
		len = otg_ep0_BufferSize;
	}

	if (len > ep0->dataLen)
	{
		len = ep0->dataLen;
	}

#ifdef DEBUG_OTG_CHIP
	w = otg_fc_ReadReg(ep0->fc, otg_ep0_Ctrl);
	if (w & otg_ep0_CtrlIn)
	{
		info("SendData: Overwrite previous IN data");
	}
#endif
	buf = (UINT8*)ep0->dataPtr;
	otg_mem_FcWrite(ep0->fc, ep0->inMem, &buf[ep0->dataPos], len);
	otg_fc_WriteReg(ep0->fc, otg_ep0_Size, len << otg_ep0_SizeInLoc);
	otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlIn);

	ep0->dataPos += len;
	ep0->dataLen -= len;
#ifdef DEBUG_OTG_CHIP
info("SendData: stage=%d dataLen=%d dataPos=%d dataPtr=%p",
	ep0->stage, ep0->dataLen, ep0->dataPos, ep0->dataPtr);
#endif
}

void
otg_ep0_RecvData(otg_ep0_t *ep0, int len)
{
	UINT8 *buf;

	/* At this point:
	 *  ep0->dataPtr points to the data buffer.
	 *  ep0->wLength is the number of bytes requested by host
	 *  ep0->dataPos is the number of bytes already received
	 *  ep0->dataLen is the number of bytes in buffer ep0->dataPtr
	 */
#ifdef DEBUG_OTG_CHIP
info("RecvData: enter stage=%d dataLen=%d dataPos=%d dataPtr=%p",
	ep0->stage, ep0->dataLen, ep0->dataPos, ep0->dataPtr);
#endif
	if (ep0->dataPos >= ep0->wLength || ep0->dataLen <= 0)
	{
		/* Done */
		return;
	}

	if (!len)
	{
		/* Zero length packet received. Keep on waiting. */
		return;
	}

	otg_mem_FcRead(ep0->fc, ep0->outMem, ep0->dataBuffer, len);

	buf = (UINT8*)ep0->dataPtr;
	if (len <= ep0->dataLen)
	{
		memcpy(&buf[ep0->dataPos], ep0->dataBuffer, len);
	}
	else
	{
		memcpy(&buf[ep0->dataPos], ep0->dataBuffer, ep0->dataLen);
	}

	ep0->dataPos += len;
	ep0->dataLen -= len;
#ifdef DEBUG_OTG_CHIP
info("RecvData: stage=%d dataLen=%d dataPos=%d dataPtr=%p",
	ep0->stage, ep0->dataLen, ep0->dataPos, ep0->dataPtr);
#endif
}

void
otg_ep0_Feature(otg_ep0_t *ep0, int set)
{
	if (otg_ep0_IsDeviceRequest(ep0->bRequestType) && set)
	{
		if (ep0->wValue == otg_ep0_FEATURE_B_HNP_ENABLE)
		{
#ifdef OTG_HNP_DEBUG
info("Feature: set bHnpEnable");
#endif
			ep0->bHnpEnable = 1;
			otg_fc_BHnpEnabled(ep0->fc);
			otg_ep0_EndOut(ep0);
			return;
		}
		else if (ep0->wValue == otg_ep0_FEATURE_A_HNP_SUPPORT)
		{
#ifdef OTG_HNP_DEBUG
info("Feature: set aHnpSupport");
#endif
			ep0->aHnpSupport = 1;
			otg_ep0_EndOut(ep0);
			return;
		}
		else if (ep0->wValue == otg_ep0_FEATURE_A_ALT_HNP_SUPPORT)
		{
#ifdef OTG_HNP_DEBUG
info("Feature: set aAltHnpSupport");
#endif
			ep0->aAltHnpSupport = 1;
			otg_ep0_EndOut(ep0);
			return;
		}
	}

	switch (otg_fc_GetState(ep0->fc))
	{
		case otg_fc_StateDefault:
			err("Feature: in Default state, stall");
			otg_ep0_SendStall(ep0);
			return;
		case otg_fc_StateAddress:
			if (!ep0->wValue)
			{
				err("Feature: non-zero ep %d, stall", ep0->wValue);
				otg_ep0_SendStall(ep0);
				return;
			}
			break;
		case otg_fc_StateConfigured:
			break;
	}

	if (otg_ep0_IsDeviceRequest(ep0->bRequestType))
	{
		if (ep0->wValue == otg_ep0_FEATURE_REMOTE_WAKEUP)
		{
			/* Clear remote wakeup */
			ep0->remoteWakeupEnable = set;
		}
		else if (set)
		{
			otg_ep0_SendStall(ep0);
			return;
		}
	}
	else if (otg_ep0_IsEndpointRequest(ep0->bRequestType))
	{
		if (ep0->wValue == otg_ep0_FEATURE_ENDPOINT_HALT)
		{
			if (ep0->wIndex == 0)
			{
				ep0->halted = set;
			}
			else
			{
				otg_ep_t *ep;

#ifdef DEBUG_OTG_CHIP
info("Feature: wIndex=0x%0x", ep0->wIndex);
#endif
				ep = otg_fc_GetEp(ep0->fc, ep0->wIndex);
				if (ep)
				{
					otg_ep_SetHalted(ep, set);
				}
				else if (set)
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
		}
		else if (set)
		{
			otg_ep0_SendStall(ep0);
			return;
		}
	}

	otg_ep0_CallStandardSetup(ep0);

	otg_ep0_EndOut(ep0);
}

void
otg_ep0_ClassRequest(otg_ep0_t *ep0)
{
	otg_device_request_t request;

#ifdef DEBUG_OTG_CHIP
info("ClassRequest: type=0x%0x", ep0->bRequestType);
#endif
	if (otg_ep0_IsDeviceToHostRequest(ep0->bRequestType))
	{
#ifdef DEBUG_OTG_CHIP
info("ClassRequest: device to host");
#endif
		request.bmRequestType = ep0->bRequestType;
		request.bRequest = ep0->bRequest;
		request.wValue = ep0->wValue;
		request.wIndex = ep0->wIndex;
		request.wLength = ep0->wLength;
		request.data = ep0->data;

		if (otg_ep0_IsClassRequest(ep0->bRequestType))
		{
#ifdef DEBUG_OTG_CHIP
info("ClassRequest: classSetup=%p", ep0->classSetup);
#endif
			if (ep0->classSetup)
			{
				if (ep0->classSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
			else
			{
				request.wLength = 0;
			}
		}
		else if (otg_ep0_IsVendorRequest(ep0->bRequestType))
		{
			if (ep0->vendorSetup)
			{
				if (ep0->vendorSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
			else
			{
				request.wLength = 0;
			}
		}
		else
		{
			if (ep0->standardSetup)
			{
				if (ep0->standardSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
			else
			{
				request.wLength = 0;
			}
		}

		ep0->dataPtr = request.data;
		ep0->wLength = request.wLength;
		ep0->dataLen = ep0->wLength;
		ep0->dataPos = 0;

		otg_ep0_SendData(ep0);
		ep0->stage = otg_ep0_StageIn;
	}
	else
	{
		if (ep0->wLength > otg_ep0_DataMax)
		{
			err("ClassRequest: too many data %d", ep0->wLength);
			ep0->wLength = otg_ep0_DataMax;
		}

		ep0->stage = otg_ep0_StageOut;
		if (!ep0->wLength)
		{
			/* Zero size package, there will be no OUT packet */
			otg_ep0_ClassRequestDone(ep0);
		}

		ep0->dataPtr = ep0->data;
		ep0->dataLen = ep0->wLength;
		ep0->dataPos = 0;
	}
}

void
otg_ep0_ClassRequestDone(otg_ep0_t *ep0)
{
	otg_device_request_t request;

	if (!otg_ep0_IsDeviceToHostRequest(ep0->bRequestType))
	{
		request.bmRequestType = ep0->bRequestType;
		request.bRequest = ep0->bRequest;
		request.wValue = ep0->wValue;
		request.wIndex = ep0->wIndex;
		request.wLength = ep0->wLength;
		request.data = ep0->data;

		if (otg_ep0_IsClassRequest(ep0->bRequestType))
		{
			if (ep0->classSetup)
			{
				if (ep0->classSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
		}
		else if (otg_ep0_IsVendorRequest(ep0->bRequestType))
		{
			if (ep0->vendorSetup)
			{
				if (ep0->vendorSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
		}
		else
		{
			if (ep0->standardSetup)
			{
				if (ep0->standardSetup(&request, ep0->privateData))
				{
					otg_ep0_SendStall(ep0);
					return;
				}
			}
		}

		otg_fc_WriteReg(ep0->fc, otg_ep0_Size, 0);
		otg_fc_WriteReg(ep0->fc, otg_ep0_Ctrl, otg_ep0_CtrlIn);
#ifdef DEBUG_OTG_CHIP
info("ClassRequestDone: dataPos=%d wLength=%d", ep0->dataPos, ep0->wLength);
#endif
	}
}

void
otg_ep0_CallStandardSetup(otg_ep0_t *ep0)
{
	if (ep0->standardSetup)
	{
		otg_device_request_t request;

		request.bmRequestType = ep0->bRequestType;
		request.bRequest = ep0->bRequest;
		request.wValue = ep0->wValue;
		request.wIndex = ep0->wIndex;
		request.wLength = ep0->wLength;
		request.data = ep0->data;

		ep0->standardSetup(&request, ep0->privateData);

		if (otg_ep0_IsDeviceToHostRequest(ep0->bRequestType))
		{
			ep0->dataPtr = request.data;
			ep0->wLength = request.wLength;
		}
	}
}
