/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_etd.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    ETD object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

void
otg_etd_SetPipe(otg_etd_t *etd, stack_urb_t *urb)
{
	stack_pipe_t pipe = stack_urb_GetPipe(urb);

	etd->urb = urb;

	etd->endpoint |= stack_pipe_GetAddressEndpoint(pipe); /* EP and device number */

	if (stack_pipe_IsIn(pipe))
	{
		etd->endpoint |= otg_etd_Direction_IN_c;
	}
	else
	{
		etd->endpoint |= otg_etd_Direction_OUT_c;
	}

	if (urb->dev->speed == USB_SPEED_LOW)
	{
		/* no need to set for full speed */
		etd->endpoint |= otg_etd_Speed_Low_c;
		etd->endpoint |= 8 << otg_etd_MaxPacketSizeLoc_c;
	}
	else
	{
		etd->endpoint |= stack_urb_GetMaxPacketSize(urb, pipe)
							<< otg_etd_MaxPacketSizeLoc_c;
	}
#ifdef DEBUG_OTG_SUBMIT
info("SetPipe: endpoint=0x%x", etd->endpoint);
#endif
}

int
otg_etd_Fill(otg_etd_t *etd, otg_hc_t *hc, stack_urb_t *urb)
{
	int rc;
	int bustime;

#ifdef DEBUG_OTG_SUBMIT
info("Fill: pipe=0x%x", stack_urb_GetPipe(urb));
#endif
	rc = 0;
	switch (stack_urb_GetPipeType(urb))
	{
		case stack_urb_PipeTypeCtrl:
			otg_etd_SetCtrlPipe(etd, urb);
			rc = otg_td_ctrl_Fill(etd, hc);
			break;
		case stack_urb_PipeTypeBulk:
			otg_etd_SetBulkPipe(etd, urb);
			rc = otg_td_bulk_Fill(etd, hc);
			break;
		case stack_urb_PipeTypeIntr:
			otg_etd_SetIntrPipe(etd, urb);
			if (!stack_urb_HasBandwidth(urb))
			{
				bustime = stack_urb_CheckBandwidth(urb);
				if (bustime < 0)
				{
					rc = bustime;
				}
				else
				{
					rc = otg_td_intr_Fill(etd, hc);
					if (!rc)
					{
						stack_urb_ClaimBandwidth(urb, bustime, 0);
					}
				}
			}
			else
			{
				rc = otg_td_intr_Fill(etd, hc);
			}
			break;
		case stack_urb_PipeTypeIso:
			otg_etd_SetIsoPipe(etd, urb);
			if (urb->number_of_packets <= 0)
			{
				rc = -EINVAL;
				break;
			}

			if (!stack_urb_HasBandwidth(urb))
			{
				bustime = stack_urb_CheckBandwidth(urb);
				if (bustime < 0)
				{
					rc = bustime;
				}
				else
				{
					rc = otg_td_iso_Fill(etd, hc);
					if (!rc)
					{
						stack_urb_ClaimBandwidth(urb, bustime, 1);
					}
				}
			}
			else
			{
				rc = otg_td_iso_Fill(etd, hc);
			}
			break;
	}

#ifdef DEBUG_OTG_ETD
otg_etd_PrintEtd(etd);
info("Fill return with rc=%d", rc);
#endif
	return rc;
}

int
otg_etd_DoDoneEtd(otg_etd_t *etd, otg_hc_etd_t *hc_etd)
{
	int rc;

	/* Process finished packets */
	if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
	{
		rc = otg_td_iso_DoDoneEtd(etd);
	}
	else
	{
		rc = otg_td_ctrl_DoDoneEtd(etd);
	}

	return rc;
}

int
otg_etd_CopyHcInfo(otg_etd_t *etd, otg_etd_t *from)
{
	etd->etdIndex = from->etdIndex;
	etd->buffer = from->buffer;
	etd->bufferSize = from->bufferSize;
	etd->addrX = from->addrX;
	etd->addrY = from->addrY;
	if (etd->addrY < 0)
	{
		if (otg_etd_NeedYBuffer(etd))
		{
			return otg_hc_Allocate(etd->hc, etd, etd->bufferSize);
		}
	}

	return 0;
}

void
otg_etd_Delete(otg_etd_t *etd)
{
#ifdef DEBUG_SOME_OTG
	info("otg_etd_Delete etd=%p 0x%0x 0x%0x", etd,
		otg_etd_GetBuffer(etd, 1), otg_etd_GetBuffer(etd, 0));
#endif
#ifdef OTG_ISO_SHARE_MEMORY
	if (otg_etd_IsSofLinked(etd))
	{
		otg_DisableSofIntr(otg_hc_GetOtg(etd->hc));

		OS_ListDelete(&etd->sofList);
		OS_ListInit(&etd->sofList);
	}
#endif
	OS_ListDelete(&etd->pipeEtdList);
	OS_ListDelete(&etd->formatEtdList);

	OS_Free(etd);
}

otg_etd_t *
otg_etd_GetNext(otg_etd_t *etd)
{
	otg_etd_t *first;
	otg_etd_t *next;

	first = otg_hc_GetFirstInPipe(etd->hc, etd);
	if (!first)
	{
		return 0;
	}

	next = OS_ListEntry(OS_ListNext(&etd->pipeEtdList), otg_etd_t, pipeEtdList);
	if (next == first)
	{
		return 0;
	}

	return next;
}

otg_etd_t *
otg_etd_GetPrev(otg_etd_t *etd)
{
	otg_etd_t *first;
	
	first = otg_hc_GetFirstInPipe(etd->hc, etd);
	if (!first || first == etd)
	{
		return 0;
	}

	return OS_ListEntry(OS_ListPrev(&etd->pipeEtdList), otg_etd_t, pipeEtdList);
}

#ifdef DEBUG_OTG_ETD
void
otg_etd_PrintEtd(otg_etd_t *etd)
{
	info("Etd %p dump", etd);
	info("---dword0=0x%0x dword1=0x%0x dword2=0x%0x dword3=0x%0x",
		etd->endpoint, etd->buffer, etd->td.ctrl.ctrl, etd->td.ctrl.info);
	if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
	{
		info("---dword4=0x%0x dword5=0x%0x dword6=0x%0x",
			etd->td.iso.data[1],etd->td.iso.data[2], etd->td.iso.data[3]);
	}

	info("---etdIndex=%d  addrX=0x%0x addrY=0x%0x unlinked=%d inHc=%d",
		etd->etdIndex, etd->addrX, etd->addrY, etd->unlinked, etd->inHc);

	info("---completed=0x%x lastCompleted=0x%0x bufferSize=0x%0x",
		etd->completed, etd->lastCompleted, etd->bufferSize);

	info("---dataLength=0x%0x data=%p", etd->dataLength, etd->data);
	
	info("---pipeEtdList=%p next=%p    formatEtdList=%p next=%p",
		&etd->pipeEtdList, etd->pipeEtdList.next,
		&etd->formatEtdList, OS_ListNext(&etd->formatEtdList));
}
#endif

#ifdef DEBUG_OTG_SUBMIT
void
otg_etd_PrintErrorCode(stack_urb_t *urb)
{
	char* msg;

	switch (stack_urb_GetStatus(urb))
	{
		case otg_NO_ERROR:
			msg = "otg_NO_ERROR";
			break;
		case -EILSEQ:
			msg = "EILSEQ";
			break;
		case -EPROTO:
			msg = "EPROTO";
			break;
		case -EPIPE:
			msg = "EPIPE";
			break;
		case -ETIMEDOUT:
			msg = "ETIMEDOUT";
			break;
		case -EOVERFLOW:
			msg = "EOVERFLOW";
			break;
		case -EREMOTEIO:
			msg = "EREMOTEIO";
			break;
		case -ECOMM:
			msg = "ECOMM";
			break;
		case -ENOSR:
			msg = "ENOSR";
			break;
		case -EXDEV:
			msg = "EXDEV";
			break;
		case otg_TRANSACTION_NOT_DONE:
			msg = "otg_TRANSACTION_NOT_DONE";
			break;
		case -EINPROGRESS:
			msg = "EINPROGRESS";
			break;
		case -ENOENT:
			msg = "ENOENT";
			break;
		case -ENODEV:
			msg = "ENODEV";
			break;
		case -EINVAL:
			msg = "EINVAL";
			break;
		case -ENXIO:
			msg = "ENXIO";
			break;
		case -EMSGSIZE:
			msg = "EMSGSIZE";
			break;
		default:
			msg = "Unknown: check linux/include/asm-i386/errno.h";
			break;
	}

	info("PrintErrorCode: %d:%s", urb->status, msg);
}

#endif
