/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_fc.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 function controller object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

static otg_fc_t *functionController = 0;

otg_fc_t *
otg_fc_New(void)
{
	otg_t *otg;

#ifndef OTG_HOST_ONLY
	if (!functionController)
	{
		otg = otg_New(NULL); /*TODO: Check that...*/
		if (!otg)
		{
			return 0;
		}
		else
		{
			functionController = otg_GetFunction(otg);
		}
	}

	return functionController;
#else
	return 0;
#endif
}

/* System startup should call otg_fc_Init.
 * An application should call otg_fc_SetDeviceDesc, otg_fc_SetConfigDescriptor
 * and otg_fc_SetStrDesc before submit transfer request.
 */
void
otg_fc_Init(otg_fc_t *fc, otg_t *otg, int start, int size)
{
	int ep0BufSize;

	fc->otg = otg;
	fc->started = 0;

#ifndef OTG_HOST_ONLY
	ep0BufSize = otg_ep0_GetChipBufferSize();
	otg_mem_Init(&fc->mem, start + ep0BufSize, size - ep0BufSize);
	fc->start = start;
	fc->selectedConfig = 0;
	fc->numConfigs = 0;

	otg_device_desc_Init(&fc->deviceDesc);
	otg_config_desc_Init(fc->configDesc);
	otg_str_desc_Init(&fc->strDesc);

	otg_fc_intr_Init(&fc->fcIntr, fc);
	otg_ep_intr_Init(&fc->epIntr, fc);
	otg_ep0_Init(&fc->ep0, fc);
	otg_ep_Init(fc->ep, fc);

	fc->timerSet = 0;
#endif
}

void
otg_fc_Reset(otg_fc_t *fc)
{
	unsigned long flags;

#ifndef OTG_HOST_ONLY
	otg_LockChipIntr(fc->otg, flags);
	otg_fc_ResetUnlocked(fc);
	otg_UnlockChipIntr(fc->otg, flags);
#endif
}

void
otg_fc_ResetUnlocked(otg_fc_t *fc)
{
	UINT32 w;
	int i;

#ifndef OTG_HOST_ONLY
	otg_fc_WriteReg(fc, otg_fc_UnlockReg, otg_fc_UnlockRegUnlock);

	w = otg_fc_ReadReg(fc, otg_fc_Command);
	for(i = 0; i < otg_fc_MaxEP; i++)
	{
		w |= otg_fc_CommandToggleReset1 << i;
	}

	otg_fc_WriteReg(fc, otg_fc_Command, w);

	otg_ClearBHnpEnable(fc->otg);

	fc->state = otg_fc_StateDefault;

	for(i = 0; i < otg_fc_MaxEP; i++)
	{
		otg_ep_Release(&fc->ep[i]);
	}
	fc->selectedConfig = 0;

	otg_fc_intr_Reset(&fc->fcIntr);
	otg_ep_intr_Reset(&fc->epIntr);
	otg_ep0_Reset(&fc->ep0);
	otg_ep_Reset(fc->ep);

#endif
}

void otg_fc_Exit(otg_fc_t *fc)
{
	if (fc->timerSet)
	{
		OS_StopTimer(&fc->timer);
		fc->timerSet = 0;
	}
}

void
otg_fc_DoIntr(otg_fc_t *fc)
{
	otg_fc_intr_t *intr;
	otg_ep_intr_t *ep_intr;
	int ep;
	UINT32 w;

#ifdef DEBUG_OTG_CHIP
info("DoIntr: enter");
#endif
	intr = &fc->fcIntr;
	otg_fc_intr_Check(intr, otg_fc_intr_All);
	while (otg_fc_intr_IsTriggered(intr, otg_fc_intr_All))
	{
#ifdef DEBUG_OTG_CHIP
		info("DoIntr: intr->status=0x%0x", intr->status);
#endif
		if (otg_fc_intr_IsTriggered(intr, otg_fc_intr_BusReset))
		{
#ifdef DEBUG_OTG_CHIP
			info("Get BusReset interrupt");
#endif
			otg_fc_intr_ClearIntr(intr, otg_fc_intr_BusReset);

			if (otg_fc_intr_Check(&fc->fcIntr, otg_fc_intr_BusReset))
			{
				/* Remove this ClearIntr will cause system dead */
				otg_fc_intr_ClearIntr(intr, otg_fc_intr_BusReset);
				otg_fc_Disable(fc);
				if (!fc->timerSet)
				{
					otg_fc_StartTimer(fc);
				}
			}
			else
			{
				otg_fc_ResetUnlocked(fc);
			}
		}
		else if (otg_fc_intr_IsTriggered(intr, otg_fc_intr_SOF))
		{
#ifdef DEBUG_OTG_CHIP
			info("Get SOF interrupt");
#endif
			otg_fc_intr_ClearIntr(intr, otg_fc_intr_SOF);
		}
		else if (otg_fc_intr_IsTriggered(intr, otg_fc_intr_SOFLost))
		{
#ifdef DEBUG_OTG_CHIP
			info("Get SOFLost interrupt");
#endif
			otg_fc_intr_ClearIntr(intr, otg_fc_intr_SOFLost);
		}
		else if (otg_fc_intr_IsTriggered(intr, otg_fc_intr_Suspend))
		{
#ifdef DEBUG_OTG_CHIP
			info("Get Suspend interrupt");
#endif
			otg_fc_intr_ClearIntr(intr, otg_fc_intr_Suspend);
			otg_fc_WriteReg(fc, otg_fc_Command, otg_fc_CommandSuspend);
		}
		else if (otg_fc_intr_IsTriggered(intr, otg_fc_intr_Resume))
		{
#ifdef DEBUG_OTG_CHIP
			info("Get Resume interrupt");
#endif
			otg_fc_intr_ClearIntr(intr, otg_fc_intr_Resume);
			otg_fc_WriteReg(fc, otg_fc_UnlockReg, otg_fc_UnlockRegUnlock);
		}
		else
		{
			err("DoIntr: Unknown interrupt");
		}
	}

	ep_intr = &fc->epIntr;
	if (!otg_device_desc_IsSet(&fc->deviceDesc) || (fc->numConfigs <= 0))
	{
#ifdef DEBUG_OTG_CHIP
info("DoIntr: device not ready");
#endif
		/* Descriptor not set. Device not ready. */
		otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_All);
		return;
	}

	if (otg_ep_intr_Check(ep_intr, otg_ep_intr_All))
	{
#ifdef DEBUG_OTG_CHIP
info("DoIntr: do ep0 interrupts");
#endif
		if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_SetupOverWritten))
		{
			otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_SetupOverWritten);
			otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_Setup);
			otg_ep0_DoSetupOverWritten(&fc->ep0);
		}

		if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_Setup))
		{
			otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_Setup);
			otg_ep0_DoSetup(&fc->ep0);
		}

		if (otg_ep0_IsStageIn(&fc->ep0))
		{
			if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_EP0In))
			{
				otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_EP0In);
				otg_ep0_DoIn(&fc->ep0);
			}

			if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_EP0Out))
			{
				otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_EP0Out);
				otg_ep0_DoOut(&fc->ep0);
			}
		}
		else
		{
			if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_EP0Out))
			{
				otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_EP0Out);
				otg_ep0_DoOut(&fc->ep0);
			}

			if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_EP0In))
			{
				otg_ep_intr_ClearIntr(ep_intr, otg_ep_intr_EP0In);
				otg_ep0_DoIn(&fc->ep0);
			}
		}

		if (otg_ep_intr_IsTriggered(ep_intr, otg_ep_intr_AllEPn))
		{
			UINT32 filled;

#ifdef DEBUG_OTG_CHIP
info("DoIntr: do ep interrupts");
#endif
			w = otg_ep_intr_EP1;
			filled = otg_fc_ReadReg(fc, otg_ep_FillStatus);
			for (ep = 0; ep < otg_fc_MaxEP; ep++)
			{
				if (otg_ep_intr_IsTriggered(ep_intr, w))
				{
					otg_ep_intr_ClearIntr(ep_intr, w);
					otg_ep_DoIntr(&fc->ep[ep], ep_intr, filled);
				}

				w <<= 1;
			}
		}
	}
}

int
otg_fc_SetConfig(otg_fc_t *fc, int config)
{
	int i;

	if (config)
	{
		if (config < 1)
		{
			return -1;
		}

		fc->selectedConfig = otg_fc_FindConfig(fc, config);
		if (!fc->selectedConfig)
		{
			return -1;
		}

		/* Release previously configured endpoints */
		for(i = 0; i < otg_fc_MaxEP; i++)
		{
			otg_ep_Release(&fc->ep[i]);
		}

		return otg_config_desc_Set(fc->selectedConfig, fc);
	}
	else
	{
		fc->selectedConfig = 0;
	}

	return 0;
}

void
otg_fc_Stall(otg_fc_t *fc, int ep_num)
{
	otg_ep_t *ep;

	ep = otg_fc_GetEp(fc, ep_num);
	if (!ep)
	{
		return;
	}

	otg_ep_SetHalted(ep, 1);
}

void
otg_fc_ClearStall(otg_fc_t *fc, int ep_num)
{
	otg_ep_t *ep;

	ep = otg_fc_GetEp(fc, ep_num);
	if (!ep)
	{
		return;
	}

	otg_ep_SetHalted(ep, 0);
}

int
otg_fc_Submit(otg_fc_t *fc, otg_urb_t *urb)
{
	otg_ep_t *ep;
	int rc;
	unsigned long flags;

#ifdef DEBUG_OTG_SUBMIT
UINT8 *bp;
bp = urb->transfer_buffer;
info("Submit: urb=%p ep=0x%0x length=%d", urb, urb->ep, urb->transfer_length);
info("Submit: 0x%0x 0x%0x 0x%0x 0x%0x 0x%0x 0x%0x 0x%0x 0x%0x",
	*bp++, *bp++, *bp++, *bp++, *bp++, *bp++, *bp++, *bp++);
#endif
	ep = otg_fc_GetEp(fc, urb->ep);
	if (!ep)
	{
		return -ENODEV;
	}

#ifdef DEBUG_OTG_SUBMIT
info("Submit: find ep=%p", ep);
#endif

	OS_DisableInterrupts(flags);

	rc = otg_ep_Submit(ep, urb);
	OS_EnableInterrupts(flags);

	return rc;
}

int
otg_fc_SetEp(otg_fc_t *fc, int ep_addr, otg_ep_desc_t *desc)
{
	return otg_ep_Set(&fc->ep[(ep_addr & 0xF) - 1], desc);
}

int
otg_fc_SetDeviceDesc(otg_fc_t *fc, UINT8 *desc, int len)
{
	return otg_device_desc_SetDesc(&fc->deviceDesc, desc, len);
}

int
otg_fc_SetConfigDesc(otg_fc_t *fc, UINT8 *desc, int len)
{
	UINT8 *buf;
	int rc;
	int p;
	int count;

	buf = OS_Malloc(len);
	if (!buf)
	{
		return -ENOMEM;
	}

	memcpy(buf, desc, len);
	fc->numConfigs = 0;

	p = 0;
	count = 0;
	while (p < len)
	{
		if (count >= otg_fc_MaxConfig)
		{
			err("SetConfigDesc: too many configurations");
			OS_Free(buf);
			return -EINVAL;
		}

		rc = otg_config_desc_SetDesc(&fc->configDesc[count], &buf[p], len - p);

		if (rc < 0)
		{
			OS_Free(buf);
			return rc;
		}

		count++;
		p += rc;
	}

	fc->numConfigs = count;
	fc->otherDesc = buf;
	fc->selectedConfig = fc->configDesc;

	otg_ep0_Cleanup(&fc->ep0);

	return 0;
}

int
otg_fc_SetStrDesc(otg_fc_t *fc, UINT8 *desc, int len)
{
	return otg_str_desc_SetDesc(&fc->strDesc, desc, len);
}

int
otg_fc_SetCallBack
(
	otg_fc_t *fc,
	otg_device_request_CallBack_t standardReq,
	otg_device_request_CallBack_t classReq,
	otg_device_request_CallBack_t vendorReq
)
{
	return otg_ep0_SetCallBack(&fc->ep0, standardReq, classReq, vendorReq);
}

int
otg_fc_GetConfiguration(otg_fc_t *fc)
{
	if (!fc->selectedConfig)
	{
		return -1;
	}

	return otg_config_desc_GetConfiguration(fc->selectedConfig);
}

int
otg_fc_WaitStart(otg_fc_t *fc)
{
	int wait;

	wait = 0;
	while (otg_ReadChipReg(fc->otg, otg_IOReady_c) & otg_IOReady_NotReady)
	{
#ifdef SIM_OTG
		/* This code is for debugging */
		OS_Wait(1);
#endif

		wait++;
		if (wait > 10000)
		{
			err("WaitStart: hardware not responding");
			return -1;
		}
	}

	return 0;
}

int
otg_fc_RequestBus(otg_fc_t *fc)
{
#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	if (!otg_ep0_GetAHnpSupport(&fc->ep0))
	{
		if (otg_ep0_GetAAltHnpSupport(&fc->ep0))
		{
			return otg_fc_ErrorAltHnpSupported;
		}

		return otg_fc_ErrorHnpNotSupported;
	}

	otg_device_RequestBus(otg_GetDevice(fc->otg));
#endif
#else
	return otg_fc_ErrorFunctionOnly;
#endif

	return 0;
}

void
otg_fc_ReleaseBus(otg_fc_t *fc)
{
#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	otg_device_ReleaseBus(otg_GetDevice(fc->otg));
#else
#endif
#endif
}

void
otg_fc_StartTimer(otg_fc_t *fc)
{
	OS_DeclareTimer(*timer);

#ifdef DEBUG_OTG_CHIP
info("StartTimer: ");
#endif
	timer = &fc->timer;
	OS_StartTimer(timer, otg_fc_Timeout, fc, 5);
	fc->timerSet = 1;
}

void
otg_fc_Timeout(unsigned long data)
{
	unsigned long flags;

	otg_fc_t *fc = (otg_fc_t *)data;

	otg_LockChipIntr(fc->otg, flags);

#ifdef DEBUG_OTG_CHIP
info("Timeout: ");
#endif
	fc->timerSet = 0;
	otg_fc_intr_ClearIntr(&fc->fcIntr, otg_fc_intr_BusReset);
	if (otg_fc_intr_Check(&fc->fcIntr, otg_fc_intr_BusReset))
	{
		otg_fc_intr_ClearIntr(&fc->fcIntr, otg_fc_intr_BusReset);
		otg_fc_StartTimer(fc);
	}
	else
	{
		otg_fc_ResetUnlocked(fc);
		otg_fc_Enable(fc);
	}

	otg_UnlockChipIntr(fc->otg, flags);
}

int
otg_fc_Enable(otg_fc_t *fc)
{
	UINT32 w;

#ifdef DEBUG_OTG_CHIP
info("Enable: enable fc interrupt");
#endif
	if (otg_device_desc_IsSet(&fc->deviceDesc) && (fc->numConfigs > 0))
	{
#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	if (otg_device_IsDevice(otg_GetDevice(fc->otg)))
	{
#endif
#endif
		w = otg_fc_intr_BusReset | otg_fc_intr_Resume;
		otg_fc_intr_EnableIntr(&fc->fcIntr, w);
		otg_ep_intr_EnableIntr(&fc->epIntr, otg_ep_intr_AllEPn);
		otg_intr_EnableFunctionController(otg_GetIntr(fc->otg));

#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	}
#endif
#endif
		return 0;
	}

	return -1;
}

int
otg_fc_Disable(otg_fc_t *fc)
{
#ifdef DEBUG_OTG_CHIP
info("Disable: ");
#endif
	if (fc->timerSet)
	{
#ifdef DEBUG_OTG_CHIP
info("Disable: disable timer");
#endif
		fc->timerSet = 0;
		OS_StopTimer(&fc->timer);
	}

	otg_fc_intr_DisableIntr(&fc->fcIntr, otg_fc_intr_All);
	otg_ep_intr_DisableIntr(&fc->epIntr, otg_ep_intr_AllEPn);
	otg_intr_DisableFunctionController(otg_GetIntr(fc->otg));

	return 0;
}

int
otg_fc_Start(otg_fc_t *fc)
{
	unsigned long flags;

#ifndef OTG_HOST_ONLY
#ifdef DEBUG_OTG_CHIP
	info("Start: started=%d", fc->started);
#endif

	if (fc->started)
	{
		return 0;
	}

#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	otg_LockChipIntr(fc->otg, flags);
	otg_SetHardwareHNP(fc->otg);
	otg_UnlockChipIntr(fc->otg, flags);
#endif
#endif

	otg_LockChipIntr(fc->otg, flags);
	otg_fc_Enable(fc);
	fc->started = 1;
	otg_UnlockChipIntr(fc->otg, flags);
#endif

	return 0;
}

int
otg_fc_Stop(otg_fc_t *fc)
{
	unsigned long flags;

#ifndef OTG_HOST_ONLY
#ifdef DEBUG_OTG_CHIP
	info("Stop: started=%d", fc->started);
#endif

	if (!fc->started)
	{
		return 0;
	}

#ifndef OTG_FUNCTION_ONLY
#ifndef OTG_HOST_ONLY
	otg_LockChipIntr(fc->otg, flags);

	otg_SetSoftwareHNP(fc->otg);
	otg_device_Disconnect(otg_GetDevice(fc->otg));

	otg_UnlockChipIntr(fc->otg, flags);
#endif
#endif

	otg_LockChipIntr(fc->otg, flags);
	otg_fc_Disable(fc);
	fc->started = 0;
	otg_UnlockChipIntr(fc->otg, flags);
#endif

	return 0;
}

void otg_fc_Delete(otg_fc_t *fc)
{
	unsigned long flags;

#ifdef DEBUG_OTG_CHIP
info("Delete: ");
#endif

	otg_LockChipIntr(fc->otg, flags);

	otg_Delete(fc->otg);

	functionController = 0;

	otg_UnlockChipIntr(fc->otg, flags);
}

void
otg_fc_BHnpEnabled(otg_fc_t *fc)
{
#ifdef OTG_HNP_DEBUG
info("BHnpEnabled: ");
#endif
	otg_BHnpEnabled(fc->otg);
}

void
otg_fc_Wakeup(otg_fc_t *fc)
{
	otg_fc_WriteReg(fc, otg_fc_Command, otg_fc_CommandRemoteWakeup);
}
