/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_fd.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    Function driver object.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"
#include "otg_fd.h"

static otg_fc_t *fc = 0;
static otg_fdevice_t **devices = 0;
static otg_fdevice_t *currentDevice = 0;

void
otg_fd_Stall(int ep)
{
	if (!fc)
	{
		return;
	}

	otg_fc_Stall(fc, ep);
}

void
otg_fd_ClearStall(int ep)
{
	if (!fc)
	{
		return;
	}

	otg_fc_ClearStall(fc, ep);
}

int
otg_fd_RequestBus(void)
{
	if (!fc)
	{
		return otg_fc_ErrorNoFunctionDevice;
	}

	return otg_fc_RequestBus(fc);
}

void
otg_fd_ReleaseBus(void)
{
	if (!fc)
	{
		return;
	}

	return otg_fc_ReleaseBus(fc);
}

int
otg_fd_Submit(otg_urb_t *urb)
{
	if (currentDevice && currentDevice == otg_urb_GetDevice(urb))
	{
		if (otg_fc_GetConfiguration(fc) >= 0)
		{
			return otg_fc_Submit(fc, urb);
		}
	}

	return -1;
}

static int
otg_fd_StandardSetup(otg_device_request_t *request, void *private)
{
	otg_fdevice_t *otg;

#ifdef DEBUG_FD
	info("StandardSetup: ");
#endif
	if (request->bRequest == otg_ep0_RequestGetDescriptor)
	{
		if (otg_ep0_IsDeviceToHostRequest(request->bmRequestType) &&
			otg_ep0_IsStandardRequest(request->bmRequestType) &&
			otg_ep0_IsDeviceRequest(request->bmRequestType) &&
			(request->wValue & 0xFF00) == 0x900)
		{
			/* Get OTG descriptor */
			otg = *devices;
			if (otg)
			{
				return otg->standardSetup(request, private);
			}
		}
	}

#ifdef DEBUG_FD
	info("StandardSetup: currentDevice=%p", currentDevice);
#endif
	if (currentDevice)
	{
		if (currentDevice->standardSetup)
		{
			return currentDevice->standardSetup(request, private);
		}
		else
		{
			return 0;
		}
	}

	return -1;
}

static int
otg_fd_ClassSetup(otg_device_request_t *request, void *private)
{
#ifdef DEBUG_FD
	info("ClassSetup: currentDevice=%p", currentDevice);
#endif
	if (currentDevice)
	{
		if (currentDevice->classSetup)
		{
			return currentDevice->classSetup(request, private);
		}
		else
		{
			return 0;
		}
	}

	return -1;
}

static int
otg_fd_VendorSetup(otg_device_request_t *request, void *private)
{
#ifdef DEBUG_FD
	info("VendorSetup: currentDevice=%p", currentDevice);
#endif
	if (currentDevice)
	{
		if (currentDevice->vendorSetup)
		{
			return currentDevice->vendorSetup(request, private);
		}
		else
		{
			return 0;
		}
	}

	return -1;
}


static int
otg_fd_SetDriver(int handle, otg_fdevice_t *dev)
{
	int rc;
	otg_fdevice_t *d;

#ifdef DEBUG_FD
	info("SetDriver: handle=%d", handle);
#endif
	d = dev;
	d->handle = handle;
	if (currentDevice)
	{
		otg_fc_Stop(fc);
		otg_fc_Reset(fc);
		currentDevice = 0;
	}

	OS_Wait(1);

	rc = otg_fc_SetStrDesc(fc, d->strDesc, d->strDescLength);
	if (rc)
	{
		return rc;
	}

	rc = otg_fc_SetDeviceDesc(fc, d->deviceDesc, d->deviceDescLength);
	if (rc)
	{
		return rc;
	}

	rc = otg_fc_SetConfigDesc(fc, d->configDesc, d->configDescLength);
	if (rc)
	{
		return rc;
	}

	rc = otg_fc_SetCallBack
		 (
		 	fc,
			otg_fd_StandardSetup,
			otg_fd_ClassSetup,
			otg_fd_VendorSetup
		 );
	if (rc)
	{
		return rc;
	}

	currentDevice = d;
	d->handle = handle;
	devices[handle] = d;

	otg_fc_Start(fc);

	return 0;
}

static int
otg_fd_ResetDriver(int handle, otg_fdevice_t *dev)
{
	otg_fdevice_t *d;

#ifdef DEBUG_FD
	info("ResetDriver: handle=%d", handle);
#endif
	d = devices[handle];
	if (d != dev)
	{
		return 0;
	}

	if (currentDevice == d)
	{
		if (d == *devices)
		{
			/* reset OTG dev */
			otg_fc_Stop(fc);
			otg_fc_Reset(fc);
			currentDevice = 0;
		}
		else
		{
			if (*devices)
			{
				otg_fd_SetDriver(0, *devices);
			}
			else
			{
				otg_fc_Stop(fc);
				otg_fc_Reset(fc);
				currentDevice = 0;
			}
		}
	}

	devices[handle] = 0;

	return 0;
}

int
otg_fd_Register(otg_fdevice_t *dev)
{
	int i;

#ifdef DEBUG_FD
	info("Register: dev=%p", dev);
#endif
	for(i = 1; i < otg_fd_DEVICE_MAX; i++)
	{
		if (!devices[i])
		{
			if (otg_fd_SetDriver(i, dev))
			{
				return -1;
			}

			return i;
		}
	}

	return -1;
}

void
otg_fd_Unregister(otg_fdevice_t *dev)
{
	int i;

#ifdef DEBUG_FD
	info("Unregister: dev=%p", dev);
#endif
	for(i = 1; i < otg_fd_DEVICE_MAX; i++)
	{
		if (devices[i] == dev)
		{
			otg_fd_ResetDriver(i, dev);
			return;
		}
	}
}

int
otg_fd_RegisterOtg(otg_fdevice_t *dev)
{
#ifdef DEBUG_FD
	info("RegisterOtg: dev=%p", dev);
#endif
	if (*devices)
	{
		return -1;
	}

	if (!currentDevice)
	{
		otg_fd_SetDriver(0, dev);
	}

	return 0;
}

void
otg_fd_UnregisterOtg(otg_fdevice_t *d)
{
#ifdef DEBUG_FD
	info("UnregisterOtg: dev=%p", d);
#endif
	if (d == *devices)
	{
		otg_fd_ResetDriver(0, d);
	}
}

static int __init
otg_fd_init(void)
{
	fc = otg_fc_New();
	if (!fc)
	{
		return -1;
	}

	devices = OS_Malloc(sizeof(otg_fdevice_t *) * otg_fd_DEVICE_MAX);
	if (!devices)
	{
		otg_fc_Reset(fc);
		otg_fc_Delete(fc);
		fc = 0;
		return -1;
	}

	OS_MemSet(devices, 0, sizeof(otg_fdevice_t *) *otg_fd_DEVICE_MAX);
	currentDevice = 0;

	return 0;
}

static void __exit
otg_fd_exit(void)
{
	int i;

	if (devices)
	{
		for(i = otg_fd_DEVICE_MAX - 1; i >= 0; i--)
		{
			if (devices[i])
			{
				otg_fd_Unregister(devices[i]);
			}
		}

		OS_Free(devices);
		devices = 0;
	}

	if (fc)
	{
		otg_fc_Reset(fc);
		otg_fc_Delete(fc);
		fc = 0;
	}
}

module_init(otg_fd_init);
module_exit(otg_fd_exit);

EXPORT_SYMBOL(otg_fd_Register);
EXPORT_SYMBOL(otg_fd_Unregister);
EXPORT_SYMBOL(otg_fd_RegisterOtg);
EXPORT_SYMBOL(otg_fd_UnregisterOtg);
EXPORT_SYMBOL(otg_fd_Submit);
EXPORT_SYMBOL(otg_fd_RequestBus);
EXPORT_SYMBOL(otg_fd_ReleaseBus);
EXPORT_SYMBOL(otg_fd_Stall);
EXPORT_SYMBOL(otg_fd_ClearStall);
