/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_hc.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    Host controller object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

#ifdef DEBUG_DONEQUEUE
extern int done_queue_found;
#endif

static otg_hc_t *hostController = 0;

#ifdef DEBUG_MEMORY
static int memoryError = 0;
static int isoPreStartFrame = -1;
#endif

#ifdef USE_KERNEL_THREAD
static OS_WaitQueueDeclare(khcdWait);
static OS_CompletionCreate(khcdExited);
static int khcdPid = 0;
static int otg_hc_Thread(void *nothing);
#endif

#ifdef USE_TASKLET
OS_TaskletDeclare(hcTasklet, otg_hc_Tasklet);
static int taskletWaitCount = 0;
#endif

int
otg_hc_New(otg_hc_t * otg_hc)
{
#ifndef OTG_FUNCTION_ONLY
	otg_t *otg;
#ifdef USE_KERNEL_THREAD
	int pid;
#endif

	info("otg_hc_New: hc=%p", otg_hc);

	if (!hostController)
	{
		otg = otg_New(otg_hc);
		if (!otg)
		{
			return -1;
		}

		hostController = otg_GetHost(otg);

#ifdef USE_KERNEL_THREAD
		pid = kernel_thread
			(
				otg_hc_Thread,
				NULL,
				CLONE_FS | CLONE_FILES | CLONE_SIGHAND
			);

		if (pid < 0)
		{
			otg_Delete(otg);
			hostController = 0;
			err("New: failed to start hc daemon");
			return 0;
		}

		khcdPid = pid;
#endif
	}

	return 0;
#else
	return -1;
#endif
}

void
otg_hc_Delete(otg_hc_t *hc)
{
	otg_Delete(hc->otg);
	hostController = 0;

#ifdef USE_KERNEL_THREAD
	kill_proc(khcdPid, SIGTERM, 1);
	OS_CompletionWait(&khcdExited);
#endif
}

#ifndef OTG_FUNCTION_ONLY
int
otg_hc_Init(otg_hc_t *hc, otg_t *otg, int start, int size)
{
	int i;

	info("otg_hc_Init: hc=%p", hc);

	hc->otg = otg;
	hc->started = 0;

	otg_mem_Init(&hc->mem, start, size);
	otg_hc_intr_Init(&hc->intr, hc);

	OS_ListInit(&hc->isoList);
	OS_ListInit(&hc->ctrlList);
	OS_ListInit(&hc->bulkList);
	
	hc->intrInterval = 0; /* no timer */
	for(i = 0; i < 8; i++)
	{
		OS_ListInit(&hc->intrList[i]);
	}

#ifdef OTG_ISO_SHARE_MEMORY
	OS_ListInit(&hc->sofList);
#endif

	/* Initially, there is two ports available. The third port 
	 * will be added after a device is connected and the MNP
	 * decide that this will act as the host.
	 */
	otg_hc_rh_Init(&hc->rh, hc);
	otg_hc_etd_Init(&hc->etd_bank, hc);

	return 0;
}
#endif

void
otg_hc_Reset(otg_hc_t *hc)
{
#ifndef OTG_FUNCTION_ONLY
	unsigned long flags;

	otg_LockChipIntr(hc->otg, flags);
	otg_hc_ResetUnlocked(hc);
	otg_UnlockChipIntr(hc->otg, flags);
#endif
}

#ifndef OTG_FUNCTION_ONLY
void
otg_hc_ResetUnlocked(otg_hc_t *hc)
{
	UINT32 w;

#ifdef DEBUG_OTG_CHIP
info("Reset: hc=%p", hc);
#endif
	info("Reset: hc->otg=%p", hc->otg);

	w = otg_ReadHostReg(hc->otg, otg_hc_Control_c);
	w &= ~otg_hc_HostState_c;
	w |= otg_hc_USBReset_c;
	otg_WriteHostReg(hc->otg, otg_hc_Control_c, w);

	otg_hc_UnlinkAllUnlocked(hc, 0);

	OS_Wait(10);

	otg_hc_intr_Reset(&hc->intr);
	otg_hc_rh_Reset(&hc->rh);
	otg_hc_etd_Reset(&hc->etd_bank);
}
#endif

void
otg_hc_GoState(otg_hc_t *hc, int state)
{
	UINT32 s;
	int current_state;

#ifdef DEBUG_OTG_CHIP
info("GoState: hc=%p state=0x%0x otg=%p", hc,state,hc->otg);
#endif
	s = otg_ReadHostReg(hc->otg, otg_hc_Control_c);
	current_state = s & otg_hc_HostState_c;
	s &= ~otg_hc_HostState_c;
	switch (state)
	{
		case otg_hc_USBReset_c:
			break;
		case otg_hc_USBResume_c:
			if (current_state == otg_hc_USBOperational_c)
			{
				return;
			}

			if (current_state != otg_hc_USBSuspend_c)
			{
				return;
			}
			break;
		case otg_hc_USBOperational_c:
			break;
		case otg_hc_USBSuspend_c:
			if (current_state == otg_hc_USBSuspend_c)
			{
				return;
			}

			otg_hc_rh_GoSuspend(&hc->rh);
			OS_Wait(5);

			break;
		default:
			err("Unknown host controller state: %d", state);
			return;
	}

	s |= state;
	otg_WriteHostReg(hc->otg, otg_hc_Control_c, s);
}

int
otg_hc_Start(otg_hc_t *hc)
{
	int timeout = 100;

#ifndef OTG_FUNCTION_ONLY
#ifdef DEBUG_OTG_CHIP
	info("Start: started=%d", hc->started);
#endif

	if (hc->started)
	{
		return 0;
	}

	do
	{
		otg_hc_GoState(hc, otg_hc_USBOperational_c);
		if (otg_hc_IsInState(hc, otg_hc_USBOperational_c))
		{
			otg_hc_rh_DoStart(&hc->rh);
			hc->started = 1;
			otg_intr_EnableHostController(otg_GetIntr(hc->otg));
			return 0;
		}

		OS_Wait(1);
	} while (timeout--);

	return -1;
#else
	return 0;
#endif
}

void
otg_hc_Exit(otg_hc_t *hc)
{
#ifdef DEBUG_OTG_CHIP
	info("Exit:");
#endif

	if (hc->intrInterval)
	{
		OS_StopTimer(&hc->intrTimer);
		hc->intrInterval = 0;
	}

	/* Cleanup, reset etc. */
	otg_hc_rh_Stop(&hc->rh);
}

void
otg_hc_DoIntr(otg_hc_t *hc)
{
	otg_hc_intr_t *intr;

	otg_hc_ReadFrameNo(hc);
	intr = &hc->intr;

	/* Need to check done queue first. In case that last filled and done queue
	 * comes at the same time but after the processing of filled.
	 */
	otg_hc_intr_Check(intr, otg_hc_intr_All_c);
	otg_hc_etd_DoIntr(&hc->etd_bank);

	while (otg_hc_intr_IsTriggered(intr, otg_hc_intr_All_c))
	{
		/* Process host controller system level interrupts */
		if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_SOF_c))
		{
			/* Start Of Frame */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_SOF_c);
			otg_DoSofIntr(hc->otg);
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_SO_c))
		{
			/* Scheduling Overrun */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_SO_c);
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_RDQ_c))
		{
			/* Readback Done Queue*/
			otg_hc_etd_UploadDoneQueue(&hc->etd_bank);

#ifdef USE_KERNEL_THREAD
			OS_WaitQueueWakeup(&khcdWait);
#else
#ifdef USE_TASKLET
			taskletWaitCount++;
#ifdef DEBUG_TASKLET
info("Tasklet scheduled (INT)...");
#endif
			OS_TaskletSchedule(&hcTasklet);
#else
			otg_hc_etd_DoDoneQueue(&hc->etd_bank);
			otg_hc_DoWaitingEtd(hc);
#endif
#endif
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_RD_c))
		{
			/* Resume Detected */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_RD_c);
			otg_hc_rh_GoResume(&hc->rh);
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_UE_c))
		{
			/* Unrecoverable Error */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_UE_c);
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_FNO_c))
		{
			/* Frame Number Overflow */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_FNO_c);
		}
		else if (otg_hc_intr_IsTriggered(intr, otg_hc_intr_RHSC_c))
		{
			/* Root Hub Status Change */
			otg_hc_intr_ClearIntr(intr, otg_hc_intr_RHSC_c);
		}
	}
}

int
otg_hc_Allocate(otg_hc_t *hc, otg_etd_t *etd, int length)
{
	int addrX;
	int addrY;
	int etdIndex;

#ifdef DEBUG_SOME_MEMORY
info("Allocate: hc=%p etd=%p length=%d",hc,etd,length);
#endif
	addrX = -1;
	if (!otg_etd_HasBuffer(etd, 0))
	{
#ifdef DEBUG_SOME_MEMORY
info("Allocate: allocating X buffer");
#endif
		addrX = otg_mem_Allocate(&hc->mem, length);
		if (addrX < 0)
		{
			err("No memory on host controller");
#ifdef DEBUG_MEMORY
memoryError = 1;
#endif
			return -1;
		}

		otg_etd_SetBufferX(etd, addrX);
	}

	addrY = -1;
	if (otg_etd_NeedYBuffer(etd) && !otg_etd_HasBuffer(etd, 1))
	{
#ifdef DEBUG_SOME_MEMORY
info("Allocate: allocating Y buffer");
#endif
		addrY = otg_mem_Allocate(&hc->mem, length);
		if (addrY < 0)
		{
			err("No memory on host controller");
			if (addrX >= 0)
			{
				otg_mem_Release(&hc->mem, addrX);
				otg_etd_ClearBuffer(etd, 0);
			}

#ifdef DEBUG_MEMORY
memoryError = 1;
#endif
			return -1;
		}

#ifdef DEBUG_SOME_MEMORY
info("Allocate: addrY=0x%0x", addrY);
#endif
		otg_etd_SetBufferY(etd, addrY);
	}

	if (otg_etd_NeedEtd(etd))
	{
		etdIndex = otg_hc_etd_Allocate(&hc->etd_bank, etd);
#ifdef DEBUG_SOME_MEMORY
info("Allocate: etdIndex=%d", etdIndex);
#endif
		if (etdIndex < 0)
		{
			if (addrY >= 0)
			{
				otg_mem_Release(&hc->mem, addrY);
				otg_etd_ClearBuffer(etd, 1);
			}

			if (addrX)
			{
				otg_mem_Release(&hc->mem, addrX);
				otg_etd_ClearBuffer(etd, 0);
			}

			return otg_OUT_OF_ETDS;
		}
		else
		{
			otg_etd_SetEtdIndex(etd, etdIndex);
		}
	}

#ifdef DEBUG_SOME_MEMORY
info("Allocate: etd->buffer=0x%0x", etd->buffer);
#endif

#ifdef DEBUG_SOME_OTG
	info("otg_hc_Allocate etd=%p 0x%0x 0x%0x", etd,
		otg_etd_GetBuffer(etd, 1), otg_etd_GetBuffer(etd, 0));
#endif
	return 0;
}

void
otg_hc_Release(otg_hc_t *hc, otg_etd_t *etd)
{
	int addr;
#ifdef DEBUG_SOME_OTG
	info("otg_hc_Release etd=%p 0x%0x 0x%0x", etd,
		otg_etd_GetBuffer(etd, 1),
		otg_etd_GetBuffer(etd, 0));
#endif

#ifdef OTG_ISO_SHARE_MEMORY
	if (!otg_etd_IsMemoryUsed(etd))
	{
#endif
	addr = otg_etd_GetBuffer(etd, 1);
#ifdef DEBUG_SOME_MEMORY
info("Release: etd=%p addr1=0x%0x", etd, addr);
#endif
	if (addr >= 0)
	{
		otg_mem_Release(&hc->mem, addr);
		otg_etd_ClearBuffer(etd, 1);
	}

	addr = otg_etd_GetBuffer(etd, 0);
#ifdef DEBUG_SOME_MEMORY
info("Release: addr0=0x%0x", addr);
#endif
	if (addr >= 0)
	{
		otg_mem_Release(&hc->mem, addr);
		otg_etd_ClearBuffer(etd, 0);
	}
#ifdef OTG_ISO_SHARE_MEMORY
	}
#endif

	addr = otg_etd_GetEtdIndex(etd);
#ifdef DEBUG_SOME_MEMORY
info("Release: index=%d", addr);
#endif
	if (addr >= 0)
	{
		otg_hc_etd_Release(&hc->etd_bank, addr);
		otg_etd_ClearEtdIndex(etd);
	}
#ifdef DEBUG_SOME_MEMORY
info("Release: done");
#endif
}

int
otg_hc_SubmitEtd(otg_hc_t *hc, otg_etd_t *etd, stack_urb_t *urb)
{
	int rc;
	otg_etd_t *e;

#ifdef DEBUG_OTG_SUBMIT
info("SubmitEtd: pipe=0x%x", stack_urb_GetPipe(urb));
#endif
	rc = otg_etd_Fill(etd, hc, urb);
	if (rc)
	{
		if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
		{
			otg_hc_DeleteEtdForUrb(hc, etd);
		}

		return rc;
	}

	rc = 0;
	/* For first iso packet, download and start it */
	if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
	{
		e = etd;
		while (OS_ListNext(&e->formatEtdList) != &hc->isoList)
		{
			e = OS_ListEntry(OS_ListNext(&e->formatEtdList), otg_etd_t, formatEtdList);
			rc = otg_etd_Fill(e, hc, urb);
			if (rc)
			{
				break;
			}
		}

		if (rc)
		{
			otg_hc_DeleteEtdForUrb(hc, etd);
			return rc;
		}

		if (otg_etd_NeedEtd(etd))
		{
			/* No ETD allocated, need to wait */
			return 0;
		}

		rc = otg_hc_etd_Transfer(&hc->etd_bank, etd);	
		if (rc)
		{
			otg_hc_DeleteEtdForUrb(hc, etd);
		}
		else
		{
			if (OS_ListNext(&etd->formatEtdList) != &hc->isoList)
			{
				e = OS_ListEntry(OS_ListNext(&etd->formatEtdList), otg_etd_t, formatEtdList);
				if (otg_etd_NeedEtd(e))
				{
					return 0;
				}

				rc = otg_hc_etd_Transfer(&hc->etd_bank, e);	
				if (rc)
				{
					otg_hc_DeleteEtdForUrb(hc, etd);
				}
			}
		}

#ifdef DEBUG_OTG_DUMPREG
if (hc->intr.enabledIntr != otg_hc_ReadReg(hc, otg_hc_intr_Enable_c))
{
	info("SubmitEtd: dump reg ");
	otg_DumpReg(hc->otg);
}
#endif
	}
	else
	{
		rc = otg_hc_Allocate(hc, etd, otg_etd_GetBufferSize(etd));
#ifdef DEBUG_OTG_OTHERS
info("SubmitEtd: allocate with rc=%d", rc);
#endif
		if (rc)
		{
			return 0; /* no resource. wait */
		}

		rc = otg_hc_etd_Transfer(&hc->etd_bank, etd);
	}

	/* NB: Reg's wrong here */
#ifdef DEBUG_OTG_DUMPREG
	info("SubmitEtd: dump reg ");
	otg_DumpReg(hc->otg);
#endif

	if (rc)
	{
		/* Transfer failed immediately. Cleanup */
		otg_hc_Release(hc, etd);
	}

	otg_hc_intr_Enable
	(
		&hc->intr,
		otg_hc_intr_SO_c | otg_hc_intr_RDQ_c | otg_hc_intr_RD_c |
		otg_hc_intr_UE_c
	);

	return 0;
}

int
otg_hc_Submit(otg_hc_t *hc, stack_urb_t *urb)
{
	otg_etd_t *etd;
	int rc;
	unsigned long flags;

#ifdef DEBUG_MEMORY
if (memoryError)
{
	return -ENOMEM;
}
#endif

#ifdef DEBUG_OTG_SUBMIT
info("Submit: dev=%p urb=%p pipe=0x%x urb_addr=0x%x rh_addr=0x%x",
	stack_urb_GetDevice(urb), urb, stack_urb_GetPipe(urb), stack_urb_GetAddress(urb), otg_hc_rh_GetAddress(&hc->rh));
#endif
	if (!stack_urb_GetDevice(urb) || !stack_urb_GetBus(urb) || !hc->started)
	{
#ifdef DEBUG_OTG_OTHERS
info("Submit ended -ENODEV");
#endif
		return -ENODEV;
	}

	/* NB: Roothub is not handled that way anymore... TODO: check */
/*	if (stack_urb_GetAddress(urb) == otg_hc_rh_GetAddress(&hc->rh))
	{
		rc = otg_hc_rh_Submit(&hc->rh, urb);
	}
	else
	{*/
#ifdef DEBUG_DELAY
if (stack_urb_IsCtrlPipe(urb))
{
	OS_Wait(1000);
}
#endif
		etd = otg_etd_New(hc);
		if (!etd)
		{
#ifdef DEBUG_OTG_OTHERS
info("Submit ended -ENOMEM");
#endif
			return -ENOMEM;
		}

		rc = 0;
		otg_LockChipIntr(hc->otg, flags);

#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
		otg_hc_ReadFrameNo(hc);
		switch (stack_urb_GetPipeType(urb))
		{
			case stack_urb_PipeTypeCtrl:
#ifdef DEBUG_OTG_OTHERS
				info("Submit LinkCtrl");
#endif
				rc = otg_hc_LinkCtrl(hc, etd, urb);
				break;
			case stack_urb_PipeTypeBulk:
#ifdef DEBUG_OTG_OTHERS
				info("Submit LinkBulk");
#endif
				rc = otg_hc_LinkBulk(hc, etd, urb);
				break;
			case stack_urb_PipeTypeIntr:
#ifdef DEBUG_OTG_OTHERS
				info("Submit LinkIntr");
#endif
				rc = otg_hc_LinkIntr(hc, etd, urb);
				break;
			case stack_urb_PipeTypeIso:
#ifdef DEBUG_OTG_OTHERS
				info("Submit LinkIso");
#endif
				rc = otg_hc_LinkIso(hc, etd, urb);
				break;
		}

		/* NB: storage bug: still the same reg dumps here */
#ifdef DEBUG_OTG_OTHERS
		info("Submit pipe ended");
#endif
		if (rc)
		{
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
			otg_UnlockChipIntr(hc->otg, flags);
			otg_etd_Delete(etd);
			return rc;
		}

		stack_IncDeviceUse(stack_urb_GetDevice(urb));
#ifdef DEBUG_OTG_OTHERS
		info("Submit ETD");
#endif
		rc = otg_hc_SubmitEtd(hc, etd, urb);
#ifdef DEBUG_OTG_OTHERS
		info("Submit end=%d", rc);
#endif
		if (rc)
		{
			otg_etd_Delete(etd);
			stack_DecDeviceUse(stack_urb_GetDevice(urb));
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
			otg_UnlockChipIntr(hc->otg, flags);
			return rc;
		}

#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif

		otg_UnlockChipIntr(hc->otg, flags);

	//}
#ifdef DEBUG_OTG_OTHERS
	info("Submit ended normally");
#endif
	return 0;
}

int
otg_hc_DoWaitingIntr(otg_hc_t *hc)
{
	OS_ListDeclare(*h);
	OS_ListDeclare(*list);
	otg_etd_t *etd;
	int min_frame;
	int max_frame;
	int frames;
	int mask;
	int start;
	int i;
	int interval;
	int rc;

	/* Convert to relative start frame of each interrupt */
	/*   1ms 0
	 *   2ms 0
	 *   4ms 1
	 *   8ms 3
	 *  16ms 7
	 *  32ms 15
	 *  64ms 31
	 * 128ms 63
	 * 256ms 127
	 * Here only process interval longer than 4ms. Interrupts with
	 * 1ms, 2ms and 4ms will be reenabled after they retired.
	 */
	
	interval = 0;
	for(i = 3; i < 8; i++)
	{
		list = &hc->intrList[i];
		for (h = OS_ListNext(list); h != list; h = OS_ListNext(h))
		{
			etd = OS_ListEntry(h, otg_etd_t, formatEtdList);
			if (otg_etd_IsInHc(etd))
			{
				/* Already in transfer */
				continue;
			}

#ifdef DEBUG_OTG_SUBMIT
info("DoWaitingIntr: etd=%p", etd);
#endif
			mask = (1 << i) - 1;
			min_frame = hc->intrPreviousFrame & mask;
			frames = (hc->frameNo - hc->intrPreviousFrame) &
						otg_hc_FmIntervalMask_c;
			max_frame = hc->frameNo & mask;
			start = (1 << (i - 1)) - 1;
#ifdef DEBUG_OTG_SUBMIT
info("DoWaitingIntr: mask=0x%0x i=%d", mask, i);
info("DoWaitingIntr: min_frame=0x%0x max_frame=0x%0x", min_frame, max_frame);
info("DoWaitingIntr: start=0x%0x", start);
#endif
			if (frames < mask && ((min_frame >= start && start > max_frame) ||
				(min_frame < start && max_frame < start)))
			{
				/* Not its time yet */
				if (!interval)
				{
					interval = (start - max_frame) & mask;
				}

				continue;
			}

			rc = otg_hc_Allocate(hc, etd, otg_etd_GetBufferSize(etd));
			if (rc)
			{
				/* No resource, delay process for three ms */
				if (!interval)
				{
					/* Return a minus number to tell some ETD is skipped */
					interval = - (mask + 1);
				}

				break;
			}

			otg_td_intr_Fill(etd, hc);
			otg_hc_etd_Transfer(&hc->etd_bank, etd);	
		}
	}

	return interval;
}

int
otg_hc_DoWaitingCtrlBulk(otg_hc_t *hc, OS_ListDeclare(*list))
{
	OS_ListDeclare(*h);
	otg_etd_t *etd;
	int rc;

	for (h = OS_ListNext(list); h != list; h = OS_ListNext(h))
	{
		etd = OS_ListEntry(h, otg_etd_t, formatEtdList);
		if (otg_etd_IsInHc(etd))
		{
			/* Already in transfer */
			continue;
		}

		rc = otg_hc_Allocate(hc, etd, otg_etd_GetBufferSize(etd));
		if (rc)
		{
			/* No resource, delay process */
			return rc;
		}

		otg_hc_etd_Transfer(&hc->etd_bank, etd);	
	}

	return 0;
}

int
otg_hc_DoWaitingEtd(otg_hc_t *hc)
{
	int rc;

	/* Search for all TD which is waiting to be transfered */
	rc = otg_hc_DoWaitingCtrlBulk(hc, &hc->ctrlList);
	if (rc)
	{
		return rc;
	}

	rc = otg_hc_DoWaitingCtrlBulk(hc, &hc->bulkList);

	return rc;
}

void
otg_hc_EtdDone(otg_hc_t *hc, otg_etd_t *etd)
{
	stack_urb_t *urb;
	otg_etd_t *new_etd;
	stack_device_t *dev;
	int rc;
	int wait;

	if (otg_etd_IsUnlink(etd))
	{
#ifdef DEBUG_UNLINK
info("EtdDone: etd=%p is dead", etd);
#endif
		/* ETD is dead */
		otg_hc_Release(hc, etd);
		otg_etd_Delete(etd);
		return;
	}

	/* Urb finished. Proceed to the next urb */
	urb = otg_etd_GetUrb(etd);
	if (otg_etd_IsTransferType(etd, otg_etd_Format_Intr_c))
	{
		new_etd = etd;
	}
	else
	{
		new_etd = otg_etd_GetWaiting(etd);
	}

#ifdef DEBUG_BULK
if (otg_etd_IsTransferType(etd, otg_etd_Format_Bulk_c))
{
	if (stack_urb_GetStatus(urb) == -EAGAIN)
	{
		otg_hc_etd_Transfer(&hc->etd_bank, etd);	
		return;
	}
}
#endif

#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: do the next new_etd=%p (etd=%p)", new_etd, etd);
#endif
	/* Process next ETD first because of time constraint */
	if (!new_etd)
	{
		/* All urb for this EP has been processed release resouce */
		otg_hc_Release(hc, etd);
		if (stack_urb_HasBandwidth(urb))
		{
			if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
			{
#ifdef DEBUG_DONEQUEUE
if (done_queue_found)
{
	new_etd = OS_ListEntry(OS_ListNext(&etd->pipeEtdList), otg_etd_t, pipeEtdList);
	printk("buffer=0x%0x etd%d etd=%p urb=%p\n",
		new_etd->buffer, new_etd->etdIndex, new_etd, new_etd->urb);
}
#endif
				stack_urb_ReleaseBandwidth(urb, 1);
				if (!otg_td_iso_IsLastForUrb(etd))
				{
					/* Not the last ETD for urb, just return */
					otg_etd_Delete(etd);
					return;
				}
			}
			else if (otg_etd_IsTransferType(etd, otg_etd_Format_Intr_c) &&
					 stack_urb_GetIntrInterval(urb))
			{
				stack_urb_ReleaseBandwidth(urb, 0);
			}
		}
#ifdef DEBUG_OTG_SUBMIT
if (otg_etd_IsTransferType(etd, otg_etd_Format_Ctrl_c))
{
	int i;
	UINT8 *cp;

	info("EtdDone: Ctrl Transfer Done");
	cp = (UINT8 *)urb->transfer_buffer;
	for(i = 0; i < urb->actual_length; i++)
	{
		printk("0x%02x ", *cp);
		cp++;
	}

	if (urb->actual_length)
	{
		printk("\n");
	}
}
#endif
	}
	else /* new_etd != NULL */
	{
		if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
		{
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: transfer iso");
#endif
#ifdef OTG_ISO_SHARE_MEMORY
			otg_etd_CopyEtdInfo(new_etd, etd);
			otg_hc_etd_Transfer(&hc->etd_bank, new_etd);	
#else
			if (otg_etd_CopyHcInfo(new_etd, etd))
			{
				stack_urb_t *new_urb;
				stack_device_t *new_dev;

				otg_hc_Release(hc, etd);
				otg_etd_Delete(new_etd);

				new_urb = otg_etd_GetUrb(etd);
				stack_urb_SetStatus(new_urb, -ENOMEM);
				new_dev = stack_urb_GetDevice(new_urb);
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: no memory for new_urb=%p", new_urb);
#endif
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
				otg_UnlockChip(hc->otg);
				otg_hcd_giveback_urb(hc, urb);
				otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif

				stack_DecDeviceUse(new_dev);
			}
			else
			{
				otg_hc_etd_Transfer(&hc->etd_bank, new_etd);	
			}
#endif //OTG_ISO_SHARE_MEMORY

			if (!otg_td_iso_IsLastForUrb(etd))
			{
				/* Not the last ETD for urb, just return */
				otg_etd_Delete(etd);
				return;
			}
#ifdef DEBUG_UNLINK
info("EtdDone: etd=%p ctrl=0x%08x", etd, etd->td.iso.ctrl);
#endif
		}
		else if (otg_etd_IsTransferType(etd, otg_etd_Format_Intr_c))
		{
			/* NB: Added, Release doesn't work, Delete does */
#ifdef DEBUG_SOME_OTG
	info("otg_etd_Delete location 1");
#endif
			otg_hc_Release(hc, etd);
			otg_etd_Delete(etd);

#ifdef DEBUG_OTG_SUBMIT
info("EtdDone (intr): giving back the urb (interrupt type) urb->status=%d actual_length=%d",
	stack_urb_GetStatus(urb), stack_urb_GetActualLength(urb));
#endif
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
			otg_UnlockChip(hc->otg);
			otg_hcd_giveback_urb(hc, urb);
			otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
			

			/* NB: Do not reenable intr ETD...
			 * This is now done at an upper layer in the kernel */
#if 0
			if (stack_urb_GetIntrInterval(urb) < 8)
			{
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: reenable intr ETD");
#endif
				/* For int ep with less than 8ms interval, renable it*/
				otg_td_intr_Fill(etd, hc);
				otg_hc_etd_Transfer(&hc->etd_bank, etd);	
			}
			else
			{
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: retire intr ETD");
#endif
				/* Retire ETD, wait for another time */
				otg_hc_Release(hc, etd);
				wait = otg_hc_GetStartFrameNo(hc, stack_urb_GetIntrInterval(urb))
						- hc->frameNo - 2;
				otg_hc_StartIntrTimer(hc, wait);
			}
#endif

			return;
		}
		else if (otg_etd_IsTransferType(etd, otg_etd_Format_Ctrl_c))
		{
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: do next packet in Ctrl transfer");
#endif
			if (stack_urb_IsErrorStatus(urb))
			{
				rc = stack_urb_GetStatus(urb);
			}
			else
			{
				rc = otg_etd_CopyHcInfo(new_etd, etd);
			}

			if (rc)
			{
				otg_hc_Release(hc, etd);
				otg_td_ctrl_Delete(etd);

				stack_urb_SetStatus(urb, rc);
			}
			else
			{
				OS_ListAddTail(&new_etd->formatEtdList, &hc->ctrlList);
				otg_hc_etd_Transfer(&hc->etd_bank, new_etd);	

				/* For ctrl transfer, only the last ETD finish urb */
#ifdef DEBUG_SOME_OTG
	info("otg_etd_Delete location 1a");
#endif
				otg_etd_Delete(etd);
				return;
			}
		}
		else
		{
			/* For Bulk transfer, just link them to formatEtdList
			 * DoWaiting function will get it
			 */
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: link bulk");
#endif
			OS_ListAddTail(&new_etd->formatEtdList, &hc->bulkList);
		}
	}

#ifdef DEBUG_SOME_OTG
	info("otg_etd_Delete location 2");
#endif
	otg_etd_Delete(etd);

	/* complete function may clear dev */
	dev = stack_urb_GetDevice(urb);

	/* Now let's finish urb */
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: giving back the urb urb->status=%d actual_length=%d",
	stack_urb_GetStatus(urb), stack_urb_GetActualLength(urb));
#endif
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
	otg_UnlockChip(hc->otg);
	otg_hcd_giveback_urb(hc, urb);
	otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
#ifdef DEBUG_OTG_SUBMIT
info("EtdDone: gave back the urb");
#endif

	stack_DecDeviceUse(dev);

#if 0 // TODO: Maybe it's not necessary anymore...
	/* Process urbs on the next list */
	urb_list = stack_urb_GetNextUrb(urb);
	if (urb_list)
	{
		int count;
		int fail;

		count = 0;
		fail = 0;
		rc = 0;
		for (; urb_list != urb && urb_list; urb_list = OS_ListNext(urb_list))
		{
			count++;
			if (count >= otg_hc_MaxUrbCount)
			{
				err("too many urbs in urb list");
				break;
			}

			if (urb_list->status == -ENOENT)
			{
				break;
			}

			if (!fail && urb_list->status != -EINPROGRESS)
			{
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
				otg_UnlockChip(hc->otg);
				rc = otg_hc_Submit(hc, urb_list);
				otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
				if (rc)
				{
					fail = 1;;
				}
			}
		}

		if (urb_list == urb && stack_urb_GetStatus(urb) != -ENOENT)
		{
			/* urb is on the next list */
			stack_urb_SetDevice(urb, dev);
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
			otg_UnlockChip(hc->otg);
#ifdef DEBUG_DONEQUEUE
if (done_queue_found)
{
	printk("urb=%p start_frame=0x%0x\n",
		urb, stack_urb_GetStartFrame(urb));
}
#endif
			rc = otg_hc_Submit(hc, urb);
			otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
		}
#ifdef DEBUG_DONEQUEUE
if (rc || fail)
{
	urb_list = stack_urb_GetNextUrb(urb);
	info("EtdDone: frame=0x%0x rc=%d urb=%p start_frame=0x%0x",
		hc->frameNo, rc, urb, stack_urb_GetStartFrame(urb));
	for (; urb_list != urb && urb_list; urb_list = OS_ListNext(urb_list))
	{
		info("EtdDone: urb_list=%p start_frame=0x%0x",
			urb_list, stack_urb_GetStartFrame(urb_list));
	}
}

if (done_queue_found)
{
	done_queue_found = 0;
}
#endif
	}
#endif
}

int
otg_hc_UnlinkList(otg_hc_t *hc, stack_urb_t *urb, OS_ListDeclare(*list))
{
	OS_ListDeclare(*h);
	otg_etd_t *etd;
	int rc;

#ifdef DEBUG_UNLINK
info("UnlinkList: ");
#endif
	/* Have to do from first to last because of ISO ETDs */
	for (h = OS_ListNext(list); h != list; )
	{
		etd = OS_ListEntry(h, otg_etd_t, formatEtdList);
#ifdef DEBUG_UNLINK
info("UnlinkList: etd=%p", etd);
#endif
		h = OS_ListNext(h);
		if (!otg_etd_IsUnlink(etd) && etd->urb == urb)
		{
#ifdef DEBUG_UNLINK
info("UnlinkList: delete");
#endif
			rc = otg_hc_UnlinkEtd(hc, etd);
			if (rc)
			{
				return rc;
			}
		}
	}

#ifdef DEBUG_UNLINK
info("UnlinkList: exit");
OS_Wait(2000);
#endif
	return 0;
}

int
otg_hc_UnlinkUrb(otg_hc_t *hc, stack_urb_t *urb)
{
	unsigned long flags;
	int i;
	int rc = 0;

	otg_LockChipIntr(hc->otg, flags);
#ifdef DEBUG_UNLINK
info("UnlinkUrb: delete urb=%p", urb);
#endif
	/* NB: Roothub is not handled that way anymore
	if (stack_urb_GetAddress(urb) == otg_hc_rh_GetAddress(&hc->rh))
	{
		rc = otg_hc_rh_UnlinkUrb(&hc->rh, urb);
	}
	else
	{*/	
		switch (stack_urb_GetPipeType(urb))
		{
			case stack_urb_PipeTypeIso:
				rc = otg_hc_UnlinkList(hc, urb, &hc->isoList);
				break;
			case stack_urb_PipeTypeIntr:
				for (i = 0; i < 8; i++)
				{
					rc = otg_hc_UnlinkList(hc, urb, &hc->intrList[i]);
					if (rc)
					{
						break;
					}

				}
				break;
			case stack_urb_PipeTypeCtrl:
				rc = otg_hc_UnlinkList(hc, urb, &hc->ctrlList);
				break;
			case stack_urb_PipeTypeBulk:
				rc = otg_hc_UnlinkList(hc, urb, &hc->bulkList);
				break;
			default:
				rc = 0;
		}
	//}

	otg_UnlockChipIntr(hc->otg, flags);

	return rc;
}

void
otg_hc_IntrTimeout(unsigned long data)
{
	int interval;
	otg_hc_t *hc;
	unsigned long flags;

	OS_DisableInterrupts(flags);

	hc = (otg_hc_t *)data;
	otg_hc_ReadFrameNo(hc);

	interval = otg_hc_DoWaitingIntr(hc);
#ifdef DEBUG_OTG_SUBMIT
info("IntrTimeout: interval=%d", interval);
#endif

	hc->intrInterval = 0;

	if (!interval)
	{
		/* No waiting ETD */
		OS_EnableInterrupts(flags);
		return;
	}
	else if (interval < 0)
	{
		/* There are unprocessed ETD. Do not change original frame no */
		interval = -interval;
	}
	else
	{
		hc->intrPreviousFrame = hc->frameNo;
	}

	OS_EnableInterrupts(flags);

	otg_hc_StartIntrTimer(hc, interval);
}

void
otg_hc_StartIntrTimer(otg_hc_t *hc, int interval)
{
	OS_DeclareTimer(*timer);
	unsigned long expire;
	unsigned long flags;

	timer = &hc->intrTimer;
	if (interval <= 0)
	{
#ifdef DEBUG_OTG_SUBMIT
info("StartIntrTimer: interval=0x%0x", interval);
#endif
		interval = 1;
	}

#ifdef DEBUG_OTG_SUBMIT
info("StartIntrTimer: interval=%d", interval);
#endif

	OS_DisableInterrupts(flags);

#ifdef SIM_OTG
	expire = interval * 10;
#else
	expire = interval;
#endif
	if (hc->intrInterval)
	{
		if (!OS_IsTimerExpired(timer, expire))
		{
			/* No need to set new expire */
			OS_EnableInterrupts(flags);
			return;
		}

		/* Stop previous timer. */
		OS_StopTimer(timer);
	}
	else
	{
		hc->intrPreviousFrame = hc->frameNo;
	}

	hc->intrInterval = interval;

#ifdef DEBUG_OTG_SUBMIT
info("StartIntrTimer: frameNo=0x%0x interval=%d", hc->frameNo, interval);
#endif
	OS_StartTimer(timer, otg_hc_IntrTimeout, hc, expire);

	OS_EnableInterrupts(flags);
}

void
otg_hc_UnlinkAll(otg_hc_t *hc, stack_device_t *dev)
{
	unsigned long flags;

	otg_LockChipIntr(hc->otg, flags);
	otg_hc_UnlinkAllUnlocked(hc, dev);
	otg_UnlockChipIntr(hc->otg, flags);
}

void
otg_hc_UnlinkAllUnlocked(otg_hc_t *hc, stack_device_t *dev)
{
	int i;

	otg_hc_UnlinkAllEtd(hc, dev, &hc->isoList);
#ifdef DEBUG_UNLINK
	info("UnlinkAll: isoList unlinked");
#endif
	for(i = 0; i < 8; i++)
	{
		otg_hc_UnlinkAllEtd(hc, dev, &hc->intrList[i]);
	}
#ifdef DEBUG_UNLINK
	info("UnlinkAll: intrList unlinked");
#endif
	otg_hc_UnlinkAllEtd(hc, dev, &hc->ctrlList);
#ifdef DEBUG_UNLINK
	info("UnlinkAll: ctrlList unlinked");
#endif
	otg_hc_UnlinkAllEtd(hc, dev, &hc->bulkList);

#ifdef DEBUG_UNLINK
	info("UnlinkAll: exit");
#endif
}

int
otg_hc_UnlinkEtd(otg_hc_t *hc, otg_etd_t *etd)
{
	stack_urb_t *urb = etd->urb;
	stack_device_t *dev;

	if (!urb || !stack_urb_GetDevice(urb))
	{
#ifdef DEBUG_UNLINK
	info("UnlinkEtd: zero urb or zero urb->dev. etd=%p", etd);
#endif
		return 0;
	}

	if (otg_etd_IsUnlink(etd))
	{
#ifdef DEBUG_UNLINK
	info("UnlinkEtd: already unlinked etd=%p", etd);
#endif
		return 0;
	}

	switch (otg_etd_GetTransferType(etd))
	{
		case otg_etd_Format_Ctrl_c:
			/* Control transfers may have data and status stage */
			otg_td_ctrl_Delete(etd);
			break;
		case otg_etd_Format_Iso_c:
			stack_urb_ReleaseBandwidth(urb, 1);
			break;
		case otg_etd_Format_Intr_c:
			stack_urb_ReleaseBandwidth(urb, 0);
			break;
		default:
			break;
	}

#ifdef DEBUG_UNLINK
	info("UnlinkEtd: ");
#endif
	if (otg_etd_IsInHc(etd))
	{
		/* ETD is in HC. */
		otg_etd_MarkUnlink(etd);
	}
	else
	{
		otg_hc_Release(hc, etd);
#ifdef DEBUG_UNLINK
	info("UnlinkEtd: after release");
#endif
		otg_etd_Delete(etd);
	}

	if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
	{
		if (!otg_td_iso_IsLastForUrb(etd))
		{
			return 0;
		}
	}

#ifdef DEBUG_UNLINK
	info("UnlinkEtd: prepare to call give back urb=%p", urb);
#endif
	stack_urb_SetStatus(urb, -ENOENT);
	dev = stack_urb_GetDevice(urb);

	stack_urb_SetDevice(urb, 0);
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif
	otg_UnlockChip(hc->otg);
	otg_hcd_giveback_urb(hc, urb);
	otg_LockChip(hc->otg);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
#ifdef DEBUG_UNLINK
	info("UnlinkEtd: gave back the urb");
#endif

	stack_DecDeviceUse(dev);
	return 0;
}

void
otg_hc_UnlinkAllEtd(otg_hc_t *hc, stack_device_t *dev, OS_ListDeclare(*list))
{
	OS_ListDeclare(*h);
	otg_etd_t *etd;
	stack_urb_t *urb;

	/* Have to do from first to last because of ISO ETDs */
	for (h = OS_ListNext(list); h != list; )
	{
#ifdef DEBUG_UNLINK
	info("UnlinkAllEtd: h=%p", h);
#endif
		etd = OS_ListEntry(h, otg_etd_t, formatEtdList);
#ifdef DEBUG_UNLINK
	info("UnlinkAllEtd: etd=%p", etd);
#endif
		h = OS_ListNext(h);
		if (otg_etd_IsUnlink(etd))
		{
#ifdef DEBUG_UNLINK
	info("UnlinkAllEtd: already unlinked etd=%p", etd);
#endif
			continue;
		}

		urb = etd->urb;
		if (urb)
		{
			if (dev && !stack_urb_IsDevice(urb, dev))
			{
#ifdef DEBUG_UNLINK
	info("UnlinkAllEtd: dev=%p urb->dev=%p", dev, stack_urb_GetDevice(urb));
#endif
				continue;
			}
		}

		otg_hc_UnlinkEtd(hc, etd);
	}
}

int
otg_hc_WaitStart(otg_hc_t *hc)
{
	int wait;

	/* We need to avoid dead loop due to hardware failure */
	wait = 0;
	while (!otg_hc_IsIOReady(hc))
	{
#ifdef SIM_OTG
		/* This code is for debugging */
		OS_Wait(1);
#endif
		wait++;
		if (wait > 10000)
		{
			err("WaitStart: hardware not responding");
			return -1;
		}
	}

	return 0;
}

#ifdef USE_KERNEL_THREAD
static int
otg_hc_Thread(void *nothing)
{
	lock_kernel();
	daemonize();
	strcpy(current->comm, "khcd");

	do
	{
		unlock_kernel();
		OS_WaitQueueSleep(&khcdWait);
		
		otg_hc_DoDoneQueue();

		lock_kernel();
	} while (!signal_pending(current));

	unlock_kernel();
	info("Thread: khcd exiting");
	OS_CompletionSignal(&khcdExited, 0);
}

void
otg_hc_DoDoneQueue(void)
{
	otg_hc_t *hc;
	unsigned long flags;

	if (!hostController)
	{
		err("DoDoneQueue: no host controller");
		return;
	}

	hc = hostController;

	otg_LockChipIntr(hc->otg, flags);

#ifdef DEBUG_OTG_CHIP
info("DoDoneQueue: ");
#endif
	otg_hc_etd_DoDoneQueue(&hc->etd_bank);
	otg_hc_DoWaitingEtd(hc);

	otg_UnlockChipIntr(hc->otg, flags);
}
#endif

#ifdef USE_TASKLET
void
otg_hc_Tasklet(unsigned long nothing)
{
	otg_hc_t *hc;
	unsigned long flags;

	if (!hostController)
	{
		err("Tasklet: no host controller");
		return;
	}

	hc = hostController;
	while (taskletWaitCount)
	{
#ifdef DEBUG_TASKLET
info("Tasklet: wait count = %d", taskletWaitCount);
#endif
		OS_DisableInterrupts(flags);
#ifdef DO_PROFILING
	otg_TimeStart(hc->otg);
#endif
		taskletWaitCount = 0;

		otg_hc_etd_DoDoneQueue(&hc->etd_bank);
		otg_hc_DoWaitingEtd(hc);
#ifdef DO_PROFILING
	otg_TimeStop(hc->otg);
#endif

		OS_EnableInterrupts(flags);
	}
#ifdef DEBUG_TASKLET
info("Tasklet: exiting");
#endif
}
#endif

#ifdef OTG_ISO_SHARE_MEMORY
void
otg_hc_DoSofIntr(otg_hc_t *hc)
{
	OS_ListDeclare(*list);
	otg_etd_t *etd;
	int rc;

	list = &hc->sofList;
	list = OS_ListNext(list);
	while(list != &hc->sofList)
	{
		etd = OS_ListEntry(list, otg_etd_t, sofList);
		list = OS_ListNext(list);
#ifdef DEBUG_OTG_SUBMIT
info("DoSofIntr: etd=%p", etd);
#endif
		rc = otg_td_iso_DoPreDownload(etd, &hc->etd_bank);
		if (rc)
		{
#ifdef DEBUG_OTG_SUBMIT
info("DoSofIntr: delete etd from sofList");
#endif
			/* etd data is ready, delete it from sof list */
			OS_ListDelete(&etd->sofList);
			OS_ListInit(&etd->sofList);
			otg_DisableSofIntr(hc->otg);
			if (rc < 0)
			{
				otg_etd_SetInDoneQueue(etd);
#ifdef DEBUG_ISO
info("DoSofIntr: Start frame passed start_frame=0x%0x frame=0x%0x",
	stack_urb_GetStartFrame(etd->urb), hc->frameNo);
#endif
#ifdef USE_KERNEL_THREAD
				OS_WaitQueueWakeup(&khcdWait);
#else
#ifdef USE_TASKLET
				taskletWaitCount++;
#ifdef DEBUG_TASKLET
info("Tasklet scheduled... (SOF)");
#endif
				OS_TaskletSchedule(&hcTasklet);
#else
				otg_hc_etd_DoDoneQueue(&hc->etd_bank);
#endif
#endif
			}
		}
	}
}
#endif

int
otg_hc_LinkIso(otg_hc_t *hc, otg_etd_t *etd, stack_urb_t *urb)
{
	OS_ListDeclare(*h);
	otg_etd_t *first;
	otg_etd_t *last;
	stack_pipe_t pipe;
	int diff;
	int rc;
	otg_etd_t *e;
	
#ifdef DEBUG_OTG_SUBMIT
info("LinkIso:");
#endif
	pipe = stack_urb_GetPipe(urb);

	if (!stack_urb_GetBuffer(urb) || !stack_urb_GetBufferLength(urb))
	{
		return -EINVAL;
	}

	first = 0;
	last = 0;
	h = &hc->isoList;
	for(h = OS_ListPrev(h); h != &hc->isoList; h = OS_ListPrev(h))
	{
		otg_etd_t *tmp;

		tmp = OS_ListEntry(h, otg_etd_t, formatEtdList);
		if (otg_etd_IsUnlink(tmp))
		{
			continue;
		}

		if (tmp->urb == urb)
		{
			err("Return -ENXIO because tmp->urb == urb.");
			return -ENXIO;
		}

#ifdef DEBUG_OTG_SUBMIT
info("LinkIso: tmp=%p urb=%p dev=%p",
	tmp, tmp->urb, stack_urb_GetDevice(tmp->urb));
#endif
		if  (
				stack_urb_GetDevice(tmp->urb) == stack_urb_GetDevice(urb) &&
				stack_urb_GetPipe(tmp->urb) == pipe &&
				!otg_etd_IsInDoneQueue(tmp)
			)
		{
			first = tmp;
			if (!last)
			{
				last = tmp;
			}
		}
	}

#ifdef DEBUG_ISO
info("LinkIso: frame=0x%0x", hc->frameNo);
#endif
	if (stack_urb_IsIsoASAP(urb))
	{
		if (last)
		{
			otg_etd_LinkEtd(etd, first);
		}

		if (last && stack_urb_GetStatus(last->urb) == -EINPROGRESS)
		{
#ifdef DEBUG_OTG_SUBMIT
info("LinkIso: find last=%p", last);
#endif
			stack_urb_SetNextStartFrame(urb, last->urb);

			/* If start_frame passed already */
			diff = (hc->frameNo + 2 - stack_urb_GetStartFrame(urb)) & otg_hc_FmMask_c;
			if (diff < 512)
			{
#if 0
//FB: This seems to be a recoverable error. It happens often, so printing
//an error message each time slows things to a crawl.
				err("LinkIso: gap start_frame=0x%0x current_frame=0x%0x",
					stack_urb_GetStartFrame(urb), hc->frameNo);
#endif
				stack_urb_SetStartFrame(urb, hc->frameNo + 3);
			}
		}
		else
		{
			stack_urb_SetStartFrame(urb, hc->frameNo + 10);
		}

		stack_urb_MaskStartFrame(urb, otg_hc_FmMask_c);
	}
	else
	{
		stack_urb_MaskStartFrame(urb, otg_hc_FmMask_c);
		diff = ((hc->frameNo + 2) - stack_urb_GetStartFrame(urb)) & otg_hc_FmMask_c;
		if (diff < 512)
		{
			return -EAGAIN;
		}
	}

#ifdef DEBUG_ISO
info("LinkIso: start_frame=0x%0x", stack_urb_GetStartFrame(urb));
#endif
#ifdef DEBUG_MEMORY
if (isoPreStartFrame == stack_urb_GetStartFrame(urb))
{
	info("LinkIso: stop ");
	memoryError = 1;
}
isoPreStartFrame = stack_urb_GetStartFrame(urb);
#endif
	OS_ListAddTail(&etd->formatEtdList, &hc->isoList);
	if (stack_urb_GetPacketCount(urb) > otg_td_iso_MaxPackets)
	{
		int count;

		if (!first)
		{
			first = etd;
		}

		/* Need to add ETDs for packets more than 8 */
		count = otg_td_iso_MaxPackets;
		rc = 0;
		while (count < stack_urb_GetPacketCount(urb))
		{
			e = otg_etd_New(hc);
			if (!e)
			{
				rc = -ENOMEM;
				break;
			}

			otg_td_iso_SetStart(e, count);
			OS_ListAddTail(&e->formatEtdList, &hc->isoList);
			otg_etd_LinkEtd(e, first);

			count += otg_td_iso_MaxPackets;
		}

		if (rc)
		{
			otg_hc_DeleteEtdForUrb(hc, etd);
			return rc;
		}
	}

#ifdef OTG_ISO_SHARE_MEMORY
	if (last)
	{
		/* Need to set start frame, in case interrupted earlier */
		otg_td_iso_SetStartFrame(etd, stack_urb_GetStartFrame(urb));
		OS_ListAddTail(&etd->sofList, &hc->sofList);
		otg_EnableSofIntr(hc->otg);
	}

	if (stack_urb_GetPacketCount(urb) > otg_td_iso_MaxPackets)
	{
		e = etd;
		while (OS_ListNext(&e->formatEtdList) != &hc->isoList)
		{
			e = OS_ListEntry(OS_ListNext(&e->formatEtdList), otg_etd_t, formatEtdList);
			otg_td_iso_SetStartFrame(e, stack_urb_GetStartFrame(urb));
			OS_ListAddTail(&e->sofList, &hc->sofList);
			otg_EnableSofIntr(hc->otg);
		}
	}
#endif

#ifdef DEBUG_OTG_SUBMIT
info("LinkIso: start frame=0x%0x current=0x%0x",
	stack_urb_GetStartFrame(urb), hc->frameNo);
#endif
	return 0;
}

#ifdef DEBUG_MEMORY
void
otg_hc_Stop(void)
{
	info("Stop: ");
	memoryError = 1;
}
#endif

void
otg_hc_DeleteEtdForUrb(otg_hc_t *hc, otg_etd_t *etd)
{
	otg_etd_t *e;

	e = OS_ListEntry(OS_ListPrev(&hc->isoList), otg_etd_t, formatEtdList);
	while (e != etd)
	{
		otg_etd_Delete(e);
		e = OS_ListEntry(OS_ListPrev(&hc->isoList), otg_etd_t, formatEtdList);
	}
}

otg_etd_t *
otg_hc_GetFirstInPipe(otg_hc_t *hc, otg_etd_t *etd)
{
	OS_ListDeclare(*h);
	OS_ListDeclare(*list);
	stack_pipe_t pipe;
	stack_device_t *dev;
	int index;
	otg_etd_t *e;

	pipe = stack_urb_GetPipe(etd->urb);
	dev = stack_urb_GetDevice(etd->urb);
	if (otg_etd_IsTransferType(etd, otg_etd_Format_Iso_c))
	{
		h = &hc->isoList;
	}
	else if (otg_etd_IsTransferType(etd, otg_etd_Format_Ctrl_c))
	{
		h = &hc->ctrlList;
	}
	else if (otg_etd_IsTransferType(etd, otg_etd_Format_Bulk_c))
	{
		h = &hc->bulkList;
	}
	else
	{
		h = &hc->intrList[0];
		for(index = 7; index >= 0; index--)
		{
			if (stack_urb_GetIntrInterval(etd->urb) == (1 << index))
			{
				h = &hc->intrList[index];
			}
		}
	}

	list = h;
	while (h != OS_ListNext(list))
	{
		e = OS_ListEntry(OS_ListNext(list), otg_etd_t, formatEtdList);
#ifdef DEBUG_DONEQUEUE
	if (done_queue_found)
	{
		info("GetFirst: etd=%p buffer=0x%0x urb=%p",
			e, e->buffer, e->urb);
	}
#endif
		if (stack_urb_GetPipe(e->urb) == pipe && stack_urb_IsDevice(e->urb, dev))
		{
			return e;
		}

		list = OS_ListNext(list);
	}

	return 0;
}

#ifdef CONFIG_PM
void
otg_hc_GoSuspend(otg_hc_t *hc)
{
	otg_hc_GoState(hc, otg_hc_USBSuspend_c);
}

void
otg_hc_GoResume(otg_hc_t *hc)
{
	otg_hc_GoState(hc, otg_hc_USBResume_c);
}

void
otg_hc_GoStandby(otg_hc_t *hc)
{
	otg_hc_GoState(hc, otg_hc_USBSuspend_c);
}

#endif
