/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_linux.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 driver host interface to linux.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/interrupt.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/pm.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>

#define VERSTR "Revision: 0.95 $ time " __TIME__ " " __DATE__

#include "os.h"
#include "stack_urb.h"
#include "otg_linux.h"
#include "otg.h"

#define DRIVER_AUTHOR "TransDimension Inc."
#define DRIVER_DESC "TDI OTG243 Linux Host Controller Driver"

#ifdef CONFIG_PM
static struct pm_dev *otg_pm_dev = 0;
#endif

static otg_hcd_t *hcd = NULL;

static int otg_hcd_UnlinkUrb(stack_urb_t *urb)
{
	otg_hcd_t *s;
#ifdef DEBUG_OTG_CHIP
info("otg_hcd_UnlinkUrb called for %p",urb);
#endif
	if (!urb || !urb->dev || !urb->dev->bus)
	{
		return -EINVAL;
	}
	
	s = (otg_hcd_t*)urb->dev->bus->hcpriv;

	return otg_hc_UnlinkUrb(s->hc, urb);
}

static int otg_hcd_AllocDev(stack_device_t *usb_dev)
{
	return 0;
}

static int otg_hcd_FreeDev(stack_device_t *usb_dev)
{
	otg_hcd_t *s;

	if(!usb_dev || !usb_dev->bus || !usb_dev->bus->hcpriv)
	{
		return -EINVAL;
	}
	
	s=(otg_hcd_t*) usb_dev->bus->hcpriv;	
	otg_hc_UnlinkAll(s->hc, usb_dev);

	return 0;
}

static int otg_hcd_GetFrameNumber(stack_device_t *usb_dev)
{
	otg_hcd_t *s;

	s=(otg_hcd_t*) usb_dev->bus->hcpriv;	
	return otg_hc_ReadFrameNo(s->hc);
}

static int otg_hcd_SubmitUrb(stack_urb_t *urb)
{
	otg_hcd_t *s;

	if (!urb->dev || !urb->dev->bus)
	{
		return -ENODEV;
	}

	s = (otg_hcd_t *)urb->dev->bus->hcpriv;

	return otg_hc_Submit(s->hc, urb);
}

struct usb_operations otg_hcd_DeviceOperations =
{
	otg_hcd_AllocDev,
	otg_hcd_FreeDev,
	otg_hcd_GetFrameNumber,
	otg_hcd_SubmitUrb,
	otg_hcd_UnlinkUrb
};

static void otg_hcd_ResetHc(otg_hcd_t *s)
{
#ifdef DEBUG_OTG_CHIP
info("otg_hcd_ResetHc: s=%p", s);
#endif
	otg_hc_Reset(s->hc);
}

static int otg_hcd_StartHc(otg_hcd_t *s)
{
	int rc;

	rc = otg_hc_Start(s->hc);
	if (rc)
	{
		err("OTG start failed");
		return rc;
	}

	return 0;
}

void otg_hcd_Free(otg_hcd_t *s)
{
	if (s->hc)
	{
		otg_hc_Delete(s->hc);
	}

	OS_Free(s);
	hcd = 0;
}

static void otg_hcd_Remove(otg_hcd_t *s)
{
	stack_device_t *root_hub = s->bus->root_hub;

	if (root_hub)
	{
		usb_disconnect(&root_hub);
	}

	otg_hcd_ResetHc(s);

	usb_deregister_bus(s->bus);

	usb_free_bus(s->bus);
}

static int __init otg_hcd_Start(otg_hcd_t *s)
{				/* start it up */
	/* connect the virtual root hub */
	stack_device_t *usb_dev;

	usb_dev = usb_alloc_dev(NULL, s->bus);
	if (!usb_dev)
	{
		return -1;
	}

	s->bus->root_hub = usb_dev;
	usb_connect(usb_dev);

	if (usb_new_device(usb_dev) != 0)
	{
		usb_free_dev(usb_dev);
		return -1;
	}

	return 0;
}


static int __devinit otg_hcd_Alloc(void)
{


	otg_hcd_t *s;                       /* host mgm't data structure */
	stack_bus_t *bus;             /* bus info */     

	/* allocate memory to store data pertaining host */

	s = OS_Malloc(sizeof(otg_hcd_t));
	if (!s)
	{
		return -1;
	}

	/* initialize host data structure */
	OS_MemSet(s, 0, sizeof(otg_hcd_t));

	s->hc = otg_hc_New();
	if(!s->hc)
	{
		otg_hcd_Free(s);
		return -1;
	}

	bus = usb_alloc_bus(&otg_hcd_DeviceOperations);
	if (!bus)
	{
		otg_hcd_Free(s);
		return -1;
	}

	s->bus = bus;
	bus->hcpriv = s;

	otg_hcd_ResetHc(s);

	usb_register_bus(s->bus);

	if (otg_hcd_StartHc(s))
	{
		otg_hcd_Remove(s);
		otg_hcd_Free(s);
		return -1;
	}

	hcd = s;

	if(otg_hcd_Start(s) < 0)
	{
		otg_hcd_Remove(s);
		otg_hcd_Free(s);
		return -1;
	}

	return 0;
}

#ifdef CONFIG_PM
static int
otg_linux_CallBackForPM(struct pm_dev *pm_dev, pm_request_t req, void *data)
{
	otg_hc_t *hc;

	if (!hcd)
	{
		/* no host controller */
		return 0;
	}

	hc = hcd->hc;
	switch (req)
	{
		case PM_SUSPEND:
			info("OTG243 is going to suspend mode");
			otg_hc_GoSuspend(hc);
			break;
		case PM_RESUME:
			info("OTG243 is resumed");
			otg_hc_GoResume(hc);
			break;
/*
		case PM_STANDBY:
			info("OTG243 standby called");
			otg_hc_GoStandby(hc);
			break;
*/
		default:
			info("OTG243 power management called with %d", req);
			break;
	}

	return 0;
}
#endif

static int __init otg_hcd_init(void) 
{
	int retval = 0;

#ifdef CONFIG_PPC850_HANWANG
       ppc850_memory_control();
#endif

	info(VERSTR);
	info(DRIVER_DESC);

#ifdef SIM_OTG
	otg_chip_Init();
#ifndef OTG_FUNCTION_ONLY
	retval = otg_hcd_Alloc();
#endif
#else
#ifndef OTG_FUNCTION_ONLY
	retval = otg_hcd_Alloc();
#endif
#endif /* #ifdef SIM_OTG */

	
#ifdef CONFIG_PM
	if (!otg_pm_dev)
	{
		otg_pm_dev = pm_register(PM_SYS_DEV, 0, otg_linux_CallBackForPM);
	}
#endif
	return retval;
}

void __exit otg_hcd_cleanup(void) 
{      
#ifdef CONFIG_PM
	if (otg_pm_dev)
	{
		pm_unregister(otg_pm_dev);
		otg_pm_dev = 0;
	}
#endif

#ifndef OTG_FUNCTION_ONLY
	/* Remove various memory allocations and clear data structs. */
	otg_hcd_Remove(hcd);

	/* Make sure this function do not use chip memory */
	otg_hcd_Free(hcd);
#endif

#ifdef SIM_OTG
	otg_chip_Exit();
#endif
}

stack_bus_t *
otg_hcd_GetUsbBus(void)
{
	if (!hcd)
	{
		return 0;
	}

	return hcd->bus;
}

otg_hc_t *otg_hcd_GetHostController(void)
{
	if (!hcd)
	{
		return 0;
	}

	return hcd->hc;
}

#ifdef SIM_OTG
otg_t *otg_hcd_GetDevice(void)
{
	return hcd->hc->otg;
}

ISR otg_hcd_GetISR(void)
{
	return otg_DoIntr;
}
#endif

module_init(otg_hcd_init);
module_exit(otg_hcd_cleanup);

MODULE_AUTHOR( DRIVER_AUTHOR );
MODULE_DESCRIPTION( DRIVER_DESC );

#ifndef MODULE
#ifdef SIM_OTG
EXPORT_SYMBOL(otg_hcd_GetDevice);
EXPORT_SYMBOL(otg_hcd_GetISR);
#endif

EXPORT_SYMBOL(otg_Delete);

#ifndef OTG_FUNCTION_ONLY
EXPORT_SYMBOL(otg_hc_New);
EXPORT_SYMBOL(otg_hc_Delete);
EXPORT_SYMBOL(otg_hcd_GetUsbBus);
EXPORT_SYMBOL(otg_hcd_GetHostController);
#endif

#ifndef OTG_HOST_ONLY
EXPORT_SYMBOL(otg_fc_SetDeviceDesc);
EXPORT_SYMBOL(otg_fc_SetConfigDesc);
EXPORT_SYMBOL(otg_fc_SetStrDesc);
EXPORT_SYMBOL(otg_fc_Submit);
EXPORT_SYMBOL(otg_fc_SetCallBack);
EXPORT_SYMBOL(otg_urb_Init);
EXPORT_SYMBOL(otg_urb_New);
EXPORT_SYMBOL(otg_urb_Delete);
EXPORT_SYMBOL(otg_fc_GetConfiguration);
EXPORT_SYMBOL(otg_fc_New);
EXPORT_SYMBOL(otg_fc_Delete);
EXPORT_SYMBOL(otg_fc_Reset);
EXPORT_SYMBOL(otg_fc_Exit);
EXPORT_SYMBOL(otg_fc_Start);
EXPORT_SYMBOL(otg_fc_Stop);
EXPORT_SYMBOL(otg_fc_RequestBus);
EXPORT_SYMBOL(otg_fc_ReleaseBus);
EXPORT_SYMBOL(otg_fc_Stall);
EXPORT_SYMBOL(otg_fc_ClearStall);
#endif

#ifdef DEBUG_OTG_CHIP
EXPORT_SYMBOL(otg_GetRegName);
#endif
#endif
