/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_mem.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 on chip memory object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

void
otg_mem_Init(otg_mem_t *mem, int begin, int length)
{
	int limit;
	int i;
	otg_mem_block_t *block;
	int start;
	int size;

	start = (begin + otg_mem_BlockSize_c - 1) & ~(otg_mem_BlockSize_c - 1);
	size = length & ~(otg_mem_BlockSize_c - 1);

#ifdef DEBUG_OTG_CHIP
info("Init: start=0x%0x size=0x%0x", start, size);
#endif
	limit = otg_mem_Start_c + otg_mem_Size_c;
	if (start < otg_mem_Start_c || start >= limit)
	{
		err("Out of range of start %d", start);
		return;
	}

	if (size < otg_mem_BlockSize_c || size > otg_mem_Size_c)
	{
		err("bad size %d", size);
		return;
	}

	mem->start = start;
	mem->blocks = (size >> otg_mem_BlockMagnitude_c) - 1;

	block = mem->mem_blocks;
	for (i = 0; i <= mem->blocks; i++)
	{
		otg_mem_block_Init(block);
		block++;
	}

	otg_mem_block_SetBlocks(mem->mem_blocks, mem->blocks);
	otg_mem_block_SetBlocks(&mem->mem_blocks[mem->blocks], mem->blocks);
}

static otg_mem_block_t *
otg_mem_FindFirstFit(otg_mem_t *mem, int blks, int backward)
{
	otg_mem_block_t *block;
	otg_mem_block_t *end;

	if (!backward)
	{
		block = mem->mem_blocks;
		end = block + mem->blocks;
		while (block <= end)
		{
			if (!otg_mem_block_GetUsed(block) &&
				(otg_mem_block_GetBlocks(block) >= blks))
			{
				return block;
			}

			block += otg_mem_block_GetBlocks(block) + 1;
		}
	}
	else
	{
		end = mem->mem_blocks;
		block = end + mem->blocks;
		while (block >= end)
		{
			if (!otg_mem_block_GetUsed(block) &&
				(otg_mem_block_GetBlocks(block) >= blks))
			{
				return block;
			}

			block -= otg_mem_block_GetBlocks(block) + 1;
		}
	}

	return 0;
}

int
otg_mem_Allocate(otg_mem_t *mem, int size)
{
	otg_mem_block_t *found_block;
	otg_mem_block_t *old;
	int blks;
	int new_size;

	blks = otg_mem_ByteToBlock(size) - 1;
	if (size > 64)
	{
		found_block = otg_mem_FindFirstFit(mem, blks, 0);
	}
	else
	{
		found_block = otg_mem_FindFirstFit(mem, blks, 1);
	}

#if defined DEBUG_OTG_CHIP || defined DEBUG_MEM
info("Allocate: mem=%p size=0x%0x => block=%p", mem, size, found_block);
#endif

#ifdef DEBUG_SOME_OTG
	{
		otg_mem_block_t *block;
		otg_mem_block_t *end;
		end = mem->mem_blocks;
		block = end + mem->blocks;
		while (block >= end)
		{
			info("block=%p: used:%d, blocks=%d", block, otg_mem_block_GetUsed(block), otg_mem_block_GetBlocks(block));

			block -= otg_mem_block_GetBlocks(block) + 1;
		}
	}
#endif

	if (!found_block)
	{
		err("Can't allocate host controller memory of size %d (blks %d)", size, blks);

		WARN_ON(1);

		return otg_OUT_OF_DM;
	}

	if (size > 64)
	{
		/* Forward allocating */
		if (otg_mem_block_GetBlocks(found_block) != blks)
		{
			/* split the block */
			new_size = otg_mem_block_GetBlocks(found_block) - blks - 1;
			old = found_block + otg_mem_block_GetBlocks(found_block);
			otg_mem_block_SetBlocks(old, new_size);
			otg_mem_block_SetBlocks(found_block + blks + 1, new_size);
			otg_mem_block_ClearUsed(found_block + blks + 1);

			otg_mem_block_SetBlocks(found_block, blks);
			otg_mem_block_SetBlocks(found_block + blks, blks);
		}

		otg_mem_block_SetUsed(found_block);
		otg_mem_block_SetUsed(found_block + blks);
#if defined DEBUG_OTG_CHIP || defined DEBUG_MEM
info("Allocate: found block 0x%0x", found_block - mem->mem_blocks);
#endif
		return otg_mem_BlockToAddr(mem, found_block - mem->mem_blocks);
	}
	else
	{
		/* Backward allocating */
		if (otg_mem_block_GetBlocks(found_block) != blks)
		{
			new_size = otg_mem_block_GetBlocks(found_block) - blks - 1;
			old = found_block - otg_mem_block_GetBlocks(found_block);
			otg_mem_block_SetBlocks(old, new_size);
			otg_mem_block_SetBlocks(found_block - blks - 1, new_size);
			otg_mem_block_ClearUsed(found_block - blks - 1);

			otg_mem_block_SetBlocks(found_block, blks);
			otg_mem_block_SetBlocks(found_block - blks, blks);
		}

		otg_mem_block_SetUsed(found_block);
		otg_mem_block_SetUsed(found_block - blks);
#if defined DEBUG_OTG_CHIP || defined DEBUG_MEM
info("Allocate: found block 0x%0x", found_block - mem->mem_blocks - blks);
#endif
		return otg_mem_BlockToAddr(mem, found_block - mem->mem_blocks - blks);
	}
}

void
otg_mem_Release(otg_mem_t *mem, int start)
{
	otg_mem_block_t *block;
	otg_mem_block_t *other;
	int blks;
	int new_blks;

#if defined DEBUG_OTG_CHIP || defined DEBUG_MEM
info("Release: mem=%p, block=%p", mem, otg_mem_AddrToBlock(mem, start));
#endif

	block = otg_mem_AddrToBlock(mem, start);
	if (!otg_mem_block_GetUsed(block))
	{
		err("Try to deallocate unallocated memory chunk %d", start);
		return;
	}

	blks = otg_mem_block_GetBlocks(block);
	otg_mem_block_ClearUsed(block);

	if (block > mem->mem_blocks)
	{
		other = block - 1;
		new_blks = otg_mem_block_GetBlocks(other) + blks + 1;
		if (!otg_mem_block_GetUsed(other))
		{
			other -= otg_mem_block_GetBlocks(other);
			otg_mem_block_SetBlocks(other, new_blks);
			otg_mem_block_SetBlocks(block + blks, new_blks);
			block = other;
			blks = new_blks;
		}
	}

	block += blks;
	otg_mem_block_ClearUsed(block);
	if (block < (mem->mem_blocks + mem->blocks))
	{
		other = block + 1;
		new_blks = otg_mem_block_GetBlocks(other) + blks + 1;
		if (!otg_mem_block_GetUsed(other))
		{
			other += otg_mem_block_GetBlocks(other);
			otg_mem_block_SetBlocks(other, new_blks);
			otg_mem_block_SetBlocks(block - blks, new_blks);
		}
	}
}

int
otg_mem_Test(otg_t *otg)
{
	UINT8 *data;
	UINT8 *buffer;
	int i;

#ifdef DEBUG_NO_MEM_TEST
	return 0;
#endif
	data = OS_Malloc(otg_mem_Size_c);
	if (!data)
	{
		return -ENOMEM;
	}

	buffer = OS_Malloc(otg_mem_Size_c);
	if (!buffer)
	{
		OS_Free(data);
		return -ENOMEM;
	}

	for (i = 0; i < otg_mem_Size_c; i++)
	{
		data[i] = (1 << (i & 0x07)) & 0xFF;
	}

#ifdef DEBUG_OTG_CHIP
#ifndef OTG_FUNCTION_ONLY
	otg_mem_BulkWrite(otg_GetHost(otg), otg_mem_Start_c, data, 64);
	otg_mem_BulkRead(otg_GetHost(otg), otg_mem_Start_c, buffer, 64);
	for (i = 0; i < 64; i++)
#else
	otg_mem_FcWrite(otg_GetFunction(otg), otg_mem_Start_c, data, 64);
	otg_mem_FcRead(otg_GetFunction(otg), otg_mem_Start_c, buffer, 64);
	for (i = 0; i < 64; i++)
#endif
#else
#ifndef OTG_FUNCTION_ONLY
	otg_mem_BulkWrite(otg_GetHost(otg), otg_mem_Start_c, data, otg_mem_Size_c);
	otg_mem_BulkRead(otg_GetHost(otg), otg_mem_Start_c, buffer, otg_mem_Size_c);
	for (i = 0; i < otg_mem_Size_c; i++)
#else
	otg_mem_FcWrite(otg_GetFunction(otg), otg_mem_Start_c, data, otg_mem_Size_c);
	otg_mem_FcRead(otg_GetFunction(otg), otg_mem_Start_c, buffer, otg_mem_Size_c);
	for (i = 0; i < otg_mem_Size_c; i++)
#endif
#endif
	{
		if (data[i] != buffer[i])
		{
info("Test: i=%d data=0x%0x buffer=0x%0x", i, data[i], buffer[i]);
			OS_Free(buffer);
			OS_Free(data);
			return -1;
		}
	}

	for (i = 0; i < otg_mem_Size_c; i++)
	{
		data[i] = ~(1 << (i & 0x07));
	}

#ifdef DEBUG_OTG_CHIP
#ifndef OTG_FUNCTION_ONLY
	otg_mem_BulkWrite(otg_GetHost(otg), otg_mem_Start_c, data, 64);
	otg_mem_BulkRead(otg_GetHost(otg), otg_mem_Start_c, buffer, 64);
	for (i = 0; i < 64; i++)
#else
	otg_mem_FcWrite(otg_GetFunction(otg), otg_mem_Start_c, data, 64);
	otg_mem_FcRead(otg_GetFunction(otg), otg_mem_Start_c, buffer, 64);
	for (i = 0; i < 64; i++)
#endif
#else
#ifndef OTG_FUNCTION_ONLY
	otg_mem_BulkWrite(otg_GetHost(otg), otg_mem_Start_c, data, otg_mem_Size_c);
	otg_mem_BulkRead(otg_GetHost(otg), otg_mem_Start_c, buffer, otg_mem_Size_c);
	for (i = 0; i < otg_mem_Size_c; i++)
#else
	otg_mem_FcWrite(otg_GetFunction(otg), otg_mem_Start_c, data, otg_mem_Size_c);
	otg_mem_FcRead(otg_GetFunction(otg), otg_mem_Start_c, buffer, otg_mem_Size_c);
	for (i = 0; i < otg_mem_Size_c; i++)
#endif
#endif
	{
		if (data[i] != buffer[i])
		{
info("Test: i=%d data=0x%0x buffer=0x%0x", i, data[i], buffer[i]);
			OS_Free(buffer);
			OS_Free(data);
			return -1;
		}
	}

	OS_Free(buffer);
	OS_Free(data);
	return 0;
}

void
otg_mem_BulkWrite(otg_hc_t *hc, int hc_mem, void *buffer, int len)
{
	int i;
	int addr;
	void *reg;
	otg_t *otg;
#ifdef OTG_16_BIT_DATA_BUS
	UINT16 w;
#else
	UINT32 w;
#endif

#ifdef DEBUG_OTG_SUBMIT
info("BulkWrite: hc_mem=0x%0x buffer=%p len=0x%0x otg_hc_WaitStart(hc)=%d", hc_mem, buffer, len, otg_hc_WaitStart(hc));
#endif
	/* Copy data from memory to host controller */
	addr = (hc_mem >> 2) | otg_hc_StartData_c | otg_hc_StartWrite_c;

	otg_hc_WriteStart(hc, addr);
	otg = otg_hc_GetOtg(hc);
	reg = otg_PIO_dataPort_c + otg->mem_map;

	if (otg_hc_WaitStart(hc))
	{
		return;
	}

#ifdef OTG_16_BIT_DATA_BUS
	if ((int)buffer & 0x1)
	{
		/* Buffer is not Word aligned */
		UINT8 *dp;

		dp = (UINT8 *)buffer;
		w = 0; /* to shutoff warning by compiler */

		for(i=0; i<len; i++)
		{
			switch (i & 1)
			{
				case 0:
#ifdef OTG_BIG_ENDIAN
					w = *dp++ << 8;
#else
					w = *dp++;
#endif
					break;
				case 1:
#ifdef OTG_BIG_ENDIAN
					w |= *dp++;
#else
					w |= (*dp++) << 8;
#endif
					otg_WriteReg16(reg, w);
					break;
			}
		}
	
		if (len & 1)
		{
			/* Last partial dword */
			otg_WriteReg16(reg, w);
		}
	}
	else
	{
		/* Buffer is Word aligned */
		UINT16 *wp;

		wp = (UINT16 *)buffer;
		for(i = 0; i < len; i+=2)
		{
#ifdef OTG_BIG_ENDIAN
			w = *wp++;
			otg_WriteReg16(reg, OS_CpuToLe16(w));
#else
			otg_WriteReg16(reg, *wp++);
#endif
		}
	}

	len &= 3;
	if (len == 2 || len == 1)
	{
		otg_WriteReg16(reg, 0);
	}
#else
	if ((int)buffer & 0x3)
	{
		/* Buffer is not DWord aligned */
		UINT8 *dp;

		dp = (UINT8 *)buffer;
		w = 0; /* to shutoff warning by compiler */

		for(i=0; i<len; i++)
		{
			switch (i & 3)
			{
				case 0:
#ifdef OTG_BIG_ENDIAN
					w = (*dp++) << 24;
#else
					w = *dp++;
#endif
					break;
				case 1:
#ifdef OTG_BIG_ENDIAN
					w |= (*dp++) << 16;
#else
					w |= (*dp++) << 8;
#endif
					break;
				case 2:
#ifdef OTG_BIG_ENDIAN
					w |= (*dp++) << 8;
#else
					w |= (*dp++) << 16;
#endif
					break;
				case 3:
#ifdef OTG_BIG_ENDIAN
					w |= *dp++;
#else
					w |= (*dp++) << 24;
#endif
					otg_WriteReg(reg, w);
			}
		}
	
		if (len & 3)
		{
			/* Last partial dword */
			otg_WriteReg(reg, w);
		}
	}
	else
	{
		/* Buffer is DWord aligned */
		UINT32 *wp;

		wp = (UINT32 *)buffer;
		for(i = 0; i < len; i+=4)
		{
#ifdef OTG_BIG_ENDIAN
			w = *wp++;
			otg_WriteReg(reg, OS_CpuToLe32(w));
#else
			otg_WriteReg(reg, *wp++);
#endif
		}
	}
#endif //OTG_16_BIT_DATA_BUS
	#ifdef DEBUG_OTG_SUBMIT
	info("BulkWrite: finished");
	#endif
}

void
otg_mem_BulkRead(otg_hc_t *hc, int hc_mem, void *buffer, int size)
{
	int i;
	int addr;
	UINT8 *dp;
	UINT32 w = 0;
#ifdef OTG_16_BIT_DATA_BUS
	UINT16 *wp;
#else
	UINT32 *wp;
#endif
	int len;
	void *reg;
	otg_t *otg;

#ifdef DEBUG_OTG_SUBMIT
info("BulkRead: hc_mem=0x%0x buffer=%p size=0x%0x",hc_mem,buffer,size);
#endif

	/* Copy data from host controller to memory */
	addr = (hc_mem >> 2) | otg_hc_StartData_c | otg_hc_StartRead_c;
	otg_hc_WriteStart(hc, addr);

	otg = otg_hc_GetOtg(hc);
	reg = otg_PIO_dataPort_c + otg->mem_map;
	if (otg_hc_WaitStart(hc))
	{
		return;
	}

#ifdef OTG_16_BIT_DATA_BUS
	if ((int)buffer & 0x1)
	{
		/* Buffer is not Word aligned */
		dp = (UINT8 *)buffer;
		len = size;
	}
	else
	{
		wp = (UINT16 *)buffer;
		/* Buffer is Word aligned */
		len = size - 1;
		for(i = 0; i < len; i += 2)
		{
#ifdef OTG_BIG_ENDIAN
			w = otg_ReadReg16(reg);
			*wp++ = OS_Le16ToCpu(w);
#else
			*wp++ = otg_ReadReg16(reg);
#endif
		}

		len = size & 0x1;
		dp = (UINT8 *)wp;
	}


	for(i=0; i<len; i++)
	{
		switch (i & 1)
		{
			case 0:
#ifdef OTG_BIG_ENDIAN
				w = otg_ReadReg16(reg);
				*dp++ = (w >> 8) & 0xFF;
#else
				w = otg_ReadReg16(reg);
				*dp++ = w & 0xFF;
#endif
				break;
			case 1:
#ifdef OTG_BIG_ENDIAN
				*dp++ = w & 0xFF;
#else
				w >>= 8;
				*dp++ = w & 0xFF;
#endif
				break;
		}
	}

	len = size & 3;
	if (size == 2 || size == 1)
	{
		otg_ReadReg16(reg);
	}
#else
	if ((int)buffer & 0x3)
	{
		/* Buffer is not DWord aligned */
		dp = (UINT8 *)buffer;
		len = size;
	}
	else
	{
		wp = (UINT32 *)buffer;
		/* Buffer is DWord aligned */
		len = size - 3;
		for(i = 0; i < len; i += 4)
		{
#ifdef OTG_BIG_ENDIAN
			w = otg_ReadReg(reg);
			*wp++ = OS_Le32ToCpu(w);
#else
			*wp++ = otg_ReadReg(reg);
#endif
		}

		len = size & 0x3;
		dp = (UINT8 *)wp;
	}


	for(i=0; i<len; i++)
	{
		switch (i & 3)
		{
			case 0:
#ifdef OTG_BIG_ENDIAN
				w = otg_ReadReg(reg);
				*dp++ = (w >> 24) & 0xFF;
#else
				w = otg_ReadReg(reg);
				*dp++ = w & 0xFF;
#endif
				break;
			case 1:
#ifdef OTG_BIG_ENDIAN
				*dp++ = (w >> 16) & 0xFF;
#endif
			case 2:
#ifdef OTG_BIG_ENDIAN
				*dp++ = (w >> 8) & 0xFF;
#endif
			case 3:
#ifdef OTG_BIG_ENDIAN
				*dp++ = w & 0xFF;
#else
				w >>= 8;
				*dp++ = w & 0xFF;
#endif
				break;
		}
	}
#endif
}

