/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_sim_device.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 function controller simulator.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg_linux.h"
#include "otg.h"
#include "otg_chip.h"
#include "otg_sim_device.h"

extern int otg_chip_HasData(void);
extern int otg_chip_HasMoreData(int bytes);
extern void otg_chip_WriteFrame(UINT8 data);
extern UINT8 otg_chip_ReadFrame(void);

otg_sim_device_t *
otg_sim_device_New(void)
{
	otg_sim_device_t *dev;
	int i;

	dev = OS_Malloc(sizeof(otg_sim_device_t));
	if (!dev)
	{
		return 0;
	}

	dev->chip = otg_chip_GetChip();
	dev->state = otg_sim_device_StateIdle;
	for(i = 0; i < 8; i++)
	{
		dev->packetDone[i] = 0;
	}

	return dev;
}

void
otg_sim_device_Delete(otg_sim_device_t *dev)
{
	if (dev)
	{
		OS_Free(dev);
	}
}

void
otg_sim_device_SetFill(int ep, int yBuffer)
{
	UINT32 w;

	w = otg_sim_device_ReadFcReg(otg_ep_FillStatus);
	if (yBuffer)
	{
		w |= 1 << (ep + 16);
	}
	else
	{
		w |= 1 << ep;
	}

	otg_sim_device_WriteFcReg(otg_ep_FillStatus, w);
}

void
otg_sim_device_ClearFill(int ep, int yBuffer)
{
	UINT32 w;

	w = otg_sim_device_ReadFcReg(otg_ep_FillStatus);
	if (yBuffer)
	{
		w &= ~(1 << (ep + 16));
	}
	else
	{
		w &= ~(1 << ep);
	}

	otg_sim_device_WriteFcReg(otg_ep_FillStatus, w);
}

int
otg_sim_device_TestFill(int ep, int yBuffer)
{
	UINT32 w;

	w = otg_sim_device_ReadFcReg(otg_ep_FillStatus);
	if (yBuffer)
	{
		return w & (1 << (ep + 16));
	}
	else
	{
		return w & (1 << ep);
	}
}

void
otg_sim_device_SkipFrame(int size)
{
	int i;

	for(i = 0; i < size; i++)
	{
		otg_chip_ReadFrame();
	}
}

void
otg_sim_device_CheckNewTransfer(otg_sim_device_t *dev, int ep)
{
	UINT32 w;

	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1));
	if (w & otg_ep_ConfigNewTransfer)
	{
		w &= ~otg_ep_ConfigNewTransfer;
		otg_sim_device_WriteFcReg(otg_ep_Config(ep - 1), w);
		if ((w & otg_ep_ConfigType) != otg_ep_ConfigTypeIso1)
		{
			dev->packetDone[ep] = 0;
		}
	}
}

static void
otg_sim_device_ReadSetup(otg_sim_device_t *dev)
{
	UINT8 b;
	int length;
	int memory;
	UINT32 w;
	int i;
	int overwrite;

	b = otg_chip_ReadFrame();
	if (b != otg_chip_TokenDATA0 && b != otg_chip_TokenDATA1)
	{
		err("ReadSetup: Expecting data token but get 0x%0x", b);
	}

	if (!otg_chip_HasMoreData(10))
	{
		err("ReadSetup: Missing data after token 0x%0x", b);
		return;
	}

	length = otg_chip_ReadFrame();
	length |= otg_chip_ReadFrame() << 8;
	if (length < 8)
	{
		err("ReadSetup: bad length %d", length);
		otg_sim_device_SkipFrame(length);
		return;
	}

	memory = otg_chip_GetReg(otg_FncRamStart_c);
#ifdef DEBUG_OTG_CHIP
	info("ReadSetup: function memory address = 0x%0x", memory);
#endif
	for(i = 0; i < 8; i++)
	{
		otg_chip_WriteMem(memory, otg_chip_ReadFrame());
		memory++;
	}

	w = otg_sim_device_ReadFcReg(otg_ep0_Ctrl);
	overwrite = 0;
	if (dev->state != otg_sim_device_StateIdle)
	{
		if (!(w & otg_ep0_CtrlAck))
		{
			overwrite = 1;
		}
	}

	w |= otg_ep0_CtrlSetup;
	w &= ~otg_ep0_CtrlAck;
	otg_sim_device_WriteFcReg(otg_ep0_Ctrl, w);

	w = otg_sim_device_ReadFcReg(otg_ep_intr_Enable);
#ifdef DEBUG_OTG_CHIP
	info("ReadSetup: ep_intr_Enable=0x%0x", w);
#endif
	if (w & otg_ep_intr_Setup)
	{
#ifdef DEBUG_OTG_CHIP
		info("ReadSetup: set Setup interrupt");
#endif
		w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
		w |= otg_ep_intr_Setup;
		otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
	}

	if (overwrite)
	{
		w = otg_sim_device_ReadFcReg(otg_ep_intr_Enable);
		if (w & otg_ep_intr_SetupOverWritten)
		{
			w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
			w |= otg_ep_intr_SetupOverWritten;
			otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
		}
	}

	otg_chip_WriteFrame(otg_chip_TokenACK);
}

static int
otg_sim_device_ReadToken(otg_sim_device_t *dev, int token)
{
	int addr;
	int ep;
	int b;
	int length;

#ifdef DEBUG_OTG_CHIP
info("ReadToken: dev=%p token=0x%0x", dev, token);
#endif
	if (!otg_chip_HasData())
	{
		err("ReadToken: Missing addr after token 0x%0x", token);
		return -1;
	}

	addr = otg_chip_ReadFrame();
	if (!otg_chip_HasData())
	{
		err("ReadToken: Missing addr after addr low 0x%0x", addr);
		return -1;
	}

	addr |= otg_chip_ReadFrame() << 8;
	if ((addr & 0x7F) == otg_sim_device_ReadFcReg(otg_ep0_DeviceAddr))
	{
		ep = (addr >> 7) & 0x0F;
	}
	else if (!addr)
	{
		ep = 0;
#ifdef DEBUG_OTG_CHIP
		info("ReadToken: zero address is used");
#endif
	}
	else
	{
		err("ReadToken: Received data for other devices addr=%d", addr);
		b = otg_chip_ReadFrame();
		length = b;
		b = otg_chip_ReadFrame();
		length |= b << 8;
		otg_sim_device_SkipFrame(length);
		ep = -1;
	}

	return ep;
}

static void
otg_sim_device_ReadEp0Out(otg_sim_device_t *dev)
{
	UINT8 b;
	int memory;
	int i;
	int len;
	UINT32 w;

	if (!otg_chip_HasData())
	{
		err("ReadEp0Out: no data in OUT transfer");
		return;
	}

	b = otg_chip_ReadFrame();
	if (b != otg_chip_TokenDATA0 && b != otg_chip_TokenDATA1)
	{
		err("ReadEp0Out: Expecting data token but get 0x%0x", b);
	}

	if (!otg_chip_HasMoreData(2))
	{
		err("ReadEp0Out: missing data after token 0x%0x", b);
		return;
	}

	len = otg_chip_ReadFrame();
	len |= otg_chip_ReadFrame() << 8;
	if (!otg_chip_HasMoreData(len))
	{
		err("ReadEp0Out: missing data after len 0x%0x", len);
		otg_sim_device_SkipFrame(len);
		return;
	}

	if (len > 8)
	{
		err("ReadEp0Out: bad len %d in OUT packet", len);
		for(i = 0; i < len; i++)
		{
#ifdef DEBUG_OTG_CHIP
			info("ReadEp0Out: ---- 0x%0x", otg_chip_ReadFrame());
#endif
		}

		return;
	}

	w = otg_sim_device_ReadFcReg(otg_ep0_Ctrl);
	if (w & otg_ep0_CtrlOut)
	{
		otg_sim_device_SkipFrame(len);
		otg_chip_WriteFrame(otg_chip_TokenNAK);
		return;
	}

	memory = otg_chip_GetReg(otg_FncRamStart_c) + 8;
	for(i = 0; i < len; i++)
	{
		otg_chip_WriteMem(memory, otg_chip_ReadFrame());
		memory++;
	}

	w = otg_sim_device_ReadFcReg(otg_ep0_Ctrl);
	w |= otg_ep0_CtrlOut;
	otg_sim_device_WriteFcReg(otg_ep0_Ctrl, w);

	w = otg_sim_device_ReadFcReg(otg_ep_intr_Enable);
	if (w & otg_ep_intr_EP0Out)
	{
		w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
		w |= otg_ep_intr_EP0Out;
		otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
	}

	otg_chip_WriteFrame(otg_chip_TokenACK);
}

static void
otg_sim_device_ReadEpnOut(otg_sim_device_t *dev, int ep)
{
	int b;
	UINT32 w;
	int i;
	int memory;
	int len;
	int size;
	int pc;
	int packets;
	int yBuffer;
	int iso1;

	if (ep > 7)
	{
		err("ReadEpnOut: bad endpoint %d", ep);
		return;
	}

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnOut: dev=%p ep=%d", dev, ep);
#endif
	b = otg_chip_ReadFrame();
	if (b != otg_chip_TokenDATA0 && b != otg_chip_TokenDATA1)
	{
		err("ReadEpnOut: Expecting data token but get 0x%0x", b);
	}

	if (!otg_chip_HasMoreData(2))
	{
		err("ReadEpnOut: missing data after token 0x%0x", b);
		return;
	}

	size = otg_chip_ReadFrame();
	size |= otg_chip_ReadFrame() << 8;
	if (!otg_chip_HasMoreData(size))
	{
		err("ReadEpnOut: missing data after size 0x%0x", size);
		return;
	}

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnOut: size=%d", size);
#endif
	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1));
	if (!(w & otg_ep_ConfigEnable))
	{
		/* Endpoint is not enabled yet, skip and send NAK */
		otg_sim_device_SkipFrame(size);
		return;
	}

	otg_sim_device_CheckNewTransfer(dev, ep);

	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1));
	len = (w & otg_ep_ConfigMaxPacketSize) >> otg_ep_ConfigMaxPacketSizeLoc;
	if ((w & otg_ep_ConfigType) == otg_ep_ConfigTypeIso1)
	{
		iso1 = 1;
	}
	else
	{
		iso1 = 0;
	}

	if (size > len)
	{
		err("ReadEpnOut: bad len %d in OUT packet", size);
		otg_sim_device_SkipFrame(size);
		return;
	}

	w = otg_sim_device_ReadFcReg(otg_ep_Ctrl(ep - 1));
	if (iso1)
	{
		packets = 1;
	}
	else
	{
		packets = (w & otg_ep_CtrlPacketPerBuffer) >> otg_ep_CtrlNPLoc;
	}

	memory = otg_sim_device_ReadFcReg(otg_ep_Buffer(ep - 1));
	yBuffer = 0;
	if (((int)(dev->packetDone[ep] / packets)) & 1)
	{
		memory >>= otg_ep_BufferYLoc;
		yBuffer = 1;
	}

	memory &= otg_ep_BufferBuffer;
	memory += (dev->packetDone[ep] % packets) * len;

	if (otg_sim_device_TestFill(ep, yBuffer))
	{
		otg_sim_device_SkipFrame(size);
		otg_chip_WriteFrame(otg_chip_TokenNAK);
		return;
	}

	for(i = 0; i < size; i++)
	{
		otg_chip_WriteMem(memory, otg_chip_ReadFrame());
		memory++;
	}

	if (iso1)
	{
		if (yBuffer)
		{
			w &= ~otg_ep_CtrlLength1;
			w |= size << 16;
		}
		else
		{
			w &= ~otg_ep_CtrlLength0;
			w |= size;
		}

		pc = 1;
	}
	else
	{
		pc = w & otg_ep_CtrlPacketCount;
		pc--;
		w &= ~otg_ep_CtrlPacketCount;
		w |= pc;
		if (!pc)
		{
			w &= ~otg_ep_CtrlLastPacketSize;
			w |= size << otg_ep_CtrlLPSLoc;
		}
	}

	otg_sim_device_WriteFcReg(otg_ep_Ctrl(ep - 1), w);

	dev->packetDone[ep]++;

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnOut: pc=%d", pc);
#endif

	if (!pc)
	{
		w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
		w |= otg_ep_intr_CTRLn(ep) | otg_ep_intr_EPn(ep);
		otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);

		otg_sim_device_SetFill(ep, yBuffer);
	}
	else if (!(dev->packetDone[ep] % packets))
	{
		if (otg_sim_device_ReadFcReg(otg_ep_intr_Enable) & otg_ep_intr_EPn(ep))
		{
			w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
			w |= otg_ep_intr_EPn(ep);
			otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
		}

		otg_sim_device_SetFill(ep, yBuffer);
	}

	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1)) & otg_ep_ConfigType;
	if (w != otg_ep_ConfigTypeIso0 && w != otg_ep_ConfigTypeIso1)
	{
		otg_chip_WriteFrame(otg_chip_TokenACK);
	}
}

static void
otg_sim_device_ReadEp0In(otg_sim_device_t *dev)
{
	UINT32 w;
	int i;
	int memory;

	w = otg_sim_device_ReadFcReg(otg_ep0_Ctrl);
	if (!(w & otg_ep0_CtrlIn))
	{
		otg_chip_WriteFrame(otg_chip_TokenNAK);
		return;
	}

	if (w & otg_ep0_CtrlStall)
	{
		otg_chip_WriteFrame(otg_chip_TokenSTALL);
		return;
	}

	w &= ~otg_ep0_CtrlIn;
#ifdef DEBUG_OTG_CHIP
	info("ReadEp0In: Clear CtrlIn 0x%0x", w);
#endif
	otg_sim_device_WriteFcReg(otg_ep0_Ctrl, w);

	otg_chip_WriteFrame(otg_chip_TokenDATA0);
	w = otg_sim_device_ReadFcReg(otg_ep0_Size);
	w &= otg_ep0_SizeIn;
	w >>= otg_ep0_SizeInLoc;

	if (!w)
	{
		otg_chip_WriteFrame(0);
		otg_chip_WriteFrame(0);
	}
	else
	{
		otg_chip_WriteFrame(w & 0xFF);
		otg_chip_WriteFrame((w >> 8) & 0xFF);

		memory = otg_chip_GetReg(otg_FncRamStart_c) + 16;
		for(i = 0; i < w; i++)
		{
			otg_chip_WriteFrame(otg_chip_ReadMem(memory) & 0xFF);
			memory++;
		}
	}

	w = otg_sim_device_ReadFcReg(otg_ep_intr_Enable);
	if (w & otg_ep_intr_EP0In)
	{
		w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
		w |= otg_ep_intr_EP0In;
		otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
	}
}

static void
otg_sim_device_ReadEpnIn(otg_sim_device_t *dev, int ep)
{
	UINT32 w;
	int i;
	int memory;
	int len;
	int pc;
	int packets;
	int yBuffer;
	int iso;

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnIn: ep=%d", ep);
#endif
	if (ep > 7)
	{
		err("ReadEpnIn: bad endpoint %d", ep);
		return;
	}

	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1));
	if (!(w & otg_ep_ConfigEnable))
	{
#ifdef DEBUG_OTG_CHIP
		info("ReadEpnIn: ep not ready Config=0x%0x", w);
#endif
		/* Endpoint is not enabled yet, send NAK */
		otg_chip_WriteFrame(otg_chip_TokenNAK);
		return;
	}

	otg_sim_device_CheckNewTransfer(dev, ep);

	w = otg_sim_device_ReadFcReg(otg_ep_Config(ep - 1));
	len = (w & otg_ep_ConfigMaxPacketSize) >> otg_ep_ConfigMaxPacketSizeLoc;

	iso = 0;
	if ((w & otg_ep_ConfigType) == otg_ep_ConfigTypeIso1)
	{
		iso = 1;
	}

	w = otg_sim_device_ReadFcReg(otg_ep_Ctrl(ep - 1));
	if (iso)
	{
		packets = 1;
	}
	else
	{
		packets = (w & otg_ep_CtrlPacketPerBuffer) >> otg_ep_CtrlNPLoc;
	}

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnIn: packets=%d len=%d", packets, len);
#endif
	if (!packets)
	{
#ifdef DEBUG_OTG_CHIP
		info("ReadEpnIn: packets is zero");
#endif
		return;
	}

	memory = otg_sim_device_ReadFcReg(otg_ep_Buffer(ep - 1));
	yBuffer = 0;
	if (((int)(dev->packetDone[ep] / packets)) & 1)
	{
		memory >>= otg_ep_BufferYLoc;
		yBuffer = 1;
	}

	memory &= otg_ep_BufferBuffer;
	memory += (dev->packetDone[ep] % packets) * len;

#ifdef DEBUG_OTG_CHIP
	info("ReadEpnIn: memory=0x%0x yBuffer=%d", memory, yBuffer);
#endif
	if (!otg_sim_device_TestFill(ep, yBuffer))
	{
		otg_chip_WriteFrame(otg_chip_TokenNAK);
		return;
	}

	otg_chip_WriteFrame(otg_chip_TokenDATA0);

	if (iso)
	{
		if (yBuffer)
		{
			len = (w & otg_ep_CtrlLength1) >> 16;
			w &= ~otg_ep_CtrlLength1;
		}
		else
		{
			len = w & otg_ep_CtrlLength0;
			w &= ~otg_ep_CtrlLength0;
		}
		pc = 1;
	}
	else
	{
		pc = w & otg_ep_CtrlPacketCount;
		pc--;
		if (!pc)
		{
			len = (w & otg_ep_CtrlLastPacketSize) >> otg_ep_CtrlLPSLoc;
		}

		w &= ~otg_ep_CtrlPacketCount;
		w |= pc;
	}

	otg_sim_device_WriteFcReg(otg_ep_Ctrl(ep - 1), w);

	otg_chip_WriteFrame(len & 0xFF);
	otg_chip_WriteFrame((len >> 8) & 0xFF);

	for (i = 0; i < len; i++)
	{
		otg_chip_WriteFrame(otg_chip_ReadMem(memory));
		memory++;
	}

	dev->packetDone[ep]++;

	w = otg_sim_device_ReadFcReg(otg_ep_intr_Enable);
	if (!pc)
	{
		if (w & otg_ep_intr_EPn(ep))
		{
			w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
			w |= otg_ep_intr_CTRLn(ep) | otg_ep_intr_EPn(ep);
			otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
			otg_sim_device_ClearFill(ep, yBuffer);
		}
	}
	else if (!(dev->packetDone[ep] % packets))
	{
		if (w & otg_ep_intr_EPn(ep))
		{
			w = otg_sim_device_ReadFcReg(otg_ep_intr_Status);
			w |= otg_ep_intr_EPn(ep);
			otg_sim_device_WriteFcReg(otg_ep_intr_Status, w);
		}

		otg_sim_device_ClearFill(ep, yBuffer);
	}
}

void
otg_sim_device_DoFrame(otg_sim_device_t *dev)
{
	UINT8 b;
	int ep;

	b = otg_chip_ReadFrame();
	while (b == otg_chip_TokenSOF)
	{
		otg_chip_ReadFrame();
		otg_chip_ReadFrame();
		b = otg_chip_ReadFrame();
	}

#ifdef DEBUG_OTG_CHIP
	info("DoFrame: token=0x%0x", b);
#endif
	switch (b)
	{
		case otg_chip_TokenOUT:
			ep = otg_sim_device_ReadToken(dev, b);
			if (!ep)
			{
				otg_sim_device_ReadEp0Out(dev);
			}
			else if (ep > 0)
			{
				otg_sim_device_ReadEpnOut(dev, ep);
			}
			else
			{
#ifdef DEBUG_OTG_CHIP
				info("DoFrame: host token incorrect, reply NAK");
#endif
				otg_chip_WriteFrame(otg_chip_TokenNAK);
			}
			break;
		case otg_chip_TokenIN:
			ep = otg_sim_device_ReadToken(dev, b);
			if (!ep)
			{
				otg_sim_device_ReadEp0In(dev);
			}
			else if (ep > 0)
			{
				otg_sim_device_ReadEpnIn(dev, ep);
			}
			else
			{
#ifdef DEBUG_OTG_CHIP
				info("DoFrame: host token incorrect, reply NAK");
#endif
				otg_chip_WriteFrame(otg_chip_TokenNAK);
			}
			break;
		case otg_chip_TokenSETUP:
			ep = otg_sim_device_ReadToken(dev, b);
			otg_sim_device_ReadSetup(dev);
			otg_chip_WriteFrame(otg_chip_TokenACK);
			break;
		case otg_chip_TokenDATA0:
		case otg_chip_TokenDATA1:
		case otg_chip_TokenNAK:
		case otg_chip_TokenSTALL:
		case otg_chip_TokenPRE:
			err("Wrong token 0x%0x", b);
			break;
		case otg_chip_TokenACK:
			break;
		default:
			err("otg_sim_device_DoFrame: Invalid token 0x%0x", b);
	}
}

EXPORT_SYMBOL(otg_sim_device_New);
EXPORT_SYMBOL(otg_sim_device_Delete);
EXPORT_SYMBOL(otg_sim_device_DoFrame);
