/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: otg_td_iso.c
 * Author: Biao Jia		1/21/02
 * Description:
 *    OTG243 TD for iso object handler.
 */
#include "os.h"
#include "stack_urb.h"
#include "otg.h"

#ifdef DEBUG_DONEQUEUE
extern int done_queue_found;
#endif

void
otg_td_iso_FillPacketWord(otg_etd_t *etd)
{
	stack_urb_t *urb;
	int i;
	int max;
	int pos;

	urb = etd->urb;

	max = etd->totalPackets;

	/* Set values from urb */
	for(i = 0, pos = etd->completed; i < max; i++, pos++)
	{
		UINT32 w;

		/* Initially set complete code to not accessed so we can tell */
		/* if the packet has been processed */
		w = stack_urb_GetIsoLength(urb, pos) | otg_td_iso_CCn_NotAccessed_c;
		if (i & 1)
		{
			w <<= otg_td_iso_UpperDwordLoc_c;
			etd->td.iso.data[i>>1] |= w;
		}
		else
		{
			etd->td.iso.data[i>>1] = w;
		}
	}

	for(i = max; i < otg_td_iso_MaxPackets; i++)
	{
		if (!(i & 1))
		{
			etd->td.iso.data[i>>1] = 0;
		}
	}

	/* 0 in otg_td_iso_NumPacketsPerBuffLoc_c means 1 packet per buffer */
	etd->td.iso.ctrl = (stack_urb_GetStartFrame(urb) + etd->completed) & otg_hc_FmMask_c;
	etd->td.iso.ctrl |= ((max - 1) << otg_td_iso_FrameCountLoc_c) | 
		otg_etd_CC_NotAccessed_c;
}

int
otg_td_iso_Fill(otg_etd_t *etd, otg_hc_t *hc)
{
	stack_urb_t *urb;
	int max_length;
	int rc;
	int pos;
#ifdef OTG_ISO_SHARE_MEMORY
	otg_etd_t *prev_etd;
#endif

	urb = etd->urb;
	max_length = stack_urb_GetMaxPacketSize(urb, stack_urb_GetPipe(urb));
	etd->bufferSize = max_length;
	etd->totalPackets = stack_urb_GetPacketCount(urb) - etd->completed;
	if (etd->totalPackets > otg_td_iso_MaxPackets)
	{
		etd->totalPackets = otg_td_iso_MaxPackets;
	}

	pos = otg_hc_GetIsoPosition(hc, etd);
	if (pos < 0)
	{
		/* Something weird return */
		return -1;
	}

#ifdef OTG_ISO_SHARE_MEMORY
	/* If there exist an ETD for this pipe, no need to allocate memory */
	prev_etd = otg_etd_GetPrev(etd);
	if (prev_etd)
	{
		otg_etd_CopyBufferInfo(etd, prev_etd);
#ifdef DEBUG_DONEQUEUE
	if (done_queue_found)
	{
		info("Fill: pos=%d addrX=%d packets=%d lcom=%d etd=%p",
			pos, prev_etd->addrX,
			prev_etd->totalPackets, prev_etd->lastCompleted,
			prev_etd);
	}
#endif
	}
#ifdef DEBUG_ISO
info("Fill: pos=%d prev_etd=%p etd=%p", pos, prev_etd, etd);
#endif
#endif

	if (pos < otg_td_iso_EtdPerPipe)
	{
		rc = otg_hc_Allocate(hc, etd, max_length);
		if (rc)
		{
			return rc;
		}
	}

	stack_urb_SetErrorCount(urb, 0);
	stack_urb_SetActualLength(urb, 0);
	stack_urb_SetStatus(urb, -EINPROGRESS);

	otg_td_iso_FillPacketWord(etd);

	return 0;
}

int
otg_td_iso_DoDoneEtd(otg_etd_t *etd)
{
	int packets;
	int i;
	stack_urb_iso_t *iso;
	stack_urb_t *urb;
	int isInTransfer;
	int total;

	
	packets = otg_td_iso_GetNumPacketFinished(etd);
	total = otg_td_iso_GetTotalPackets(etd);
	urb = etd->urb;
	iso = stack_urb_GetIso(urb, etd->lastCompleted);
#ifdef DEBUG_OTG_SUBMIT
info("DoDoneEtd: etd=%p total=%d",etd, total);
#endif

	if (otg_td_iso_IsLastForUrb(etd) && stack_urb_GetStatus(urb) == -EINPROGRESS)
	{
		stack_urb_SetStatus(urb, otg_NO_ERROR);
	}

	isInTransfer = otg_etd_IsInEtd(etd);
	for(i = 0; i < packets; i++)
	{
		int cc;

		cc = otg_td_iso_GetCompleteCode(etd, i);
		if (cc == otg_etd_CC_NoError_c || cc == 0x9000 )
		{
			/* Successfully finished */
			if (isInTransfer)
			{
				stack_urb_SetIsoActualLength(iso, otg_td_iso_GetLength(etd, i));
			}
			else
			{
				iso->actual_length = iso->length;
				stack_urb_CopyIsoLength(iso);
			}
			stack_urb_SetIsoStatus(iso, otg_NO_ERROR);
		}
		else if (cc == otg_td_iso_CCn_BufferOverRun_c)
		{
			stack_urb_SetIsoActualLength(iso, 0);
			stack_urb_SetIsoStatus(iso, -ENOSR);
			info("DoDoneEtd: ETD%d packet%d buffer overrun", etd->etdIndex, i);
		}
		else
		{
			/* There are failed packets. need to report error */
			stack_urb_SetIsoActualLength(iso, 0);
			stack_urb_SetIsoStatus(iso, -EXDEV);
#if 0
//FB: Just for debugging.
			info("DoDoneEtd: err 0x%0x word2=0x%0x", cc, etd->td.iso.ctrl);
#endif
		}

		stack_urb_IncActualLength(urb, stack_urb_GetIsoActualLength(iso));
		if (stack_urb_GetIsoStatus(iso))
		{
			stack_urb_IncErrorCount(urb);
			if (!stack_urb_IsErrorStatus(urb))
			{
				stack_urb_SetStatus(urb, stack_urb_GetIsoStatus(iso));
			}
		}

#ifdef DEBUG_OTG_SUBMIT
info("DoDoneEtd: i=%d cc=0x%0x status=%d actual_length=%d",
	i, cc, stack_urb_GetIsoStatus(iso), stack_urb_GetIsoActualLength(iso));
#endif
		iso++;
	}

	for(i = packets; i < total; i++)
	{
		stack_urb_SetIsoActualLength(iso, 0);
		stack_urb_SetIsoStatus(iso, otg_NO_ERROR);
		iso++;
	}

	return stack_urb_GetStatus(urb);
}

void
otg_td_iso_NextPacket(otg_etd_t *etd, otg_hc_etd_t *hc_etd)
{
	int hc_addr;
	int length;
	int yBuffer;
	void *addr;

	if (!otg_td_iso_HasMorePacket(etd))
	{
		return;
	}

	yBuffer = otg_td_iso_GetBufferIndex(etd);
	if (otg_etd_IsUnlink(etd))
	{
#ifdef DEBUG_UNLINK
info("NextPacket: unlinked etd=%p", etd);
#endif
		if (otg_etd_IsInEtd(etd))
		{
			otg_hc_etd_ClearFill(hc_etd, etd->etdIndex, yBuffer);
		}
		else
		{
			otg_hc_etd_SetFill(hc_etd, etd->etdIndex, yBuffer);
		}

		otg_td_iso_IncCompletePacket(etd);

		return;
	}

	hc_addr = otg_etd_GetBuffer(etd, yBuffer);
	length = otg_td_iso_GetPacketSize(etd);
	addr = otg_td_iso_GetPacketAddr(etd);
	otg_td_iso_IncCompletePacket(etd);

	if (otg_etd_IsInEtd(etd))
	{
		otg_mem_BulkRead(etd->hc, hc_addr, addr, length);
		otg_hc_etd_ClearFill(hc_etd, etd->etdIndex, yBuffer);
	}
	else
	{
		otg_mem_BulkWrite(etd->hc, hc_addr, addr, length);
		otg_hc_etd_SetFill(hc_etd, etd->etdIndex, yBuffer);
	}

	return;
}

#ifdef OTG_ISO_SHARE_MEMORY
int
otg_td_iso_DoPreDownload(otg_etd_t *etd, otg_hc_etd_t *hc_etd)
{
	int rc;

	if (etd->etdIndex < 0)
	{
		return 0;
	}

	rc = otg_td_iso_CheckActiveFrame(etd);
	if (rc)
	{
		if (rc < 0)
		{
			/* Start frame passed, this ETD should be retired */
			return -1;
		}

		return 0;
	}

	/* Start frame reached */
#ifdef DEBUG_OTG_CHIP
info("DoPreDownload: ");
#endif
	if (!otg_etd_IsInEtd(etd))
	{
		/* For OUT transfer, prev-download the packet */
		otg_td_iso_NextPacket(etd, hc_etd);
	}

	otg_hc_etd_Enable(hc_etd, etd->etdIndex);

	return 1;
}
#endif

#ifdef OTG_ISO_SHARE_MEMORY
int
otg_td_iso_CheckActiveFrame(otg_etd_t *etd)
{
	int frames;

	frames = (etd->td.iso.ctrl - 1 - otg_hc_GetFrameNo(etd->hc))
			& otg_hc_FmMask_c;
#ifdef DEBUG_OTG_CHIP
	info("CheckActiveFrame: frames=0x%0x", frames);
#endif
	if (frames == 0 || frames == (-1 & otg_hc_FmMask_c))
	{
		return 0;
	}

	if (frames > 512)
	{
		return -1;
	}

	return 1;
}
#endif
