/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: stack_urb.h
 * Author: Biao Jia		6/24/02
 * Description:
 *    URB adapter. Convert OS specific URB to otg urb.
 */
#ifndef STACK_URB_H
#define STACK_URB_H

#include <linux/usb.h>
#include "../core/hcd.h"

typedef struct urb stack_urb_t;
typedef struct usb_device stack_device_t;
typedef struct usb_bus stack_bus_t;
/* NB: replaced devrequest with usb_ctrlrequest */
typedef struct usb_ctrlrequest stack_setup_t;
/* NB: replaced iso_packet_descriptor_t with usb_iso_packet_descriptor */
typedef struct usb_iso_packet_descriptor stack_urb_iso_t;

#include "stack_pipe.h"

#define	stack_urb_PipeTypeCtrl	PIPE_CONTROL
#define	stack_urb_PipeTypeBulk	PIPE_BULK
#define	stack_urb_PipeTypeIntr	PIPE_INTERRUPT
#define	stack_urb_PipeTypeIso	PIPE_ISOCHRONOUS

static inline stack_pipe_t
stack_urb_GetPipe(stack_urb_t *urb)
{
	return urb->pipe;
}

static inline int
stack_urb_HasSamePipe(stack_urb_t *urb, stack_urb_t *urb1)
{
	return urb->pipe == urb1->pipe;
}

static inline int
stack_urb_HasSameAddress(stack_urb_t *urb, stack_urb_t *urb1)
{
	return (urb->pipe ^ urb1->pipe) & ~USB_DIR_IN;
}

static inline stack_device_t *
stack_urb_GetDevice(stack_urb_t *urb)
{
	return urb->dev;
}

static inline void
stack_urb_SetDevice(stack_urb_t *urb, stack_device_t *dev)
{
	urb->dev = dev;
}

static inline int
stack_urb_IsDevice(stack_urb_t *urb, stack_device_t *dev)
{
	return dev == urb->dev;
}

static inline int
stack_urb_HasSameDevice(stack_urb_t *urb, stack_urb_t *urb1)
{
	return urb->dev == urb1->dev;
}

static inline stack_bus_t *
stack_urb_GetBus(stack_urb_t *urb)
{
	return urb->dev->bus;
}

static inline int
stack_urb_GetAddress(stack_urb_t *urb)
{
	return usb_pipedevice(urb->pipe);
}

static inline int
stack_urb_GetEndpoint(stack_urb_t *urb)
{
	return usb_pipeendpoint(urb->pipe);
}

static inline int
stack_urb_GetActualLength(stack_urb_t *urb)
{
	return urb->actual_length;
}

static inline void
stack_urb_SetActualLength(stack_urb_t *urb, int len)
{
	urb->actual_length = len;
}

static inline void
stack_urb_IncActualLength(stack_urb_t *urb, int len)
{
	urb->actual_length += len;
}

static inline int
stack_urb_GetBufferLength(stack_urb_t *urb)
{
	return urb->transfer_buffer_length;
}

static inline void *
stack_urb_GetBuffer(stack_urb_t *urb)
{
	return urb->transfer_buffer;
}

static inline int
stack_urb_GetPacketCount(stack_urb_t *urb)
{
	return urb->number_of_packets;
}

static inline int
stack_urb_GetIsoLength(stack_urb_t *urb, int packet)
{
	return urb->iso_frame_desc[packet].length;
}

static inline int
stack_urb_GetIsoOffset(stack_urb_t *urb, int packet)
{
	return urb->iso_frame_desc[packet].offset;
}

static inline stack_urb_iso_t *
stack_urb_GetIso(stack_urb_t *urb, int packet)
{
	return &urb->iso_frame_desc[packet];
}

static inline stack_urb_t *
stack_urb_GetNextUrb(stack_urb_t *urb)
{
#ifdef DEBUG_OTG_STUB
	warn("stack_urb_GetNextUrb stub");
#endif
	//TODO: Do something here: http://www.mail-archive.com/linux-usb-devel@lists.sourceforge.net/msg33214.html
	//return urb->next;
	return NULL;
}

static inline int
stack_urb_GetStartFrame(stack_urb_t *urb)
{
	return urb->start_frame;
}

static inline void
stack_urb_SetStartFrame(stack_urb_t *urb, int frame)
{
	urb->start_frame = frame;
}

static inline void
stack_urb_MaskStartFrame(stack_urb_t *urb, int mask)
{
	urb->start_frame &= mask;
}

static inline void
stack_urb_SetNextStartFrame(stack_urb_t *urb, stack_urb_t *prev)
{
	urb->start_frame = prev->start_frame + prev->number_of_packets;
}

static inline int
stack_urb_GetPipeType(stack_urb_t *urb)
{
	return usb_pipetype(urb->pipe);
}

static inline int
stack_urb_IsCtrlPipe(stack_urb_t *urb)
{
	return usb_pipecontrol(urb->pipe);
}

static inline int
stack_urb_IsBulkPipe(stack_urb_t *urb)
{
	return usb_pipebulk(urb->pipe);
}

static inline int
stack_urb_IsIntrPipe(stack_urb_t *urb)
{
	return usb_pipeint(urb->pipe);
}

static inline int
stack_urb_IsIsoPipe(stack_urb_t *urb)
{
	return usb_pipeisoc(urb->pipe);
}

static inline void
stack_IncDeviceUse(stack_device_t *dev)
{
	usb_get_dev(dev); // NB: was usb_inc_dev_use
}

static inline void
stack_DecDeviceUse(stack_device_t *dev)
{
	usb_put_dev(dev); //NB: was usb_dec_dev_use
}

static inline int
stack_urb_GetStatus(stack_urb_t *urb)
{
	return urb->status;
}

static inline void
stack_urb_SetStatus(stack_urb_t *urb, int status)
{
	urb->status = status;
}

int stack_urb_HasBandwidth(stack_urb_t *urb);

int stack_urb_CheckBandwidth(stack_urb_t *urb);

void stack_urb_ClaimBandwidth(stack_urb_t *urb, int bustime, int is_iso);

void stack_urb_ReleaseBandwidth(stack_urb_t *urb, int is_iso);

static inline int
stack_urb_IsValidInterval(stack_urb_t *urb)
{
	return urb->interval >= 0 && urb->interval <= 255;
}

static inline int
stack_urb_GetIntrInterval(stack_urb_t *urb)
{
	return urb->interval;
}

static inline void
stack_urb_SetIntrInterval(stack_urb_t *urb, int val)
{
	urb->interval = val;
}

/* NB: We don't have the right to call this directly, this is done by
       usb_hcd_giveback_urb. */
/*
static inline usb_complete_t
stack_urb_HasCallBack(stack_urb_t *urb)
{
	return urb->complete;
}

static inline void
stack_urb_CallBack(stack_urb_t *urb)
{
	urb->complete(urb);
}
*/

static inline int
stack_urb_IsIsoASAP(stack_urb_t *urb)
{
	return urb->transfer_flags & URB_ISO_ASAP;
}

static inline int
stack_urb_IsShortPacketAllowed(stack_urb_t *urb)
{
#ifdef DEBUG_OTG_OTHERS
	warn("stack_urb_IsShortPacketAllowed, ret=%d", !(urb->transfer_flags & URB_SHORT_NOT_OK));
#endif
	return !(urb->transfer_flags & URB_SHORT_NOT_OK); //NB: USB_DISABLE_SPD
}

static inline int
stack_urb_IsBulkQueued(stack_urb_t *urb)
{
#ifdef DEBUG_OTG_STUB
	warn("stack_urb_IsBulkQueued stub");
#endif
	//return urb->transfer_flags & URB_QUEUE_BULK; //TODO: Not sure it's available anymore...
	return 0;
}

static inline int
stack_urb_IsErrorStatus(stack_urb_t *urb)
{
	return urb->status && urb->status != -EINPROGRESS;
}

static inline void
stack_urb_AddFlagTimeout(stack_urb_t *urb)
{
#ifdef DEBUG_OTG_STUB
	warn("stack_urb_AddFlagTimeout stub");
#endif
	//urb->transfer_flags |= USB_TIMEOUT_KILLED; //TODO: Not sure it's available anymore...
}

static inline void *
stack_urb_GetContext(stack_urb_t *urb)
{
	return urb->context;
}

static inline void
stack_urb_SetErrorCount(stack_urb_t *urb, int count)
{
	urb->error_count = count;
}

static inline void
stack_urb_IncErrorCount(stack_urb_t *urb)
{
	urb->error_count++;
}

static inline int
stack_urb_GetMaxPacketSize(stack_urb_t *urb, stack_pipe_t pipe)
{
	return usb_maxpacket(urb->dev, pipe, stack_pipe_IsOut(pipe));
}

/* TODO: We might need to merge these two: See http://linmodems.technion.ac.il/archive-fifth/msg00587.html . */
static inline int
stack_urb_IsHalted(stack_urb_t *urb, stack_pipe_t pipe)
{
#ifdef DEBUG_OTG_STUB
	warn("stack_urb_IsHalted stub");
#endif
	//return usb_endpoint_halted(urb->dev, usb_pipeendpoint(pipe), stack_pipe_IsOut(pipe));
	return 0;
}

static inline void
stack_urb_Halt(stack_urb_t *urb)
{
#ifdef DEBUG_OTG_STUB
	warn("stack_urb_Halt stub");
#endif
	//usb_endpoint_halt(urb->dev, usb_pipeendpoint(urb->pipe), stack_pipe_IsOut(urb->pipe));
}
/* end */

static inline int
stack_urb_GetToggle(stack_urb_t *urb, stack_pipe_t pipe)
{
	return usb_gettoggle(urb->dev, usb_pipeendpoint(pipe), usb_pipeout(pipe));
}

static inline void
stack_urb_SetToggle(stack_urb_t *urb, int toggle)
{
	usb_settoggle
	(
		urb->dev,
		usb_pipeendpoint(urb->pipe),
		usb_pipeout(urb->pipe),
		toggle
	);
}

static inline stack_setup_t *
stack_urb_GetSetup(stack_urb_t *urb)
{
	return (stack_setup_t *)urb->setup_packet;
}

static inline int
stack_urb_GetReqestType(stack_setup_t *setup)
{
	return setup->bRequestType;
}

static inline int
stack_urb_GetReqest(stack_setup_t *setup)
{
	return setup->bRequest;
}

static inline int
stack_urb_GetReqestValue(stack_setup_t *setup)
{
	return setup->wValue;
}

static inline int
stack_urb_GetReqestIndex(stack_setup_t *setup)
{
	return setup->wIndex;
}

static inline int
stack_urb_GetReqestLength(stack_setup_t *setup)
{
	return setup->wLength;
}

static inline int
stack_urb_GetIsoActualLength(stack_urb_iso_t *iso)
{
	return iso->actual_length;
}

static inline void
stack_urb_SetIsoActualLength(stack_urb_iso_t *iso, int len)
{
	iso->actual_length = len;
}

static inline void
stack_urb_CopyIsoLength(stack_urb_iso_t *iso)
{
	iso->actual_length = iso->length;
}

static inline int
stack_urb_GetIsoStatus(stack_urb_iso_t *iso)
{
	return iso->status;
}

static inline void
stack_urb_SetIsoStatus(stack_urb_iso_t *iso, int status)
{
	iso->status = status;
}

static inline int
stack_urb_IsOneStepSetup(void)
{
	return 1;
}

#include "otg_linux.h"

#endif /* #ifndef STACK_URB_H */
