/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: storage.c
 * Author: Biao Jia		5/1/02
 * Description:
 *    Mass storage device handler.
 */

#include "os.h"
#include "stack_urb.h"

#include "otg.h"
#include "otg_fd.h"
#include "otg_device_request.h"
#include "test_t.h"

#include "storage_cbw.h"
#include "storage_csw.h"
#include "storage_user.h"

#define storage_BulkInAddr			0x82
#define storage_BulkOutAddr			0x01
#define storage_PkgSize				0x40

#define storage_StageIdle			0
#define storage_StageIn				1
#define storage_StageOut			2
#define storage_StageStatus			3
#define storage_StageStall			4

static test_t *device = 0;
static otg_fdevice_t mass_device;

#define storage_BufferMax			storage_PkgSize

static UINT8 in_buffer[storage_PkgSize];
static UINT8 out_buffer[storage_BufferMax];

static otg_urb_t in_urb;
static otg_urb_t out_urb;
static int in_urb_submitted = 0;
static int out_urb_submitted = 0;

static storage_cbw_t cbw;
static storage_csw_t csw;
static int state = storage_StageIdle;

static int maxLun = 0; /* There is only one LUN */

static int resultLength = 0;
static int set_configuration = 0;

static inline void
storage_SubmitInUrb(void)
{
	int rc;
	unsigned long flags;

	OS_DisableInterrupts(flags);

	if (in_urb_submitted)
	{
		err("SubmitInUrb: ");
		OS_EnableInterrupts(flags);
		return;
	}

	otg_urb_SetStatus(&in_urb, 0);
	in_urb_submitted = 1;
	rc = otg_fd_Submit(&in_urb);
	if (rc)
	{
		err("SubmitInUrb: failed with %d", rc);
		in_urb_submitted = 0;
	}

	OS_EnableInterrupts(flags);
}

static inline void
storage_SubmitOutUrb(int size)
{
	int rc;
	unsigned long flags;

	OS_DisableInterrupts(flags);

	if (out_urb_submitted)
	{
		err("SubmitOutUrb: ");
		OS_EnableInterrupts(flags);
		return;
	}

	otg_urb_SetStatus(&out_urb, 0);
	otg_urb_SetLength(&out_urb, size);
	out_urb_submitted = 1;
	rc = otg_fd_Submit(&out_urb);
	if (rc)
	{
		err("SubmitOutUrb: failed with %d", rc);
		out_urb_submitted = 0;
	}

	OS_EnableInterrupts(flags);
}

static void
storage_DoCommand(void)
{
	int rc;

#ifdef DEBUG_OTG_CHIP
UINT8 *bp;
bp = storage_cbw_GetCommand(&cbw);
info("DoCommand: state=%d", state);
#endif
	rc = storage_user_DoCommand
		 (
		 	storage_cbw_GetCommand(&cbw),
		 	storage_cbw_GetCommandLength(&cbw)
		 );

	if (rc)
	{
		state = storage_StageStall;
		otg_fd_Stall(storage_BulkOutAddr);
		otg_fd_Stall(storage_BulkInAddr);
	}
}

void
storage_RecvData(void)
{
	int len;
	int max;
	int remaining;

#ifdef DEBUG_OTG_CHIP
info("RecvData: state=%d", state);
#endif
	if (out_urb_submitted)
	{
		return;
	}

	max = storage_csw_GetLength(&csw);
	if (max > storage_BufferMax)
	{
		max = storage_BufferMax;
	}

	if (max > otg_urb_GetActualLength(&out_urb))
	{
		max = otg_urb_GetActualLength(&out_urb);
	}

	if (-resultLength < storage_cbw_GetLength(&cbw))
	{
		remaining = storage_cbw_GetLength(&cbw) - storage_csw_GetLength(&csw);
		remaining = -resultLength - remaining;
		if (remaining < 0)
		{
			err("RecvData: negative remaining %d", remaining);
			remaining = 0;
		}

		if (max > remaining)
		{
			max = remaining;
		}
	}

	len = max - otg_urb_GetLength(&out_urb);
	len = storage_user_PutData(&out_buffer[otg_urb_GetLength(&out_urb)], len);
	otg_urb_AddLength(&out_urb, len);
#ifdef DEBUG_OTG_CHIP
info("RecvData: len=%d max=%d", len, max);
#endif
	if (otg_urb_GetLength(&out_urb) < max)
	{
		return;
	}

	storage_csw_AddLength(&csw, otg_urb_GetLength(&out_urb));
	max = storage_csw_GetLength(&csw);
	if (-resultLength < storage_cbw_GetLength(&cbw))
	{
		remaining = storage_cbw_GetLength(&cbw) - storage_csw_GetLength(&csw);
		remaining = -resultLength - remaining;
		if (remaining < 0)
		{
			err("RecvData: remaining %d", remaining);
			remaining = 0;
		}

		if (max > remaining)
		{
			max = remaining;
		}
	}

#ifdef DEBUG_OTG_CHIP
info("RecvData: max=%d", max);
#endif
	if (max)
	{
		if (max > storage_BufferMax)
		{
			max = storage_BufferMax;
		}

		storage_SubmitOutUrb(max);

		return;
	}

	len = -resultLength;
	/* Transfer finished */
	if (len > storage_cbw_GetLength(&cbw))
	{
		/* Case 13: Ho < Do */
		state = storage_StageStall;
		storage_csw_SetStatus(&csw, 2);
		otg_fd_Stall(storage_BulkOutAddr);

		return;
	}
	if (len < storage_cbw_GetLength(&cbw))
	{
		/* Case 9, 11: Ho > Dn or Ho <> Di */
		state = storage_StageStatus;
		otg_fd_Stall(storage_BulkOutAddr);
		return;
	}

	/* Case 12: Ho = Do */
	storage_csw_FillUrb(&csw, &in_urb);
	storage_SubmitInUrb();
	state = storage_StageIdle;
}

void
storage_SendData(void)
{
	int len;
	int max;

	if (in_urb_submitted)
	{
		return;
	}

	max = storage_csw_GetLength(&csw);
	if (resultLength < storage_cbw_GetLength(&cbw))
	{
		max = resultLength - storage_cbw_GetLength(&cbw) + max;
	}

#ifdef DEBUG_OTG_CHIP
info("SendData: max=%d", max);
#endif
	if (max > 0)
	{
		if (max > storage_PkgSize)
		{
			max = storage_PkgSize;
		}

		len = max - otg_urb_GetLength(&in_urb);
		len = storage_user_GetData(&in_buffer[otg_urb_GetLength(&in_urb)], len);
		otg_urb_AddLength(&in_urb, len);
		if (otg_urb_GetLength(&in_urb) >= max)
		{
			storage_SubmitInUrb();
		}

		return;
	}

#ifdef DEBUG_OTG_CHIP
info("SendData: resultLength=%d cbw_Length=%d",
	resultLength, storage_cbw_GetLength(&cbw));
#endif
	/* Transfer finished */
	if (resultLength > storage_cbw_GetLength(&cbw))
	{
		/* Case 7: Hi < Di */
		state = storage_StageStall;
		storage_csw_SetStatus(&csw, 2);
		otg_fd_Stall(storage_BulkInAddr);

		return;
	}
	else if (resultLength < storage_cbw_GetLength(&cbw))
	{
		/* Case 4, 5: Hi > Dn or Hi > Di */
		/* Workaround a host problem */
		if (resultLength % storage_PkgSize)
		{
			storage_csw_FillUrb(&csw, &in_urb);
			storage_SubmitInUrb();
			state = storage_StageIdle;
			return;
		}

		state = storage_StageStatus;
		otg_fd_Stall(storage_BulkInAddr);

		return;
	}

	/* Case 6: Hi = Di */
	storage_csw_FillUrb(&csw, &in_urb);
	storage_SubmitInUrb();
	state = storage_StageIdle;
}

void
storage_CommandDone(int code, int len)
{
	int max;

#ifdef DEBUG_OTG_CHIP
info("CommandDone: state=%d resultLength=%d", state, len);
#endif
	resultLength = len;
	storage_csw_Fill(&csw, &cbw);
	if (state != storage_StageIdle)
	{
		err("CommandDone: not in idle state");
		return;
	}

	if (!storage_cbw_GetLength(&cbw))
	{
		if (len == 0)
		{
			/* case 1: Hn = Dn */
			storage_csw_SetStatus(&csw, code);
		}
		else
		{
			/* case 2, 3: Hn < Di or Hn < Do */
info("CommandDone: resultLength=%d cbw=%d", len, storage_cbw_GetLength(&cbw));
			storage_csw_SetStatus(&csw, 2);
			state = storage_StageStall;
		}

#ifdef DEBUG_OTG_CHIP
info("CommandDone: state=%d Hn", state);
#endif
		if (in_urb_submitted)
		{
			err("CommandDone: in urb already submitted");
			return;
		}

		storage_csw_FillUrb(&csw, &in_urb);
		storage_SubmitInUrb();

		return;
	}

	/* Host is expecting data (IN or OUT) */
	if (storage_cbw_IsIn(&cbw))
	{
#ifdef DEBUG_OTG_CHIP
info("CommandDone: Hi");
#endif
		if (len >= 0)
		{
			/* Case 6: Hi = Di */
			/* Case 4 or Case 5: Hi > Dn or Hi > Di */
			/* Case 7: Hi < Di */
			storage_csw_SetStatus(&csw, code);
			if (in_urb_submitted)
			{
				err("CommandDone: in urb already submitted");
				return;
			}

			state = storage_StageIn;
			otg_urb_SetLength(&in_urb, 0);
			storage_SendData();

			return;
		}

		/* case 8: Hi <> Do */
		state = storage_StageStatus;
		otg_fd_Stall(storage_BulkInAddr);
		storage_csw_SetStatus(&csw, 2);
	}
	else
	{
#ifdef DEBUG_OTG_CHIP
info("CommandDone: Ho");
#endif
		if (len <= 0)
		{
			storage_csw_SetStatus(&csw, code);
			max = -len;
			if (max > storage_cbw_GetLength(&cbw))
			{
				/* Case 12: Ho = Do */
				/* Case 13: Ho < Do */
				max = storage_csw_GetLength(&csw);
			}
			/* else { Case 9, 11: Ho > Dn or Ho <> Di } */

			if (max > storage_BufferMax)
			{
				max = storage_BufferMax;
			}

			state = storage_StageOut;
			storage_SubmitOutUrb(max);

			return;
		}

		/* Case 10: Ho <> Di */
		state = storage_StageStatus;
		storage_csw_SetStatus(&csw, 2);
		otg_fd_Stall(storage_BulkOutAddr);
	}
}

static void
storage_CallBackForIn(otg_urb_t *urb)
{
	unsigned long flags;

	OS_DisableInterrupts(flags);

	in_urb_submitted = 0;

#ifdef DEBUG_OTG_CHIP
info("CallBackForIn: status=%d", urb->status);
#endif
	if (urb->status)
	{
		/* Error, should stall */
		err("CalBackForIn: err=%d", urb->status);
/*
		if (urb->status == otg_urb_ErrorResetByHost)
		{
			storage_SubmitInUrb();
		}
*/

		OS_EnableInterrupts(flags);
		return;
	}

#ifdef DEBUG_OTG_CHIP
info("CallBackForIn: state=%d", state);
#endif
	if (state == storage_StageStall)
	{
		otg_fd_Stall(storage_BulkOutAddr);
		otg_fd_Stall(storage_BulkInAddr);
	}
	else if (state == storage_StageIdle)
	{
		/* For next command */
		storage_SubmitOutUrb(storage_PkgSize);
	}
	else if (state == storage_StageIn)
	{
		storage_csw_AddLength(&csw, otg_urb_GetActualLength(&in_urb));
		if (otg_urb_GetActualLength(&in_urb) != otg_urb_GetLength(&in_urb))
		{
			err("CallBackForIn: partial data sent");
		}

		otg_urb_SetLength(&in_urb, 0);
		storage_SendData();
	}

	OS_EnableInterrupts(flags);
}

static void
storage_CallBackForOut(otg_urb_t *urb)
{
	int rc;
	unsigned long flags;

	OS_DisableInterrupts(flags);

	out_urb_submitted = 0;

#ifdef DEBUG_OTG_CHIP
info("CallBackForOut: status=%d", urb->status);
#endif
	if (set_configuration && state != storage_StageIdle)
	{
		/* Restart */
#ifdef DEBUG_OTG_CHIP
info("CallBackForOut: restart");
#endif
		set_configuration = 0;
		state = storage_StageIdle;
		storage_SubmitOutUrb(storage_PkgSize);
		OS_EnableInterrupts(flags);
		return;
	}

	if (urb->status)
	{
		/* Error, should stall */
		err("CalBackForOut: err=%d", urb->status);
/*
		if (urb->status == otg_urb_ErrorResetByHost)
		{
			storage_SubmitOutUrb(storage_PkgSize);
		}
*/

		OS_EnableInterrupts(flags);
		return;
	}

#ifdef DEBUG_OTG_CHIP
info("CallBackForOut: state=%d", state);
#endif
	if (state == storage_StageStall)
	{
		otg_fd_Stall(storage_BulkOutAddr);
		otg_fd_Stall(storage_BulkInAddr);
		OS_EnableInterrupts(flags);
		return;
	}
	else if (state == storage_StageIdle)
	{
		rc = storage_cbw_Load
			 (
			 	&cbw,
				otg_urb_GetBuffer(urb),
				otg_urb_GetActualLength(urb)
			 );

		if (rc)
		{
			state = storage_StageStall;
			otg_fd_Stall(storage_BulkOutAddr);
			otg_fd_Stall(storage_BulkInAddr);
		}
		else
		{
			storage_DoCommand();
		}
	}
	else if (state == storage_StageOut)
	{
		otg_urb_SetLength(&out_urb, 0);
		storage_RecvData();
	}

	OS_EnableInterrupts(flags);
}

static void
storage_DoClearFeature(int ep)
{
#ifdef DEBUG_OTG_CHIP
info("DoClearFeature: ep=0x%0x state=%d", ep, state);
#endif
	if (state == storage_StageStatus)
	{
		otg_fd_ClearStall(ep);
		state = storage_StageIdle;
		storage_csw_FillUrb(&csw, &in_urb);
		storage_SubmitInUrb();
		return;
	}

	if (state == storage_StageStall)
	{
		otg_fd_Stall(ep);
	}
	else
	{
		otg_fd_ClearStall(ep);
	}
}

static int
storage_StandardRequest(otg_device_request_t *request, void *private)
{
#ifdef DEBUG_OTG_CHIP
info("StandardRequest: bmRequestType=0x%0x bRequest=0x%0x wValue=0x%0x",
	request->bmRequestType, request->bRequest, request->wValue);
#endif
	if (request->bRequest == otg_ep0_RequestSetConfiguration)
	{
		if (otg_ep0_IsDeviceRequest(request->bmRequestType))
		{
#ifdef DEBUG_OTG_CHIP
info("StandardRequest: SetConfiguration");
#endif
			/* SetConfiguration */
			if (!out_urb_submitted)
			{
#ifdef DEBUG_OTG_CHIP
info("StandardRequest: submit out_urb");
#endif
				if (state != storage_StageIdle)
				{
					set_configuration = 1;
				}

				storage_SubmitOutUrb(storage_PkgSize);
			}
		}
	}
	else if (otg_ep0_IsEndpointRequest(request->bmRequestType))
	{
		if (request->bRequest == otg_ep0_RequestClearFeature &&
			request->wValue == otg_ep0_FEATURE_ENDPOINT_HALT)
		{
#ifdef DEBUG_OTG_CHIP
info("StandardRequest: ClearFeature wIndex=0x%0x", request->wIndex);
#endif
			if (request->wIndex == storage_BulkInAddr)
			{
				storage_DoClearFeature(storage_BulkInAddr);
			}
			else if (request->wIndex == storage_BulkOutAddr)
			{
				storage_DoClearFeature(storage_BulkOutAddr);
			}
		}
	}

	request->wLength = 0;
	return 0;
}

static void
storage_DoReset(int intf)
{
#ifdef DEBUG_OTG_CHIP
info("DoReset: ");
#endif
	storage_user_SetReset();

	state = storage_StageIdle;
	otg_fd_ClearStall(storage_BulkInAddr);
	otg_fd_ClearStall(storage_BulkOutAddr);

	storage_SubmitOutUrb(storage_PkgSize);
}

static int
storage_ClassRequest(otg_device_request_t *request, void *private)
{
#ifdef DEBUG_OTG_CHIP
info("ClassRequest: bmRequestType=0x%0x bRequest=0x%0x wValue=0x%0x",
	request->bmRequestType, request->bRequest, request->wValue);
#endif
	if (request->bmRequestType == 0xA1)
	{
		if (request->bRequest == 0xFE && !request->wValue)
		{
#ifdef DEBUG_OTG_CHIP
info("ClassRequest: GetMaxLUN");
#endif
			/* Get Max LUN */
			request->data = &maxLun;
			request->wLength = 1;
/*
			return -1;
*/
		}
	}
	else if (request->bmRequestType == 0x21)
	{
		/* Bulk-Only Mass Storage Reset */
		if (request->bRequest == 0xFF && request->wValue == 0)
		{
			storage_DoReset(request->wIndex);
		}

		request->wLength = 0;
	}
	else
	{
		request->wLength = 0;
	}

	return 0;
}

static int
storage_VendorRequest(otg_device_request_t *request, void *private)
{
	info("VendorRequest: called");
	request->wLength = 0;
	return 0;
}

static int
storage_start(void)
{
	int i;
	int len;
	int strlen;
	int rc;

	if (!device)
	{
		return -1;
	}

	len = 0;
	device->deviceDesc[len++] = otg_device_desc_Max;
	device->deviceDesc[len++] = 0x01; /* bDescriptorType, device */
	device->deviceDesc[len++] = 0x10; /* bcdUSB */
	device->deviceDesc[len++] = 0x01;
	device->deviceDesc[len++] = 0x00; /* bDeviceClass, interface will define */
	device->deviceDesc[len++] = 0x00; /* bDeviceSubClass, ... */
	device->deviceDesc[len++] = 0x00; /* bDeviceProtocol */
	device->deviceDesc[len++] = 0x08; /* bMaxPacketSize0, 8 bytes */
	device->deviceDesc[len++] = 0x20; /* idVendor, TDI 0x0B20 */
	device->deviceDesc[len++] = 0x0B;
	device->deviceDesc[len++] = 0x00; /* idProduct */
	device->deviceDesc[len++] = 0x04;
	device->deviceDesc[len++] = 0x00; /* bcdDevice */
	device->deviceDesc[len++] = 0x01;
	device->deviceDesc[len++] = 0x01; /* iManufacturer */
	device->deviceDesc[len++] = 0x02; /* iProduct */
	device->deviceDesc[len++] = 0x03; /* iSerialNumber */
	device->deviceDesc[len++] = 0x01; /* bNumConfigurations */

	if (len != otg_device_desc_Max)
	{
		return -1;
	}

	len = 0;
	device->configDesc[len++] = otg_config_desc_Max;
	device->configDesc[len++] = 0x02; /* bDescriptorType, configuration */
	device->configDesc[len++] = 0x00; /* wTotalLength */
	device->configDesc[len++] = 0x00;
	device->configDesc[len++] = 0x01; /* bNumInterfaces */
	device->configDesc[len++] = 0x01; /* bConfigurationValue */
	device->configDesc[len++] = 0x00; /* iConfiguration */
	device->configDesc[len++] = 0x00; /* bmAttributes, self powered */
	device->configDesc[len++] = 0x00; /* bMaxPower */

	device->configDesc[len++] = otg_interface_desc_Max;
	device->configDesc[len++] = 0x04; /* bDescriptorType, interface */
	device->configDesc[len++] = 0x00; /* bInterfaceNumber */
	device->configDesc[len++] = 0x00; /* bAlternateSetting */
	device->configDesc[len++] = 0x02; /* bNumEndpoints */
	device->configDesc[len++] = 0x08; /* bInterfaceClass, MassStorage */
	device->configDesc[len++] = 0x06; /* bInterfaceSubClass, SCSI commands */
	device->configDesc[len++] = 0x50; /* bInterfaceProtocol, Bulk Only */
	device->configDesc[len++] = 0x00; /* iInterface */

	device->configDesc[len++] = otg_ep_desc_Max;
	device->configDesc[len++] = 0x05; /* bDescriptorType, endpoint */
	device->configDesc[len++] = storage_BulkOutAddr; /* bEndpointAddress */
	device->configDesc[len++] = 0x02; /* bmAttributes, bulk */
	device->configDesc[len++] = storage_PkgSize; /* wMaxPacketSize */
	device->configDesc[len++] = 0x00;
	device->configDesc[len++] = 0x00; /* bInterval, not used */

	device->configDesc[len++] = otg_ep_desc_Max;
	device->configDesc[len++] = 0x05; /* bDescriptorType, endpoint */
	device->configDesc[len++] = storage_BulkInAddr; /* bEndpointAddress */
	device->configDesc[len++] = 0x02; /* bmAttributes, bulk */
	device->configDesc[len++] = storage_PkgSize; /* wMaxPacketSize */
	device->configDesc[len++] = 0x00;
	device->configDesc[len++] = 0x00; /* bInterval, not used */

	device->configDesc[otg_config_desc_wTotalLength] = len & 0xFF;
	device->configDesc[otg_config_desc_wTotalLength + 1] = (len >> 8) & 0xFF;

	strlen = 0;
	device->strDesc[strlen++] = 4; /* bLength */
	device->strDesc[strlen++] = 3; /* bDescriptorType, string */
	device->strDesc[strlen++] = 0x09; /* LangID codes, English */
	device->strDesc[strlen++] = 0x00;

	i = strlen;
	device->strDesc[strlen++] = 0x00; /* bLength */
	device->strDesc[strlen++] = 0x03; /* bDescriptorType, string */
	device->strDesc[strlen++] = 'T';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'D';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'I';
	device->strDesc[strlen++] = 0x00;

	device->strDesc[i] = strlen - i;

	i = strlen;
	device->strDesc[strlen++] = 0x00; /* bLength */
	device->strDesc[strlen++] = 0x03; /* bDescriptorType, string */
	device->strDesc[strlen++] = 'M';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'a';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 's';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 's';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = ' ';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'S';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 't';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'o';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'r';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'a';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'g';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'e';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = ' ';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'D';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'e';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'v';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'i';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'c';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = 'e';
	device->strDesc[strlen++] = 0x00;

	device->strDesc[i] = strlen - i;

	i = strlen;
	device->strDesc[strlen++] = 0x00; /* bLength */
	device->strDesc[strlen++] = 0x03; /* bDescriptorType, string */
	device->strDesc[strlen++] = 'v';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = '1';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = '.';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = '0';
	device->strDesc[strlen++] = 0x00;
	device->strDesc[strlen++] = '0';
	device->strDesc[strlen++] = 0x00;

	device->strDesc[i] = strlen - i;

	otg_urb_Init
	(
		&in_urb,
		storage_BulkInAddr,
		in_buffer,
		storage_PkgSize,
		storage_CallBackForIn
	);
	otg_urb_SetDevice(&in_urb, &mass_device);

	otg_urb_Init
	(
		&out_urb,
		storage_BulkOutAddr,
		out_buffer,
		storage_PkgSize,
		storage_CallBackForOut
	);
	otg_urb_SetDevice(&out_urb, &mass_device);

	OS_MemSet(&mass_device, 0, sizeof(mass_device));
	mass_device.strDesc = device->strDesc;
	mass_device.strDescLength = strlen;
	mass_device.deviceDesc = device->deviceDesc;
	mass_device.deviceDescLength = otg_device_desc_Max;
	mass_device.configDesc = device->configDesc;
	mass_device.configDescLength = len;
	mass_device.standardSetup = storage_StandardRequest;
	mass_device.classSetup = storage_ClassRequest;
	mass_device.vendorSetup = storage_VendorRequest;

	state = storage_StageIdle;

	rc = otg_fd_Register(&mass_device);
	if (rc < 0)
	{
		return rc;
	}

#ifdef DEBUG_OTG_CHIP
info("start: device started");
#endif
	return 0;
}

static int __init
storage_init(void)
{
	int rc;

	device = OS_Malloc(sizeof(test_t));
	if (!device)
	{
		return -ENOMEM;
	}

	OS_MemSet(device, 0, sizeof(test_t));

	rc = storage_user_Init();
	if (rc)
	{
		OS_Free(device);
		device = 0;
		return rc;
	}

	rc = storage_start();
	if (rc)
	{
		storage_user_Exit();
		OS_Free(device);
		device = 0;

		return rc;
	}

	return 0;
}

static void __exit
storage_exit(void)
{
	storage_user_Exit();

	if (device)
	{
		OS_Free(device);
		device = 0;
	}
}

module_init(storage_init);
module_exit(storage_exit);
