/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: storage_cmd.c
 * Author: Biao Jia		5/3/02
 * Description:
 *    User mode program handles all mass storage commands
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <errno.h>

#include "os_mem.h"
#include "storage_user.h"

#define storage_cmd_Max					16
#define storage_cmd_TestReady			0x00
#define storage_cmd_RequestSense		0x03
#define storage_cmd_Inquiry				0x12
#define storage_cmd_ModeSense			0x1A
#define storage_cmd_StartStop			0x1B
#define storage_cmd_MediumRemoval		0x1E
#define storage_cmd_ReadFormat			0x23
#define storage_cmd_ReadCapacity		0x25
#define storage_cmd_Read				0x28
#define storage_cmd_Write				0x2A

const char *device_name = "/dev/usb_storage";
char command[storage_cmd_Max];

#define storage_DeviceSize						(16 * 1024 * 1024)
#define storage_cmd_BufferMax					4096
const char *device_file = "storage";

const char *log_filename = "storage.log";
FILE *log_file = 0;
int device_fd = -1;
int driver_fd = -1;
UINT8 data_buffer[storage_cmd_BufferMax];

static int media_ready = 0;

UINT8 inquiry_data[64] =
{
	0x00,
	0x80, /* RMB, removable media */
	0x02, /* ANSI version, SPC-2 */
	0x00,
	0x1F, /* Additional Length */
	0x00,
	0x00,
	0x00,
	'T','D','I',' ',' ',' ',' ',' ',
	'M','a','s','s',' ','S','t','o','r','a','g','e',' ',' ',' ',' ',
	'1','6','M','B',
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x30,
	0x30,
	0x30,
	0x30,
	0x30,
	0x30,
	0x30,
	0x30,
	0x30,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
	0x00,
};

#define storage_cmd_SenseDataMax					18
#define storage_cmd_SenseDataValid					0x80
#define storage_cmd_SenseDataErrorCode				0x70
#define storage_cmd_SenseDataCode					0
#define storage_cmd_SenseDataSenseKey				2
#define storage_cmd_SenseDataInfo					3
#define storage_cmd_SenseDataAdditionalLength		7
#define storage_cmd_SenseDataAdditionalCode			12
#define storage_cmd_SenseDataAdditionalQualifier	13

UINT8 sense_data[storage_cmd_SenseDataMax] =
{
	0x70,	/* Error Code */
	0x00,
	0x00,	/* Sense Key */
	0x00, 0x00, 0x00, 0x00,	/* Information */
	0x0B,	/* Additional Sense Length */
	0x00, 0x00, 0x00, 0x00,
	0x00,	/* Additional Sense Code */
	0x00,	/* Additional Sense Code Qualifier */
	0x00,
	0x00, 0x00, 0x00
};

UINT8 capacity_data[8] = /* Total volumn 16MB */
{
	0x00,0x00,0x69,0xFF,	/* Last LBA, 32K  - 1 */
	0x00,0x00,0x02,0x00,	/* Block length in bytes, 512 */
};

UINT8 capacity_list[12] =
{
	0x00,0x00,0x00,
	0x08,					/* Capacity List Length */
	0x00,0x00,0x6A,0x00,	/* Number of blocks */
	0x02,					/* Formatted Media */
	0x00,0x02,0x00,	/* Block length in bytes, 512 */
};

UINT8 mode_sense_data[12] =
{
	0x0B,	/* Mode data length */
	0x00,	/* Medium type */
	0x00,
	0x08,	/* Block descriptor length */
	0x00,	/* Write density */
	0x00,	/* Number of blocks */
	0x6A,
	0x00,
	0x00,
	0x00,	/* Block length in bytes */
	0x02,
	0x00,
};

int
storage_cmd_Open(void)
{
	device_fd = open(device_file, O_RDWR);
	if (device_fd < 0)
	{
		printf("Open %s with error %d\n", device_file, errno);
		return -1;
	}

	log_file = fopen(log_filename, "a");
	if (!log_file)
	{
		close(device_fd);
		device_fd = -1;
		printf("Open %s with error %d\n", log_filename, errno);
		return -1;
	}

	return 0;
}

void
storage_cmd_Close(void)
{
	close(device_fd);
	fclose(log_file);
	log_file = 0;

	device_fd = -1;
}

static void
storage_cmd_DoReadWrite(char *cmd, int len)
{
	int i;
	char *cp;
	int c;
	int rw;
	UINT32 result_len;
	UINT8 *result;
	int rc;
	off_t offset;
	off_t LBA;
	ssize_t size;

	rw = *cmd;
	result = data_buffer;
	c = cmd[7] & 0xFF;
	result_len = c << 8;
	c = cmd[8] & 0xFF;
	result_len |= c;
	result_len <<= 9;

	c = cmd[2] & 0xFF;
	LBA = c << 8;
	c = cmd[3] & 0xFF;
	LBA |= c;
	LBA <<= 8;
	c = cmd[4] & 0xFF;
	LBA |= c;
	LBA <<= 8;
	c = cmd[5] & 0xFF;
	LBA |= c;
	LBA <<= 9;

	rc = 0;
	if (result_len > storage_DeviceSize || LBA >= storage_DeviceSize ||
		result_len + LBA > storage_DeviceSize)
	{
		/* Illegal Request */
		sense_data[storage_cmd_SenseDataSenseKey] = 0x05;
		/* Illegal command operation code */
		sense_data[storage_cmd_SenseDataAdditionalCode] = 0x20;

		result_len = 0;
		rc = 1;
	}
	else
	{
		offset = lseek(device_fd, LBA, SEEK_SET);
		if (offset < 0 || offset != LBA)
		{
			/* Hardware error */
			sense_data[storage_cmd_SenseDataSenseKey] = 0x04;
			/* Illegal command operation code */
			sense_data[storage_cmd_SenseDataAdditionalCode] = 0x00;

			result_len = 0;
			rc = 1;
		}
	}

	if (rc)
	{
		ioctl(driver_fd, storage_user_CommandFailed, 0);
		return;
	}
	else
	{
		if (rw == storage_cmd_Read)
		{
			ioctl(driver_fd, storage_user_CommandSuccess, result_len);
		}
		else
		{
			ioctl(driver_fd, storage_user_CommandSuccess, -result_len);
		}
	}

	if (result_len)
	{
		while (result_len > 0)
		{
			if (result_len > storage_cmd_BufferMax)
			{
				size = storage_cmd_BufferMax;
			}
			else
			{
				size = result_len;
			}

				/* Hardware error */
				/* Illegal command operation code */
			offset = lseek(device_fd, LBA, SEEK_SET);
			if (offset < 0 || offset != LBA)
			{
				sense_data[storage_cmd_SenseDataSenseKey] = 0x04;
				sense_data[storage_cmd_SenseDataAdditionalCode] = 0x00;
			}

			if (rw == storage_cmd_Read)
			{
				size = read(device_fd, data_buffer, size);
				if (size < 0)
				{
					fprintf(log_file, "Read: read returned with %d\n", errno);
					return;
				}
				else if (size == 0)
				{
					fprintf(log_file, "Read: read reach end of file\n", errno);
					return;
				}

				write(driver_fd, data_buffer, size);

				if (size < 0)
				{
					/* Reset, abort */
					err("write failed with %d", size);
					break;
				}
			}
			else
			{
				size = read(driver_fd, data_buffer, size);
				if (size < 0)
				{
					fprintf(log_file, "Read: read returned with %d\n", errno);
					return;
				}
				else if (size == 0)
				{
					fprintf(log_file, "Read: read reach end of file\n", errno);
					return;
				}

				if (size < 0)
				{
					/* Reset, abort */
					err("write failed with %d", size);
					break;
				}

				write(device_fd, data_buffer, size);


/*
				storage_cmd_Close();
				if (storage_cmd_Open())
				{
					exit(-1);
				}
*/
			}

			LBA += size;
			result_len -= size;
		}
	}
}

void
storage_cmd_Do(char *cmd, int len)
{
	int i;
	char *cp;
	int c;
	UINT32 result_len;
	UINT8 *result;
	int rc;

	cp = cmd;
#ifdef STORAGE_LOG_COMMAND
	fprintf(log_file, "Do: command (%d bytes) = ", len);
	for(i = 0; i < len; i++)
	{
		fprintf(log_file, "0x%0x ", *cp);
		cp++;
	}
	fprintf(log_file, "\n");
#endif

	c = *cmd;
	rc = 0;
	if (c != storage_cmd_RequestSense)
	{
		OS_MemSet(sense_data, 0, storage_cmd_SenseDataMax);
		sense_data[storage_cmd_SenseDataCode] = storage_cmd_SenseDataErrorCode;
		sense_data[storage_cmd_SenseDataAdditionalLength] = 0x0B;

		sense_data[0] |= storage_cmd_SenseDataValid;
	}

	if (c == storage_cmd_Inquiry)
	{
		result = inquiry_data;
		i = cmd[4];
		if (i == 36)
		{
			result_len = 36;
			inquiry_data[4] = 31;
		}
		else
		{
			result_len = 56;
			inquiry_data[4] = 51;
		}
	}
	else if (c == storage_cmd_RequestSense)
	{
		/* Do not process this command. Stall */
		result = sense_data;
		result_len = storage_cmd_SenseDataMax;
	}
	else if (c == storage_cmd_ReadCapacity)
	{
		result = capacity_data;
		result_len = 8;
	}
	else if (c == storage_cmd_Read)
	{
		storage_cmd_DoReadWrite(cmd, len);

		return;
	}
	else if (c == storage_cmd_ReadFormat)
	{
		/* Do not process this command. Stall */
		result = 0;
		result_len = 0;


		if (!media_ready)
		{
			media_ready = 1;
			/* Not Ready */
			sense_data[storage_cmd_SenseDataSenseKey] = 0x02;
			/* Media Not Ready */
			sense_data[storage_cmd_SenseDataAdditionalCode] = 0x3A;
			rc = 1;
		}
		else
		{
/*
			sense_data[storage_cmd_SenseDataSenseKey] = 0x05;
			sense_data[storage_cmd_SenseDataAdditionalCode] = 0x20;
			rc = 1;
*/
			result = capacity_list;
			result_len = 12;
		}
	}
	else if (c == storage_cmd_TestReady)
	{
		result = 0;
		result_len = 0;
	}
	else if (c == storage_cmd_ModeSense)
	{
		result = mode_sense_data;
		if (cmd[4] > 12)
		{
			result_len = 12;
		}
		else
		{
			result_len = cmd[4];
		}
	}
	else if (c == storage_cmd_MediumRemoval)
	{
		result = 0;
		result_len = 0;
		if (cmd[4] & 1)
		{
#ifdef STORAGE_LOG_COMMAND
			fprintf(log_file, "Prevent media removal\n");
#endif
			media_ready = 1;
		}
		else
		{
#ifdef STORAGE_LOG_COMMAND
			fprintf(log_file, "Allow media removal\n");
#endif
			media_ready = 0;

			storage_cmd_Close();
			if (storage_cmd_Open())
			{
				exit(-1);
			}
		}
	}
	else if (c == storage_cmd_Write)
	{
		storage_cmd_DoReadWrite(cmd, len);

		return;
	}
	else if (c == storage_cmd_StartStop)
	{
		result_len = 0;
	}
	else
	{
		/* Illegal Request */
		sense_data[storage_cmd_SenseDataSenseKey] = 0x05;
		/* Illegal command operation code */
		sense_data[storage_cmd_SenseDataAdditionalCode] = 0x20;

		fprintf(log_file, "Unknown command 0x%0x\n", c);
		result_len = 0;
		rc = 1;
	}

	if (rc)
	{
		ioctl(driver_fd, storage_user_CommandFailed, 0);
	}
	else
	{
		ioctl(driver_fd, storage_user_CommandSuccess, result_len);
	}

	if (result_len)
	{
		write(driver_fd, result, result_len);
	}
}

int
main(int argc, char *argv[])
{
	int rc;

	driver_fd = open(device_name, O_RDWR);
	if (driver_fd < 0)
	{
		printf("Open device %s with error %d\n", device_name, errno);
		return -1;
	}

	if (storage_cmd_Open())
	{
		close(driver_fd);
		driver_fd = -1;
		return -1;
	}

	for (;;)
	{
		rc = ioctl(driver_fd, storage_user_CommandGetCommand, command);
		if (rc < 0)
		{
			fprintf(log_file, "Ioctl returned with error %d\n", errno);
			break;
		}

		if (rc)
		{
			storage_cmd_Do(command, rc);
		}
	}

	storage_cmd_Close();

	close(driver_fd);
	driver_fd = -1;

	return 0;
}
