/* 
 * Copyright (C) 2002 TransDimension, Inc. All Rights Reserved
 * Name: storage_user.c
 * Author: Biao Jia		5/02/02
 * Description:
 *    Mass storage driver interface to user space.
 */

#include "os.h"
#include "stack_urb.h"

#include "otg.h"
#include "storage.h"
#include "storage_user.h"
#include "otg_buffer.h"

#define storage_user_Major		184
#define storage_user_DataMax	4096

static DECLARE_WAIT_QUEUE_HEAD(storage_user_Wait);
static spinlock_t storage_user_Lock;

static int storage_user_Reset = 0;
static UINT8 *storage_user_Command = 0;
static UINT8 storage_user_CommandLength;

static otg_buffer_t *storage_user_DataBuffer = 0;

static int storage_user_Used = 0;

int
storage_user_DoCommand(UINT8 *cmd, int len)
{
	if (storage_user_Command)
	{
		OS_Free(storage_user_Command);
	}

	storage_user_Command = OS_Malloc(len);
	if (!storage_user_Command)
	{
		return -ENOMEM;
	}
	
	memcpy(storage_user_Command, cmd, len);
	storage_user_CommandLength = len;

#ifdef DEBUG_OTG_CHIP
info("DoCommand: len=%d cmd=0x%0x", len, *cmd);
#endif
	wake_up(&storage_user_Wait);

	return 0;
}

void
storage_user_SetReset(void)
{
	if (!otg_buffer_IsEmpty(storage_user_DataBuffer))
	{
		info("SetReset: readPos=%d writePos=%d",
			storage_user_DataBuffer->readPos,
			storage_user_DataBuffer->writePos);
		if (storage_user_CommandLength)
		{
			info("    cmd=%2x %2x %2x %2x %2x",
				storage_user_Command[0],
				storage_user_Command[1],
				storage_user_Command[2],
				storage_user_Command[3],
				storage_user_Command[4]);
		}
	}

	storage_user_Reset = 1;
	storage_user_CommandLength = 0;

	otg_buffer_Init(storage_user_DataBuffer);
}

void
storage_user_DoReset(void)
{
	storage_user_Reset = 0;
}

int
storage_user_GetData(UINT8 *buf, int len)
{
	int count;

	count = otg_buffer_GetData(storage_user_DataBuffer, buf, len);
	wake_up(&storage_user_Wait);

	return count;
}

int
storage_user_PutData(UINT8 *buf, int len)
{
	int count;

	count = otg_buffer_PutData(storage_user_DataBuffer, buf, len);
	wake_up(&storage_user_Wait);

	return count;
}

static ssize_t
storage_user_Read(struct file *filp, char *buffer, size_t count, loff_t *ppos)
{
	char *cp;	
	int len;
	int max;
	unsigned long flags;

#ifdef DEBUG_OTG_CHIP
info("Read: count=%d", count);
#endif
	spin_lock_irqsave(&storage_user_Lock, flags);

	cp = buffer;
	len = count;
	while(len > 0)
	{
		if (storage_user_Reset)
		{
			storage_user_DoReset();
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -ECONNRESET;
		}

		if (otg_buffer_IsEmpty(storage_user_DataBuffer))
		{
			storage_RecvData();
		}

		if (otg_buffer_IsEmpty(storage_user_DataBuffer))
		{
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			interruptible_sleep_on(&storage_user_Wait);
			spin_lock_irqsave(&storage_user_Lock, flags);

			continue;
		}

		if (signal_pending(current))
		{
			/* May need to inform storage.c */
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -ENODEV;
		}

		max = otg_buffer_GetMaxReadLength(storage_user_DataBuffer);
		if (max > len)
		{
			max = len;
		}

		if (OS_CopyToUser
			(
				cp,
				otg_buffer_GetReadPtr(storage_user_DataBuffer),
				max
			))
		{
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -EFAULT;
		}

		otg_buffer_IncReadPos(storage_user_DataBuffer, max);
		cp += max;
		len -= max;
#ifdef DEBUG_OTG_CHIP
info("Read: cp=%p len=%d max=%d", cp, len, max);
#endif
	}

	spin_unlock_irqrestore(&storage_user_Lock, flags);

	return count - len;
}

static ssize_t
storage_user_Write
(
	struct file *filp,
	const char *buffer,
	size_t count,
	loff_t *ppos
)
{
	const char *cp;	
	int len;
	int max;
	unsigned long flags;

#ifdef DEBUG_OTG_CHIP
info("Write: count=%d", count);
#endif
	spin_lock_irqsave(&storage_user_Lock, flags);

	cp = buffer;
	len = count;
	while(len > 0)
	{
		if (storage_user_Reset)
		{
			storage_user_DoReset();
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -ECONNRESET;
		}

		if (otg_buffer_IsFull(storage_user_DataBuffer))
		{
			storage_SendData();
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			interruptible_sleep_on(&storage_user_Wait);
			spin_lock_irqsave(&storage_user_Lock, flags);

			continue;
		}

		if (signal_pending(current))
		{
			/* May need to inform storage.c */
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -ENODEV;
		}

		max = otg_buffer_GetMaxWriteLength(storage_user_DataBuffer);
		if (max > len)
		{
			max = len;
		}

		if (OS_CopyFromUser
			(
				otg_buffer_GetWritePtr(storage_user_DataBuffer),
				cp,
				max
			))
		{
			spin_unlock_irqrestore(&storage_user_Lock, flags);
			return -EFAULT;
		}

		otg_buffer_IncWritePos(storage_user_DataBuffer, max);
		cp += max;
		len -= max;
#ifdef DEBUG_OTG_CHIP
info("Write: cp=%p len=%d max=%d", cp, len, max);
#endif
	}

	storage_SendData();

	spin_unlock_irqrestore(&storage_user_Lock, flags);

	return count - len;
}

static int
storage_user_Ioctl
(
	struct inode *inode,
	struct file *filp,
	unsigned int cmd,
	unsigned long arg
)
{
	UINT8 *bp;
	unsigned long flags;
	int rc;
	int len;

#ifdef DEBUG_OTG_CHIP
info("Ioctl: cmd=%d", cmd);
#endif
	rc = 0;
	if (cmd == storage_user_CommandGetCommand)
	{
		do
		{
			spin_lock_irqsave(&storage_user_Lock, flags);

			if (storage_user_Reset)
			{
				storage_user_DoReset();
			}

			if (!storage_user_CommandLength)
			{
				spin_unlock_irqrestore(&storage_user_Lock, flags);
				interruptible_sleep_on(&storage_user_Wait);
				spin_lock_irqsave(&storage_user_Lock, flags);
			}

			if (signal_pending(current))
			{
				/* May need to inform storage.c */
				spin_unlock_irqrestore(&storage_user_Lock, flags);
				return -ENODEV;
			}

			bp = (UINT8 *)arg;
			if (OS_CopyToUser
					(
						bp,
						storage_user_Command,
						storage_user_CommandLength
					)
				)
			{
				spin_unlock_irqrestore(&storage_user_Lock, flags);
				return -EFAULT;
			}

			rc = storage_user_CommandLength;
			spin_unlock_irqrestore(&storage_user_Lock, flags);
		} while (!rc);
	}
	else if (cmd == storage_user_CommandSuccess ||
		cmd == storage_user_CommandFailed)
	{
		spin_lock_irqsave(&storage_user_Lock, flags);
		if (storage_user_Reset)
		{
			storage_user_DoReset();
			spin_unlock_irqrestore(&storage_user_Lock, flags);

			return 0;
		}
		else
		{
			/* positive length is for IN and negative is for OUT */
			len = (int)arg;
			if (cmd == storage_user_CommandSuccess)
			{
				storage_CommandDone(0, len);
			}
			else
			{
				storage_CommandDone(1, len);
			}

			storage_user_CommandLength = 0;
		}

		rc = len;

		spin_unlock_irqrestore(&storage_user_Lock, flags);
	}
	else
	{
		return -EINVAL;
	}

	return rc;
}

static int
storage_user_Open(struct inode *inode, struct file *filp)
{
	unsigned long flags;

#ifdef DEBUG_OTG_CHIP
info("Open: ");
#endif
	spin_lock_irqsave(&storage_user_Lock, flags);
	if (storage_user_Used)
	{
		spin_unlock_irqrestore(&storage_user_Lock, flags);
		return -EBUSY;
	}

	storage_user_Used = 1;

	spin_unlock_irqrestore(&storage_user_Lock, flags);

	return 0;
}

static ssize_t
storage_user_Release(struct inode *inode, struct file *filp)
{
	unsigned long flags;

#ifdef DEBUG_OTG_CHIP
info("Release: ");
#endif
	spin_lock_irqsave(&storage_user_Lock, flags);

	storage_user_Used = 0;

	spin_unlock_irqrestore(&storage_user_Lock, flags);

	return 0;
}

static struct file_operations storage_user_fops =
{
	read:		storage_user_Read,
	write:		storage_user_Write,
	ioctl:		storage_user_Ioctl,
	open:		storage_user_Open,
	release:	storage_user_Release,
};

int
storage_user_Init(void)
{
	storage_user_DataBuffer = otg_buffer_New(storage_user_DataMax);
	if (!storage_user_DataBuffer)
	{
		return -ENOMEM;
	}

#if 0 // TODO: fix devfs
	if (devfs_register_chrdev(storage_user_Major,"storage",&storage_user_fops))
	{
		err("unable to get major %d for storage", storage_user_Major);
		return -EBUSY;
	}
#endif

	storage_user_DoReset();
	storage_user_Used = 0;
	storage_user_CommandLength = 0;
	otg_buffer_Init(storage_user_DataBuffer);

	spin_lock_init(&storage_user_Lock);

	return 0;
}

void
storage_user_Exit(void)
{
#if 0 // TODO: fix devfs
	devfs_unregister_chrdev(storage_user_Major, "storage");
#endif
	if (storage_user_Command)
	{
		OS_Free(storage_user_Command);
		storage_user_Command = 0;
	}

	if (storage_user_DataBuffer)
	{
		otg_buffer_Delete(storage_user_DataBuffer);
		storage_user_DataBuffer = 0;
	}
}
